namespace NXyn
  module JSON =
    open System
    open System.Collections
    open System.Collections.Generic

    open NXyn.Aux

    (* FP special cases: +infinity, -infinity, nan *)
    let (|InfNan|Finite|) d = if   Double.IsNaN(d)              then InfNan( "nan")
                              elif Double.IsPositiveInfinity(d) then InfNan( "inf")
                              elif Double.IsNegativeInfinity(d) then InfNan("-inf")
                                                                else Finite(d)

    let (|InfNanOfString|_|) s = if   String.Compare(s, "nan")  = 0 then Some(nan)
                                 elif String.Compare(s, "inf")  = 0 then Some(infinity)
                                 elif String.Compare(s, "-inf") = 0 then Some(-infinity)
                                                                    else None

    (* input helpers *)
    let invalidFmt(s) = raise(FormatException(s))

    let noEOF(c) = if c >= 0 then char(c) else failwith("JSON.noEOF")

    let (ws,isJSONwhitespace) = ([| ' '; '\t'; '\n'; '\r' |], function ' ' | '\t' | '\n' | '\r' -> true | _ -> false)

    let readJSONescapedChar r = function '\"' -> '\"'
                                       | '\\' -> '\\'
                                       | '/'  -> '/'
                                       | 'f'  -> '\f'
                                       | 'n'  -> '\n'
                                       | 'r'  -> '\r'
                                       | 't'  -> '\t'
                                       | 'u'  -> let d = (r:IO.TextReader).Read() |> noEOF |> ConvertMisc.valueOfHexdigit
                                                 let c = r.Read() |> noEOF |> ConvertMisc.valueOfHexdigit
                                                 let b = r.Read() |> noEOF |> ConvertMisc.valueOfHexdigit
                                                 let a = r.Read() |> noEOF |> ConvertMisc.valueOfHexdigit
                                                 char(((d*16 + c)*16 + b)*16 + a)
                                       | c -> failwithf "JSON.readJSONescapedChar: invalid '%c'" c

    let readJSONstr r allowEscapedSlash = let rec loop(sb) = noEOF((r:IO.TextReader).Read())
                                                             |> function '\"' -> (sb:Text.StringBuilder).ToString()
                                                                       | '\\' -> match noEOF(r.Read()) with '/' when not allowEscapedSlash -> failwith("JSON.readJSONstr")
                                                                                                          | c -> loop(sb.Append(readJSONescapedChar r c))
                                                                       | c    -> loop(sb.Append(c))
                                          loop(new Text.StringBuilder())

    let readJSONnum r firstChar = let appendNextChar(sb:Text.StringBuilder) = sb.Append(noEOF((r:IO.TextReader).Read()))
                                  let readIf s readAlso sb = let c = char(r.Peek())
                                                             if String.exists ((=)c) s then readAlso (appendNextChar sb) else sb
                                  let rec readDec n sb = if   Char.IsDigit(char(r.Peek())) then readDec (n-1) (appendNextChar sb)
                                                         elif n < 1                        then sb
                                                                                           else failwith("JSON.readJSONnum")
                                  new Text.StringBuilder()
                                  |> fun sb -> if firstChar = '-' then (noEOF(r.Read()),sb.Append('-')) else (firstChar,sb)
                                               ||> fun c sb -> if not(Char.IsDigit(c)) then invalidFmt("double")
                                                               sb.Append(c)
                                                               |> if c = '0' then id else readDec 0
                                                               |> readIf "." (readDec 1)
                                                               |> readIf "eE" (readIf "+-" id >> readDec 1)
                                                               |> fun sb -> double(sb.ToString())

    (* output helpers *)
    let writeEscapedChar w slash = function '\"' -> (w:IO.TextWriter).Write("\\\""); 2
                                          | '\\' -> w.Write("\\\\"); 2
                                          | '\b' -> w.Write("\\b"); 2
                                          | '\f' -> w.Write("\\f"); 2
                                          | '\t' -> w.Write("\\t"); 2
                                          | '\n' -> w.Write("\\n"); 2
                                          | '\r' -> w.Write("\\r"); 2
                                          | '/'  -> w.Write(slash:string); slash.Length
                                          | c -> if Char.IsControl(c) then fprintf w "\\u%04x" (int(c)); 6 else w.Write(c); 1

    let writeEscapedString w slash s = let mutable n = 0
                                       for i = 0 to String.length s - 1 do n <- n + writeEscapedChar w slash s.[i]
                                       n

    let writeQuotedEscapedString w slash s = (w:IO.TextWriter).Write('\"'); let n = writeEscapedString w slash s in w.Write('\"'); n + 2

    let writeArray w l r write1 xs = (w:IO.TextWriter).Write(l:char)
                                     if xs <> [||] then write1(xs.[0]); for i = 1 to xs.Length - 1 do w.Write(','); write1(xs.[i])
                                     w.Write(r:char)

    (* JSON Value input/output options *)
    type InputOption = AllowEscSlash of bool | AllowComments of bool
    with static member AllowCommentInput(b)      = AllowComments(b)
         static member AllowEscapedSlashInput(b) = AllowEscSlash(b)

         static member TryParse(s) = match s with StringMisc.ArgWithValue("AllowEscapedSlashInput")(StringMisc.BoolOfString(b)) -> Some(InputOption.AllowEscapedSlashInput(b))
                                                | StringMisc.ArgWithValue("AllowCommentInput")(StringMisc.BoolOfString(b))      -> Some(InputOption.AllowCommentInput(b))
                                                | _ -> None

    type InputOptions = { InputEscSlash : bool; InputComments : bool }
    with member opts.AllowEscapedSlashInput = opts.InputEscSlash
         member opts.AllowCommentInput      = opts.InputComments

         static member Default = { InputEscSlash = true; InputComments = true }

         member opts.Set(o:InputOption) = match o with AllowEscSlash(b) -> { opts with InputEscSlash = b }
                                                     | AllowComments(b) -> { opts with InputComments = b }
         member opts.Set(xs:seq<InputOption>) = Seq.fold (fun (os:InputOptions) o -> os.Set(o:InputOption)) opts xs
         member opts.Set(xs:seq<string>) = Seq.fold (fun (os:InputOptions) s -> match InputOption.TryParse(s) with None -> os | Some(o) -> os.Set(o)) opts xs

    type OutputStyle = Compact | Format | Print | PrintJagged
    with static member Default = Compact

    type OutputOption = CompLeft | CompRight | AlignVals | IndentN of int | KVsep of string | Split of int | Nls of string | CompArray of int | CompObject of int | MaxColumns of int | EscapeOutSlash
    with static member AlignValues          = AlignVals
         static member CompressLeft         = CompLeft
         static member CompressRight        = CompRight
         static member EscapeOutputSlash    = EscapeOutSlash
         static member CompressArray(n)     = CompArray(n)
         static member CompressObject(n)    = CompObject(n)
         static member PageWidth(m)         = if m <= 0 then invalidFmt("JSON.OutputOption.PageWidth")   else MaxColumns(m)
         static member NewLines(i)          = if i <  0 then invalidFmt("JSON.OutputOption.NewLines")    else Nls(String.replicate i Environment.NewLine)
         static member SplitDepth(d)        = if d <  0 then invalidFmt("JSON.OutputOption.SplitDepth")  else Split(d)
         static member Indentation(i)       = if i <= 0 then invalidFmt("JSON.OutputOption.Indentation") else IndentN(i)
         static member KeyValueSeparator(s) = let s = readJSONstr (new IO.StringReader(s + "\"")) false
                                              let split = s.Split(ws,StringSplitOptions.RemoveEmptyEntries)
                                              if [|":"|] = split then KVsep(s) else invalidFmt("JSON.OutputOption.KeyValueSeparator")

         static member TryParse(s) = match s with StringMisc.ArgWithoutValue("AlignValues")                                        -> Some(OutputOption.AlignValues)
                                                | StringMisc.ArgWithoutValue("CompressLeft")                                       -> Some(OutputOption.CompressLeft)
                                                | StringMisc.ArgWithoutValue("CompressRight")                                      -> Some(OutputOption.CompressRight)
                                                | StringMisc.ArgWithoutValue("EscapeOutputSlash")                                  -> Some(OutputOption.EscapeOutSlash)
                                                | StringMisc.ArgWithValue("NewLines")(StringMisc.Int32OfString(i))                 -> Some(OutputOption.NewLines(i))
                                                | StringMisc.ArgWithValue("PageWidth")(StringMisc.Int32OfString(i))                -> Some(OutputOption.PageWidth(i))
                                                | StringMisc.ArgWithValue("SplitDepth")(StringMisc.Int32OfString(i))               -> Some(OutputOption.SplitDepth(i))
                                                | StringMisc.ArgWithValue("Indentation")(StringMisc.Int32OfString(i))              -> Some(OutputOption.Indentation(i))
                                                | StringMisc.ArgWithValue("CompressArray")(StringMisc.Int32OfString(i))            -> Some(OutputOption.CompressArray(i))
                                                | StringMisc.ArgWithValue("CompressObject")(StringMisc.Int32OfString(i))           -> Some(OutputOption.CompressObject(i))
                                                | StringMisc.ArgWithValue("KeyValueSeparator")(StringMisc.QuotedStringOfString(s)) -> Some(OutputOption.KeyValueSeparator(s))
                                                | _ -> None

    type OutputOptions = { CompressL : bool; CompressR : bool; Align : bool; Indent : int; KVsep : string; Sdepth : int; Nl : string; SmallArraySize : int; SmallObjSize : int; Slash : string; Maxcols : int }
    with static member Default = { Slash = "/"; Maxcols = 222; CompressL = false; CompressR = false; Align = false; Indent = 4; KVsep = ":"; Sdepth = 9; Nl = Environment.NewLine; SmallArraySize = 0; SmallObjSize = 0 }
         static member Compact = { OutputOptions.Default with CompressL = true; CompressR = true; Align = false; Indent = 2 }

         member opts.AlignValues       = opts.Align
         member opts.CompressLeft      = opts.CompressL
         member opts.CompressRight     = opts.CompressR
         member opts.EscapeOutputSlash = opts.Slash <> "/"
         member opts.CompressArray     = opts.SmallArraySize
         member opts.CompressObject    = opts.SmallObjSize
         member opts.PageWidth         = opts.Maxcols
         member opts.NewLines          = opts.Nl.Length / Environment.NewLine.Length
         member opts.SplitDepth        = opts.Sdepth
         member opts.Indentation       = opts.Indent
         member opts.KeyValueSeparator = opts.KVsep

         member opts.Minimize          = { opts with Align = false; Indent = 1; KVsep =  ":"; Sdepth = 999999; SmallArraySize = 0; SmallObjSize = 0 }

         member opts.Set(o:OutputOption) = match o with Nls(s)         -> { opts with Nl             = s     }
                                                      | KVsep(s)       -> { opts with KVsep          = s     }
                                                      | AlignVals      -> { opts with Align          = true  }
                                                      | Split(d)       -> { opts with Sdepth         = d     }
                                                      | IndentN(i)     -> { opts with Indent         = i     }
                                                      | CompLeft       -> { opts with CompressL      = true  }
                                                      | CompRight      -> { opts with CompressR      = true  }
                                                      | CompObject(i)  -> { opts with SmallObjSize   = i     }
                                                      | CompArray(i)   -> { opts with SmallArraySize = i     }
                                                      | EscapeOutSlash -> { opts with Slash          = "\\/" }
                                                      | MaxColumns(m)  -> { opts with Maxcols        = m     }
         member opts.Set(xs:seq<OutputOption>) = Seq.fold (fun (os:OutputOptions) o -> os.Set(o:OutputOption)) opts xs
         member opts.Set(xs:seq<string>) = Seq.fold (fun (os:OutputOptions) s -> match OutputOption.TryParse(s) with Some(o) -> os.Set(o) | None -> os) opts xs


    (* JSON.Value type definition *)
    [<StructuralEquality; StructuralComparison>]
    type Value = Jtrue | Jfalse                                              (* boolean values *)
               | Jnull                                                       (* null *)
               | Jstr0 | Jstr of string                                      (* empty and non-empty strings *)
               | Jzero | Jone | Jtwo | Jnum of double                        (* floating-point / integer values *)
               | Js0 | Js1 of Value | Js2 of Value * Value | Js of Value []  (* value array *)
               | Jassoc0 | Jassoc of (string * Value) []                     (* association array *)
    with (* principal "constructors" *)
         static member Array(xs)  = match xs with [||] -> Js0
                                                | [|v|] -> Js1(v)
                                                | [|u;v|] -> Js2(u,v)
                                                | vs -> Js(Array.copy vs)
         static member Assoc(kvs) = if kvs = [||] then Jassoc0
                                                  else let d = new Dictionary<string,int>(kvs.Length)
                                                       let a = new ResizeArray<string * Value>(kvs.Length)
                                                       for i = 0 to kvs.Length - 1 do let (k,_) as kv = kvs.[i]
                                                                                      match d.TryGetValue(k) with (true,i)  -> a.[i] <- kv
                                                                                                                | (false,_) -> d.[k] <- a.Count; a.Add(kv)
                                                       Jassoc(a.ToArray())
         static member Bool(b) = if b then Jtrue else Jfalse
         static member Null       = Jnull
         static member Number(d)  = if   Double.IsInfinity(d) || Double.IsNaN(d) then invalidFmt("JSON.Value.Number")
                                    elif d = 0.0 then Jzero
                                    elif d = 1.0 then Jone
                                    elif d = 2.0 then Jtwo
                                                 else Jnum(d)
         static member String(s)  = if   s = null then nullArg("JSON.Value.String")
                                    elif s = ""   then Jstr0
                                                  else Jstr(s)

         (* principal type tests *)
         member v.IsBool   = match v with Jtrue | Jfalse               -> true                  | Jnull | Jstr0 | Jstr _ | Jzero | Jone | Jtwo | Jnum _ | Js0 | Js1 _ | Js2 _ | Js _ | Jassoc0 | Jassoc _ -> false
         member v.IsNull   = match v with Jnull                        -> true | Jtrue | Jfalse         | Jstr0 | Jstr _ | Jzero | Jone | Jtwo | Jnum _ | Js0 | Js1 _ | Js2 _ | Js _ | Jassoc0 | Jassoc _ -> false
         member v.IsString = match v with Jstr0 | Jstr _               -> true | Jtrue | Jfalse | Jnull                  | Jzero | Jone | Jtwo | Jnum _ | Js0 | Js1 _ | Js2 _ | Js _ | Jassoc0 | Jassoc _ -> false
         member v.IsNumber = match v with Jzero | Jone | Jtwo | Jnum _ -> true | Jtrue | Jfalse | Jnull | Jstr0 | Jstr _                                | Js0 | Js1 _ | Js2 _ | Js _ | Jassoc0 | Jassoc _ -> false
         member v.IsArray  = match v with Js0 | Js1 _ | Js2 _ | Js _   -> true | Jtrue | Jfalse | Jnull | Jstr0 | Jstr _ | Jzero | Jone | Jtwo | Jnum _                              | Jassoc0 | Jassoc _ -> false
         member v.IsAssoc  = match v with Jassoc0 | Jassoc _           -> true | Jtrue | Jfalse | Jnull | Jstr0 | Jstr _ | Jzero | Jone | Jtwo | Jnum _ | Js0 | Js1 _ | Js2 _ | Js _                      -> false

         (* direct read-only access to arrays, assoc lists, and strings *)
         member v.Item with get(k:obj) = v.[k.ToString()]
         member v.Item with get(i:int) = match v with Js1(v1)   when i = 0 -> v1
                                                    | Js2(v1,_) when i = 0 -> v1
                                                    | Js2(_,v2) when i = 1 -> v2
                                                    | Js(xs) -> xs.[i]
                                                    | Jstr(s) -> Jstr(string(s.[i]))  (* don't depend on Value.Char here *)
                                                    | Jassoc _ -> v.[string(i)]
                                                    | Jassoc0 | Js0 | Jfalse | Jnull | Jstr0 | Jzero | Jone | Jtwo -> failwith "JSON.Value.Item<System.Int32>.get"
                                                    | Jstr _ | Js1 _ | Js2 _ | Jtrue | Jnum _ -> failwith "JSON.Value.Item<System.Int32>.get"
         member v.Item with get(s:string) = match v with Jassoc(kvs) -> Array.pick (fun(k,v) -> if k = s then Some(v) else None) kvs
                                                       | Js1 _ | Js2 _ | Js _ | Jstr _ -> v.[int(s)]
                                                       | Jassoc0 | Js0 | Jfalse | Jtrue | Jnull | Jstr0 | Jzero | Jone | Jtwo | Jnum _ -> failwith "JSON.Value.Item<System.String>.get"

         (* get maximum depth of JSON.Value tree *)
         member v.GetMaximumDepth() = match v with Js0 | Jassoc0 | Jnull | Jzero | Jone | Jtwo | Jnum _ | Jtrue | Jfalse | Jstr0 | Jstr _ -> 0
                                                 | Js1(v1)     -> v1.GetMaximumDepth() + 1
                                                 | Js2(v1,v2)  -> max (v1.GetMaximumDepth()) (v2.GetMaximumDepth()) + 1
                                                 | Js(vs)      -> 1 + Array.fold (fun m   (v:Value) -> max m (v.GetMaximumDepth())) (-1)  vs
                                                 | Jassoc(kvs) -> 1 + Array.fold (fun m (_,v:Value) -> max m (v.GetMaximumDepth())) (-1) kvs

         (* normalize JSON.Value *)
         member v.GetNormalized() = match v with Js1(v1)     -> Js1(v1.GetNormalized())
                                               | Js2(v1,v2)  -> Js2(v1.GetNormalized(), v2.GetNormalized())
                                               | Js(vs)      -> Js(Array.map (fun(v:Value) -> v.GetNormalized()) vs)
                                               | Jassoc(kvs) -> Jassoc(Array.sortBy fst (Array.map (fun(k,v:Value) -> (k,v.GetNormalized())) kvs))
                                               | Jassoc0 | Js0 | Jtrue | Jfalse | Jnull | Jstr0 | Jzero | Jone | Jtwo | Jstr _ | Jnum _ -> v

         (* return all subnodes of some JSON.Value *)
         member v.GetContents() = seq { yield v; match v with Jnull | Jzero | Jone | Jtwo | Jnum _ | Jtrue | Jfalse | Jstr0 | Jstr _ | Jassoc0 | Js0 -> ()
                                                            | Js1(v1)     -> yield! (v1.GetContents())
                                                            | Js2(v1,v2)  -> yield! (v1.GetContents()); yield! (v2.GetContents())
                                                            | Js(vs)      -> for v in vs do yield! (v.GetContents())
                                                            | Jassoc(kvs) -> for (k,v:Value) in kvs do yield (Value.String(k)); yield! (v.GetContents()) }

         (* get all key-value pairs in an JSON.Value *)
         member v.GetPairs() = seq { match v with Jnull | Jtrue | Jfalse | Jzero | Jone | Jtwo | Jnum _ | Jstr0 | Jstr _ | Js0 | Jassoc0 -> ()
                                                | Js1(v1)     -> yield! (v1.GetPairs())
                                                | Js2(v1,v2)  -> yield! (v1.GetPairs()); yield! (v2.GetPairs())
                                                | Js(vs)      -> for v in vs do yield! (v.GetPairs())
                                                | Jassoc(kvs) -> for (k:string,v:Value) in kvs do yield(k,v); yield! (v.GetPairs()) }

         (* return length of an JSON.Value that can be accessed with "Item" (array, assoc list, string) *)
         member v.Length = match v with Jstr0 | Jassoc0 | Js0 -> 0
                                      | Js1(_) -> 1
                                      | Js2(_) -> 2
                                      | Jassoc(kvs) -> kvs.Length
                                      | Js(vs) -> vs.Length
                                      | Jstr(s) -> s.Length
                                      | Jnull | Jtrue | Jfalse | Jzero | Jone | Jtwo | Jnum _ -> failwith "JSON.Value.Length"

         (* compact output, no additional formatting/indentation *)
         member x.OutputI(w,opts) = match x with Jstr(s)     -> writeQuotedEscapedString w opts.Slash s |> ignore
                                               | Jstr0       -> w.Write("\"\"")
                                               | Jtrue       -> w.Write("true")
                                               | Jfalse      -> w.Write("false")
                                               | Jnull       -> w.Write("null")
                                               | Js0         -> w.Write("[]")
                                               | Js1(v)      -> w.Write('['); v.OutputI(w,opts); w.Write(']')
                                               | Js2(u,v)    -> w.Write('['); u.OutputI(w,opts); w.Write(','); v.OutputI(w,opts); w.Write(']')
                                               | Jzero       -> w.Write('0')
                                               | Jone        -> w.Write('1')
                                               | Jtwo        -> w.Write('2')
                                               | Jassoc0     -> w.Write("{}")
                                               | Jnum(d)     -> w.Write(d)
                                               | Js(vs)      ->  vs |> writeArray w '[' ']' (fun(v:Value) -> v.OutputI(w,opts))
                                               | Jassoc(kvs) -> kvs |> writeArray w '{' '}' (fun(k,v:Value) -> writeQuotedEscapedString w opts.Slash k |> ignore
                                                                                                               w.Write(opts.KVsep)
                                                                                                               v.OutputI(w,opts))
         override v.ToString() = let w = new IO.StringWriter() in v.OutputI(w,OutputOptions.Compact); w.ToString()

    (* principal views *)
    let (|Array|Assoc|Bool|Null|Number|String|) = function Jnull      -> Null
                                                         | Jtrue      -> Bool(true)
                                                         | Jfalse     -> Bool(false)
                                                         | Jzero      -> Number(0.0)
                                                         | Jone       -> Number(1.0)
                                                         | Jtwo       -> Number(2.0)
                                                         | Jnum(d)    -> Number(d)
                                                         | Jassoc0    -> Assoc([||])
                                                         | Jassoc(xs) -> Assoc(Array.copy xs)
                                                         | Js0        -> Array([||])
                                                         | Js1(v)     -> Array([|v|])
                                                         | Js2(v1,v2) -> Array([|v1;v2|])
                                                         | Js(xs)     -> Array(Array.copy xs)
                                                         | Jstr0      -> String("")
                                                         | Jstr(s)    -> String(s)
    let Array_of_Value  = function Array(x) -> x            | Assoc _ | Bool _ | Null | Number _ | String _ -> failwith "JSON.Array_of_Value"
    let Assoc_of_Value  = function Assoc(x) -> x  | Array _           | Bool _ | Null | Number _ | String _ -> failwith "JSON.Assoc_of_Value"
    let Bool_of_Value   = function Bool(x) -> x   | Array _ | Assoc _          | Null | Number _ | String _ -> failwith "JSON.Bool_of_Value"
    let Null_of_Value   = function Null -> null   | Array _ | Assoc _ | Bool _        | Number _ | String _ -> failwith "JSON.Null_of_Value"
    let Number_of_Value = function Number(x) -> x | Array _ | Assoc _ | Bool _ | Null            | String _ -> failwith "JSON.Number_of_Value"
    let String_of_Value = function String(x) -> x | Array _ | Assoc _ | Bool _ | Null | Number _            -> failwith "JSON.String_of_Value"

    type Value with static member MappedArray(m,xs) = match xs with [||] -> Js0
                                                                  | [|v|] -> Js1(m(v))
                                                                  | [|v1;v2|] -> Js2(m(v1),m(v2))
                                                                  | vs -> Js(Array.map m vs)
    let                         (|MappedArray|_|) m = function Js0        -> Some([||])
                                                             | Js1(v1)    -> Some([| m(v1) |])
                                                             | Js2(v1,v2) -> Some([| m(v1); m(v2) |])
                                                             | Js(vs)     -> Some(Array.map m vs)
                                                             | Jstr0 | Jassoc0 | Jassoc _ | Jstr _ | Jnull | Jtrue | Jfalse | Jzero | Jone | Jtwo | Jnum _ -> None
    type Value with    member v.IsMappedArray = v.IsArray
    let                           MappedArray_of_Value m = function MappedArray m xs -> xs | _ -> failwith "JSON.MappedArray_of_Value"

    (* secondary JSON.Value constructors and views: may depend on principal ones *)
    type Value with static member BigInt(i)         = Jstr(string(i:bigint))
    let                         (|BigInt|_|)        = function Jstr(s) -> begin try Some(Aux.ConvertMisc.base10bigintconv.Parse(s)) with _ -> None end | _ -> None
    type Value with    member v.IsBigInt            = match v with BigInt(_) -> true | _ -> false
    let                           BigInt_of_Value   = function BigInt(i) -> i | v -> failwithf "JSON.BigInt_of_Value %A" v

    type Value with static member BitArray(b)       = if (b:Collections.BitArray).Count = 0 then Jstr0 else Jstr(ConvertMisc.String_of_BitArray(b))
    let                         (|BitArray|_|)      = function Jstr0 -> Some(new BitArray(0)) | Jstr(s) -> begin try Some(ConvertMisc.BitArray_of_String(s)) with _ -> None end | _ -> None
    type Value with    member v.IsBitArray          = match v with BitArray(_) -> true | _ -> false
    let                           BitArray_of_Value = function BitArray(i) -> i | v -> failwithf "JSON.BitArray_of_Value(%A)" v

    type Value with static member Byte(u)           = if u = 0uy then Jzero elif u = 1uy then Jone elif u = 2uy then Jtwo else Jnum(double(u))
    let                         (|Byte|_|)          = function Jzero -> Some 0uy | Jone -> Some 1uy | Jtwo -> Some 2uy | Jnum d -> begin try Some(byte(d)) with _ -> None end | _ -> None
    type Value with    member v.IsByte              = match v with Byte(_) -> true | _ -> false
    let                           Byte_of_Value     = function Byte(i) -> i | v -> failwithf "JSON.Byte_of_Value(%A)" v

    type Value with static member Char(c)           = Jstr(string(c:char))
    let                         (|Char|_|)          = function Jstr(s) when s.Length > 0 -> Some(s.[0]) | _ -> None
    type Value with    member v.IsChar              = match v with Char(_) -> true | _ -> false
    let                           Char_of_Value     = function Char(i) -> i | v -> failwithf "JSON.Char_of_Value(%A)" v

    type Value with static member Decimal(d)        = Jstr(string(d:decimal))
    let                         (|Decimal|_|)       = function Jstr(s) -> begin try Some(decimal(s)) with _ -> None end | _ -> None
    type Value with    member v.IsDecimal           = match v with Decimal(_) -> true | _ -> false
    let                           Decimal_of_Value  = function Decimal(i) -> i | v -> failwithf "JSON.Decimal_of_Value(%A)" v

    type Value with static member Double(f)         = match (f:double) with InfNan(s) -> Value.String(s) | Finite(d) -> Value.Number(d)
    let                         (|Double|_|)        = function Number(d) -> Some(d) | String(InfNanOfString(k)) -> Some(k) | _ -> None
    type Value with    member v.IsDouble            = match v with Double(_) -> true | _ -> false
    let                           Double_of_Value   = function Double(i) -> i | v -> failwithf "JSON.Double_of_Value(%A)" v

    type Value with static member False             = Jfalse
    let                         (|False|_|)         = function Jfalse -> Some() | _ -> None
    type Value with    member v.IsFalse             = match v with False -> true | _ -> false
    let                           False_of_Value    = function False -> false | v -> failwithf "JSON.False_of_Value(%A)" v

    type Value with static member Int16(s)          = if s = 0s then Jzero elif s = 1s then Jone elif s = 2s then Jtwo else Jnum(double(s))
    let                         (|Int16|_|)         = function Jzero -> Some 0s | Jone -> Some 1s | Jtwo -> Some(2s) | Jnum(s) -> begin try Some(int16(s)) with _ -> None end | _ -> None
    type Value with    member v.IsInt16             = match v with Int16(_) -> true | _ -> false
    let                           Int16_of_Value    = function Int16(i) -> i | v -> failwithf "JSON.Int16_of_Value(%A)" v

    type Value with static member Int32(s)          = if s = 0 then Jzero elif s = 1 then Jone elif s = 2 then Jtwo else Jnum(double(s))
    let                         (|Int32|_|)         = function Jzero -> Some 0 | Jone -> Some 1 | Jtwo -> Some(2) | Jnum(s) -> begin try Some(int(s))   with _ -> None end | _ -> None
    type Value with    member v.IsInt32             = match v with Int32(_) -> true | _ -> false
    let                           Int32_of_Value    = function Int32(i) -> i | v -> failwithf "JSON.Int32_of_Value(%A)" v

    type Value with static member Int64(s)          = Jstr(string(s:int64))
    let                         (|Int64|_|)         = function Jstr(s) -> begin try Some(int64(s)) with _ -> None end | _ -> None
    type Value with    member v.IsInt64             = match v with Int64(_) -> true | _ -> false
    let                           Int64_of_Value    = function Int64(i) -> i | v -> failwithf "JSON.Int64_of_Value(%A)" v

    type Value with static member SByte(s)          = if s = 0y then Jzero elif s = 1y then Jone elif s = 2y then Jtwo else Jnum(double(s))
    let                         (|SByte|_|)         = function Jzero -> Some 0y | Jone -> Some 1y | Jtwo -> Some(2y) | Jnum(s) -> begin try Some(sbyte(s)) with _ -> None end | _ -> None
    type Value with    member v.IsSByte             = match v with SByte(_) -> true | _ -> false
    let                           SByte_of_Value    = function SByte(i) -> i | v -> failwithf "JSON.SByte_of_Value(%A)" v

    type Value with static member String0(s)        = if s = null then Value.Null else Value.String(s)
    let                         (|String0|_|)       = function String(s) -> Some(s) | Null -> Some(null) | _ -> None
    type Value with    member v.IsString0           = match v with String0(_) -> true | _ -> false
    let                           String0_of_Value  = function String0(i) -> i | v -> failwithf "JSON.String0_of_Value(%A)" v

    type Value with static member True              = Jtrue
    let                         (|True|_|)          = function Jtrue -> Some() | _ -> None
    type Value with    member v.IsTrue              = match v with True -> true | _ -> false
    let                           True_of_Value     = function True -> true | v -> failwithf "JSON.True_of_Value(%A)" v

    type Value with static member UInt16(u)         = if u = 0us then Jzero elif u = 1us then Jone elif u = 2us then Jtwo else Jnum(double(u))
    let                         (|UInt16|_|)        = function Jzero -> Some 0us | Jone -> Some 1us | Jtwo -> Some 2us | Jnum d -> begin try Some(uint16(d)) with _ -> None end | _ -> None
    type Value with member    v.IsUInt16            = match v with UInt16(_) -> true | _ -> false
    let                           UInt16_of_Value   = function UInt16(i) -> i | v -> failwithf "JSON.UInt16_of_Value(%A)" v

    type Value with static member UInt32(u)         = if u = 0u then Jzero elif u = 1u then Jone elif u = 2u then Jtwo else Jnum(double(u))
    let                         (|UInt32|_|)        = function Jzero -> Some 0u | Jone -> Some 1u | Jtwo -> Some 2u | Jnum d -> begin try Some(uint32(d)) with _ -> None end | _ -> None
    type Value with    member v.IsUInt32            = match v with UInt32(_) -> true | _ -> false
    let                           UInt32_of_Value   = function UInt32(i) -> i | v -> failwithf "JSON.UInt32_of_Value(%A)" v

    type Value with static member UInt64(u)         = Jstr(string(u:uint64))
    let                         (|UInt64|_|)        = function Jstr(s) -> begin try Some(uint64(s)) with _ -> None end | _ -> None
    type Value with    member v.IsUInt64            = match v with UInt64(_) -> true | _ -> false
    let                           UInt64_of_Value   = function UInt64(i) -> i | v -> failwithf "JSON.UInt64_of_Value(%A)" v

    (* tertiary constructors/views: may depend on primary and secondary ones *)
    type Value with static member Single(f)         = Value.Double(double(f:single))
    let                         (|Single|_|)        = function Double(d) -> begin try Some(single(d)) with _ -> None end | _ -> None
    type Value with    member v.IsSingle            = match v with Single(_) -> true | _ -> false
    let                           Single_of_Value   = function Single(i) -> i | v -> failwithf "JSON.Single_of_Value(%A)" v


    (* input JSON values from TextReader/String *)
    let readI(r,o) = let next() = (r:IO.TextReader).Read() |> ignore
                     let eatWhile(test) = while test(noEOF(r.Peek())) do next()
                     let rec eatWS() = match noEOF(r.Peek()) with '/' when o.InputComments -> next(); eatComment(); next(); eatWS()
                                                                | '/' -> failwith "JSON.Value.Read: support of comment disabled"
                                                                | c -> if isJSONwhitespace(c) then next(); eatWS()
                     and eatComment() = match noEOF(r.Read()) with '/' -> eatWhile((<>)'\n')
                                                                 | '*' -> while noEOF(r.Peek()) <> '/' do eatWhile((<>)'*'); next()
                                                                 | c -> failwithf "JSON.eatWS: expected '/' or '*', got '%c'" c
                     let eat1(c) = let c0 = noEOF(r.Read()) in if c <> c0 then failwithf "JSON.eat1: expected '%c', got '%c'" c c0
                     let eatN(s) = for i = 0 to String.length s - 1 do eat1(s.[i])
                     let rec readPair() = eat1('\"'); let k = readJSONstr r o.InputEscSlash in eatWS(); eat1(':'); eatWS(); (k,readVal())
                     and readVals chClose read1 = let is_closing() = chClose = noEOF(r.Peek())
                                                  let rec loop() = read1(); eatWS(); if is_closing() then next() else eat1(','); eatWS(); loop()
                                                  eatWS(); if is_closing() then next() else loop()
                     and readArray chClose read1 = let a = new ResizeArray<'a>() in readVals chClose (read1 >> a.Add); a.ToArray()
                     and readVal() = match noEOF(r.Read()) with '\"' -> Value.String(readJSONstr r o.InputEscSlash)
                                                              | 'n'  -> eatN("ull");  Jnull
                                                              | 't'  -> eatN("rue");  Jtrue
                                                              | 'f'  -> eatN("alse"); Jfalse
                                                              | '['  -> Value.Array(readArray ']' readVal)
                                                              | '{'  -> Value.Assoc(readArray '}' readPair) (* keep original order *)
                                                              | chr  -> Jnum(readJSONnum r chr)
                     eatWS(); readVal()

    let parseI(s,o) = readI(new IO.StringReader(s), o)

    (* formatted output/pretty printing of JSON.Value to TextWriter *)
    let maxEscapedKeyLength slash xs = let mutable mx = 0
                                       for i = 0 to Array.length xs - 1 do mx <- max mx (writeQuotedEscapedString IO.StreamWriter.Null slash (fst xs.[i]))
                                       mx

    let alignedKeyLength o xs = if o.Align then maxEscapedKeyLength o.Slash xs else 0 (* 0 = dummy value *)

    let indent (w:IO.TextWriter) n = for i = 1 to n do w.Write(' ')

    let showClosing w o (c,xs) = if xs <> [] then (w:IO.TextWriter).Write(o.Nl); ignore(List.fold (fun i0 i -> indent w (i-i0); w.Write(c:char); i+1) 0 xs)

    let addClosing w o c0 (l0:int) (c,ps) = (c0,l0::(if o.CompressR && c = c0 then ps else (showClosing w o (c,ps); [])))

    let formatMany w o chOpen chClose l0 format1 xs = let l1 = l0+o.Indent
                                                      (w:IO.TextWriter).Write(chOpen:char)
                                                      indent w (if o.CompressL then o.Indent-1 else w.Write(o.Nl); l1)
                                                      for i = 0 to Array.length xs - 2 do showClosing w o (format1 l1 xs.[i])
                                                                                          w.Write(',')
                                                                                          w.Write(o.Nl)
                                                                                          indent w l1
                                                      format1 l1 xs.[xs.Length-1] |> addClosing w o chClose l0

    let formatObj1 w o mx format1 l1 (k,v) = let k_len = writeQuotedEscapedString w o.Slash k
                                             let len = if o.Align then (indent w (mx - k_len); mx) else k_len
                                             w.Write(o.KVsep)
                                             format1 (l1 + o.KVsep.Length + len) v

    let formatI(v,w,o) = let rec formatVal d1 l0 = function Array(xs) when d1 > 0 && xs.Length > o.SmallArraySize -> formatMany w o '[' ']' l0 (formatVal (d1-1)) xs
                                                          | Assoc(xs) when d1 > 0 && xs.Length > o.SmallObjSize   -> formatMany w o '{' '}' l0 (formatObj1 w o (alignedKeyLength o xs) (formatVal (d1-1))) xs
                                                          | v -> v.OutputI(w,o); (' ',[])
                         showClosing w o (formatVal o.Sdepth 0 v)

    let fitsI(v,o) = try  (v:Value).OutputI(new Aux.LimitedTextWriter.ExnTextWriter(o.Maxcols),o); true
                     with Failure(_) -> false

    let formatJaggedI(v,w,o) = let rec formatVal l0 x = match x with Array(xs) when not(fitsI(x,{o with Maxcols = o.Maxcols - l0})) -> formatMany w o '[' ']' l0 formatVal xs
                                                                   | Assoc(xs) when not(fitsI(x,{o with Maxcols = o.Maxcols - l0})) -> formatMany w o '{' '}' l0 (formatObj1 w o (alignedKeyLength o xs) formatVal) xs
                                                                   | v -> v.OutputI(w,o); (' ',[])
                               showClosing w o (formatVal 0 v)

    let tuneSplitDepthI(v,o) = let check(d) = try  let o = {o with Sdepth = d}
                                                   formatI(v,new LimitedTextWriter.ExnTextWriter(o.Maxcols),o)
                                                   Some(d,o)
                                              with Failure(_) -> None
                               let ub = v.GetMaximumDepth()
                               let rec loop(d) = match check(d) with Some(_,o)        -> Some(o)
                                                                   | None when d < ub -> loop(d+1)
                                                                   | None             -> None
                               match loop(0) with None    -> (o.Minimize, false)
                                                | Some(o) -> (o, true)

    let writeI(v,s,w,o) = match s with Compact     -> (v:Value).OutputI(w,o)
                                     | Format      -> formatI(v,w,o)
                                     | PrintJagged -> formatJaggedI(v,w,o)
                                     | Print       -> formatI(v,w,fst(tuneSplitDepthI(v,o)))
    let writeLineI(v,s,w,o) = writeI(v,s,w,o); w.Write(o.Nl)

    (* overloaded methods *)
    type Value with static member Parse(s:string               ) = parseI(s, InputOptions.Default)
                    static member Parse(s:string,o:InputOptions) = parseI(s,                    o)

                    static member Read(                              ) = readI(Console.In, InputOptions.Default)
                    static member Read(r:IO.TextReader               ) = readI(r,          InputOptions.Default)
                    static member Read(                o:InputOptions) = readI(Console.In,                    o)
                    static member Read(r:IO.TextReader,o:InputOptions) = readI(         r,                    o)

                    member v.Write(                                             ) = writeI(v, OutputStyle.Default, Console.Out, OutputOptions.Default)
                    member v.Write(                              o:OutputOptions) = writeI(v, OutputStyle.Default, Console.Out,                     o)
                    member v.Write(              w:IO.TextWriter                ) = writeI(v, OutputStyle.Default,           w, OutputOptions.Default)
                    member v.Write(              w:IO.TextWriter,o:OutputOptions) = writeI(v, OutputStyle.Default,           w,                     o)
                    member v.Write(s:OutputStyle                                ) = writeI(v,                   s, Console.Out, OutputOptions.Default)
                    member v.Write(s:OutputStyle,                o:OutputOptions) = writeI(v,                   s, Console.Out,                     o)
                    member v.Write(s:OutputStyle,w:IO.TextWriter                ) = writeI(v,                   s,           w, OutputOptions.Default)
                    member v.Write(s:OutputStyle,w:IO.TextWriter,o:OutputOptions) = writeI(v,                   s,           w,                     o)

                    member v.WriteLine(                                             ) = writeLineI(v, OutputStyle.Default, Console.Out, OutputOptions.Default)
                    member v.WriteLine(                              o:OutputOptions) = writeLineI(v, OutputStyle.Default, Console.Out,                     o)
                    member v.WriteLine(              w:IO.TextWriter                ) = writeLineI(v, OutputStyle.Default,           w, OutputOptions.Default)
                    member v.WriteLine(              w:IO.TextWriter,o:OutputOptions) = writeLineI(v, OutputStyle.Default,           w,                     o)
                    member v.WriteLine(s:OutputStyle                                ) = writeLineI(v,                   s, Console.Out, OutputOptions.Default)
                    member v.WriteLine(s:OutputStyle,                o:OutputOptions) = writeLineI(v,                   s, Console.Out,                     o)
                    member v.WriteLine(s:OutputStyle,w:IO.TextWriter                ) = writeLineI(v,                   s,           w, OutputOptions.Default)
                    member v.WriteLine(s:OutputStyle,w:IO.TextWriter,o:OutputOptions) = writeLineI(v,                   s,           w,                     o)

    (* example of user-defined JSON parser/unparser compatible with module "EmitJSONio" *)
    type Value with static member ToJSON(v:Value)   = v
                    static member FromJSON(v:Value) = v

(* END *)
