namespace NXyn

  module JSON =


    type OutputStyle = Compact
                     | Format
                     | Print
                     | PrintJagged
    with static member Default : OutputStyle

    type [<Sealed>] InputOption =
         static member AllowEscapedSlashInput : System.Boolean -> InputOption
         static member AllowCommentInput      : System.Boolean -> InputOption

         static member TryParse               : System.String -> option<InputOption>

    type [<Sealed>] InputOptions =
         member AllowEscapedSlashInput        : System.Boolean
         member AllowCommentInput             : System.Boolean

         static member Default                : InputOptions

         member        Set                    : option:InputOption  -> InputOptions
         member        Set                    : options:System.Collections.Generic.IEnumerable<InputOption>  -> InputOptions
         member        Set                    : options:System.Collections.Generic.IEnumerable<System.String> -> InputOptions

    (***)

    type [<Sealed>] OutputOption =
         static member  AlignValues           : OutputOption
         static member  CompressLeft          : OutputOption
         static member  CompressRight         : OutputOption
         static member  Indentation           : n:System.Int32  -> OutputOption
         static member  KeyValueSeparator     : s:System.String -> OutputOption
         static member  SplitDepth            : n:System.Int32  -> OutputOption
         static member  NewLines              : n:System.Int32  -> OutputOption
         static member  CompressArray         : n:System.Int32  -> OutputOption
         static member  CompressObject        : n:System.Int32  -> OutputOption (* Assoc? *)

         static member  TryParse              : s:System.String -> option<OutputOption>

    type [<Sealed>] OutputOptions =
         static member Default                : OutputOptions
         static member Compact                : OutputOptions

         member        AlignValues            : bool
         member        CompressLeft           : bool
         member        CompressRight          : bool
         member        EscapeOutputSlash      : bool
         member        CompressArray          : int (* name? *)
         member        CompressObject         : int (* name ? *)
         member        PageWidth              : int
         member        NewLines               : int
         member        SplitDepth             : int
         member        Indentation            : int
         member        KeyValueSeparator      : string

         member        Set                    : option:OutputOption  -> OutputOptions
         member        Set                    : options:System.Collections.Generic.IEnumerable<OutputOption>  -> OutputOptions
         member        Set                    : options:System.Collections.Generic.IEnumerable<System.String> -> OutputOptions


    type [<Sealed>] Value =
         static member ToJSON : v:Value -> Value
         static member FromJSON : v:Value -> Value

         member Item : index:System.Int32  -> Value with get
         member Item : index:System.String -> Value with get
         member Item : index:System.Object -> Value with get

         member Length : int

         member GetNormalized : unit -> Value

         member GetPairs : unit -> seq<System.String * Value>

         member GetContents : unit -> seq<Value>

         /// <example>Value.Array([| Value.True; other; Value.Int32(kInt) |])</example>
         static member Array : xs:array<Value> -> Value

         member IsArray : System.Boolean

         (* keep original assoc order. if entries with identical keys occur, then later ones replace the ones inserted before. *)
         /// <example>Value.Assoc([| "a",Value.Number(3.1514); "a",Value.True |]) = Value.Assoc([| "a",Value.True |])</example>
         static member Assoc : xs:array<System.String * Value> -> Value

         member IsAssoc : System.Boolean

         /// <example>Value.BigInt(99999999999999999999999999999999999999999999I)</example>
         static member BigInt : x:System.Numerics.BigInteger -> Value

         member IsBigInt : System.Boolean

         /// <example>Value.BitArray("01011011101111101111111011111111111011111111111110")</example>
         static member BitArray : b:System.Collections.BitArray -> Value

         member IsBitArray : System.Boolean

         /// <example>Value.Bool(true) = Value.True && Value.Bool(false) = Value.False</example>
         static member Bool : b:System.Boolean -> Value

         member IsBool : System.Boolean

         /// <example>Value.Byte(255uy)</example>
         static member Byte : b:System.Byte -> Value

         member IsByte : System.Boolean

         /// <example>Value.Char('\0')</example>
         static member Char : c:System.Char -> Value

         member IsChar : System.Boolean

         /// <example>Value.Decimal(2233344445555566666677777777888888889999999999M ??)</example>
         static member Decimal : d:System.Decimal -> Value

         member IsDecimal : System.Boolean

         /// <example>Value.False <> Value.True && Value.Bool(true) <> Value.Bool(false)</example>
         static member False : Value

         member IsFalse : System.Boolean

         /// <example>Value.Int16(32767y) = Value.Int16(MaxInt16)</example>
         static member Int16 : i:System.Int16 -> Value

         member IsInt16 : System.Boolean

         /// <example>Value.Int32(-2G) = Value.Int32(MinInt32)</example>
         static member Int32 : i:System.Int32 -> Value

         member IsInt32 : System.Boolean

         /// <example>Value.Int64(MAX) = Value.Int64(MaxInt64)</example>
         static member Int64 : i:System.Int64 -> Value

         member IsInt64 : System.Boolean

         static member MappedArray : ('a -> Value) * 'a [] -> Value
         member IsMappedArray : System.Boolean
         
         /// <example>Value.False <> Value.Null</example>
         static member Null : Value

         member IsNull : System.Boolean

         /// <example>Value.Number(1.0)</example>
         static member Number : n:System.Double -> Value

         member IsNumber: System.Boolean

         /// <remark>Value.Number(nan) = Value.Number(nan)</remark>
         static member Double : f:System.Double -> Value

         member IsDouble : System.Boolean

         /// <example>Value.SByte(127y) = Value(SByte(Intmax.Sbyte))</example>
         static member SByte : s:System.SByte -> Value

         member IsSByte : System.Boolean

         /// <example>Value.Single(Value.Number(infinity)) = Value.Double(Value.Number(infinity))</example>
         static member Single : s:System.Single -> Value

         member IsSingle : System.Boolean

         /// <example>Value.String("foobarbaz")</example>
         static member String : s:System.String -> Value

         member IsString : System.Boolean

         /// <example>Value.String0(null)</example>
         static member String0 : s:System.String -> Value

         member IsString0 : System.Boolean

         /// <example>Value.True = Value.True && Value.True <> Value.False</example>
         static member True : Value

         member IsTrue : System.Boolean

         /// <example>Value.UInt16(65535uy)</example>
         static member UInt16 : u:System.UInt16 -> Value

         member IsUInt16 : System.Boolean

         /// <example>Value.UInt32(3000000000u)</example>
         static member UInt32 : u:System.UInt32 -> Value

         member IsUInt32 : System.Boolean

         /// <example>Value.UInt64(-Min)</example>
         static member UInt64 : u:System.UInt64 -> Value

         member IsUInt64 : System.Boolean

         /// <summary>Reads a string from the given text data source, parses the string as a JSON value, and returns the JSON value.</summary>
         /// <exception>Raises an exception if the string does not represent a valid JSON value. (TBD more specific)</exception>
         /// <param name="r">Data source.</param>
         static member Read    : unit                                          -> Value
         static member Read    : r:System.IO.TextReader                        -> Value
         static member Read    :                          options:InputOptions -> Value
         static member Read    : r:System.IO.TextReader * options:InputOptions -> Value

         (* convert JSON value from string *)
         static member Parse   : s:System.String                        -> Value
         static member Parse   : s:System.String * options:InputOptions -> Value

         (* convert JSON value to string *)
         override ToString : unit -> System.String

         (* output JSON value *)
         member Write                          : unit                                                                    -> unit
         member Write                          :                                                   options:OutputOptions -> unit
         member Write                          :                     writer:System.IO.TextWriter                         -> unit
         member Write                          :                     writer:System.IO.TextWriter * options:OutputOptions -> unit
         member Write                          : style:OutputStyle                                                       -> unit
         member Write                          : style:OutputStyle *                               options:OutputOptions -> unit
         member Write                          : style:OutputStyle * writer:System.IO.TextWriter                         -> unit
         member Write                          : style:OutputStyle * writer:System.IO.TextWriter * options:OutputOptions -> unit

         (* output JSON value plus newline *)
         member WriteLine                      : unit                                                                    -> unit
         member WriteLine                      :                                                   options:OutputOptions -> unit
         member WriteLine                      :                     writer:System.IO.TextWriter                         -> unit
         member WriteLine                      :                     writer:System.IO.TextWriter * options:OutputOptions -> unit
         member WriteLine                      : style:OutputStyle                                                       -> unit
         member WriteLine                      : style:OutputStyle *                               options:OutputOptions -> unit
         member WriteLine                      : style:OutputStyle * writer:System.IO.TextWriter                         -> unit
         member WriteLine                      : style:OutputStyle * writer:System.IO.TextWriter * options:OutputOptions -> unit

    (***)

    (* basic views *)
    val (|Array|Assoc|Bool|Null|Number|String|) : v:Value -> Choice<array<Value>, array<System.String * Value>, System.Boolean, unit, System.Double, System.String>

    (* derived views *)
    val (|String0|_|)  : v:Value -> option<System.String>

    val (|BitArray|_|) : v:Value -> option<System.Collections.BitArray>
    val (|BigInt|_|)   : v:Value -> option<System.Numerics.BigInteger>
    val (|Decimal|_|)  : v:Value -> option<System.Decimal>
    val (|Double|_|)   : v:Value -> option<System.Double>

    [<System.CLSCompliant(false)>]
    val (|SByte|_|)    : v:Value -> option<System.SByte>
    val (|Int16|_|)    : v:Value -> option<System.Int16>
    val (|Int32|_|)    : v:Value -> option<System.Int32>
    val (|Int64|_|)    : v:Value -> option<System.Int64>

    val (|Byte|_|)     : v:Value -> option<System.Byte>
    [<System.CLSCompliant(false)>]
    val (|UInt16|_|)   : v:Value -> option<System.UInt16>
    [<System.CLSCompliant(false)>]
    val (|UInt32|_|)   : v:Value -> option<System.UInt32>
    [<System.CLSCompliant(false)>]
    val (|UInt64|_|)   : v:Value -> option<System.UInt64>

    val (|Char|_|)     : v:Value -> option<System.Char>

    val (|Single|_|)   : v:Value -> option<System.Single>

    (***)

    val (|MappedArray|_|) : m:(Value -> 'a) -> v:Value -> option<'a []>
    val MappedArray_of_Value    : m:(Value -> 'a) -> v:Value -> 'a []

    val Array_of_Value                   : v:Value -> Value []
    val Assoc_of_Value                   : v:Value -> (System.String * Value) []
    val Bool_of_Value                    : v:Value -> System.Boolean
    val Null_of_Value<'a when 'a : null> : Value -> 'a when 'a : null
    val Number_of_Value                  : v:Value -> System.Double
    val String_of_Value                  : v:Value -> System.String

    (***)

    val BigInt_of_Value   : v:Value -> System.Numerics.BigInteger
    val BitArray_of_Value : v:Value -> System.Collections.BitArray
    val Byte_of_Value     : v:Value -> System.Byte
    val Char_of_Value     : v:Value -> System.Char
    val Decimal_of_Value  : v:Value -> System.Decimal
    val Double_of_Value   : v:Value -> System.Double
    val False_of_Value    : v:Value -> System.Boolean
    val Int16_of_Value    : v:Value -> System.Int16
    val Int32_of_Value    : v:Value -> System.Int32
    val Int64_of_Value    : v:Value -> System.Int64
    val SByte_of_Value    : v:Value -> System.SByte
    val String0_of_Value  : v:Value -> System.String
    val True_of_Value     : v:Value -> System.Boolean
    val UInt16_of_Value   : v:Value -> System.UInt16
    val UInt32_of_Value   : v:Value -> System.UInt32
    val UInt64_of_Value   : v:Value -> System.UInt64
    val Single_of_Value   : v:Value -> System.Single

(* END *)
