

module Basics


open System
open System.Collections.Generic

open Ast
open X64instr

open InstrIO
open MoreInstrIO

open AstIO
open MoreAstIO


open NXyn
open NXyn.Aux
open NXyn.Aux.Global

open NXyn.Aux.IntMisc
open NXyn.Aux.OtherMisc
open NXyn.Aux.SetMisc


let (|Contains|_|)(part:String) (whole:String) = if whole.Contains(part) then Some() else None

let RawOptions = ref [""]


let LookupOptionWithArg (prefix:String) (argMapper:String -> option<'a>) (defaultVal:'a) : 'a =
                        !RawOptions
                        |> List.tryFind (fun s -> s.StartsWith(prefix))
                        |> Option.bind (fun s -> s.Substring(prefix.Length).ToLower() |> argMapper)
                        |> function None -> defaultVal
                                  | Some(v) -> v


(***)


type Target = AMDk10
            | IntelCore2
            | IntelCoreI7
with static member Default = IntelCore2


let Target_of_String = function Contains("amd")   | Contains("k10")    | Contains("barcelona") | Contains("shanghai") -> Some(Target.AMDk10)
                              | Contains("intel") | Contains("core2")  | Contains("conroe")    | Contains("merom")    -> Some(Target.IntelCore2)
                              | Contains("intel") | Contains("corei7") | Contains("nehalem")   | Contains("westmere") -> Some(Target.IntelCoreI7)
                              | _ -> None

(* isel-target= *)
let target() = LookupOptionWithArg "--nxyn-isel-target=" Target_of_String Target.Default

(* sched-target *)
let scheduling_target() = LookupOptionWithArg "--nxyn-sched-target=" Target_of_String Target.Default

let TargetIsAMD()   = match target() with AMDk10 -> true  | IntelCoreI7 | IntelCore2 -> false
let TargetIsIntel() = match target() with AMDk10 -> false | IntelCoreI7 | IntelCore2 -> true





type RedoEA = No
            | Yes
                
let RedoEA_of_String = function "no"  | "false" | "none" -> Some(RedoEA.No)
                              | "yes" | "true"  | "some" -> Some(RedoEA.Yes)
                              | _ -> None  

let redo_EA() = LookupOptionWithArg "--nxyn-redo-ea=" RedoEA_of_String RedoEA.Yes



type ElimVecCopies = No
                   | Yes

let ElimVecCopies_of_String = function "no"  | "false" | "none" -> Some(ElimVecCopies.No)
                                     | "yes" | "true"  | "some" -> Some(ElimVecCopies.Yes)
                                     | _ -> None  


let elim_veccopies() = LookupOptionWithArg "--nxyn-elim-veccopies=" ElimVecCopies_of_String ElimVecCopies.Yes





type ReassignXMM = No (* none at all *)
                 | HC (* hill climbing (good) *)
                 | GA (* genetic algorithm (best, but slow) *)


let ReassignXMM_of_String = function "no"  | "false" | "none"                                                   -> Some(ReassignXMM.No)
                                   | "yes" | "true"  | "some" | "hc" | "hill_climbing"     | "hillclimbing"     -> Some(ReassignXMM.HC)
                                   | "yes" | "true"  | "full" | "ga" | "genetic_algorithm" | "geneticalgorithm" -> Some(ReassignXMM.GA)
                                   | _ -> None



let reassign_xmm() = LookupOptionWithArg "--nxyn-reassign-xmm=" ReassignXMM_of_String (if TargetIsIntel() then ReassignXMM.GA else ReassignXMM.No)




type ReassignStack = No      (* no reassignment *)
                   | Yes     (* reassign, use `%rsp' register *)
                   | Yes_Xsp (* reassign, try using `%rsp' *and* second register with good encoding *)


let ReassignStack_of_String = function "no" | "false" | "none" -> Some(ReassignStack.No)
                                     | "yes" | "true" | "some" -> Some(ReassignStack.Yes)
                                     | "yes" | "true" | "full" -> Some(ReassignStack.Yes_Xsp)
                                     | _ -> None

let reassign_stack() = LookupOptionWithArg "--nxyn-reassign-stack=" ReassignStack_of_String (if TargetIsIntel() then ReassignStack.Yes else ReassignStack.No)



let DumpOptions() = eprintfn "target         = %A" (target())
                    eprintfn "reassign_stack = %A" (reassign_stack())
                    eprintfn "reassign_xmm   = %A" (reassign_xmm())
                    eprintfn "sched_target   = %A" (scheduling_target())

                    
    

(* => Options *)



let konstSizeDict = new Dictionary<String,Int32>()


let dodump = ref false

let find_near_power_of_two (m:Int32) = 
                           match Array.tryFind is_power_of_two [| m; m-1; m+1; m-2; m+2; m-3; m+3 |]
                            with | Some(x) -> x
                                 | None -> m

(***)

(* module opt_basics *)

type IBlock = { Name            : String;
                Alignment       : Int32;
                LexPred         : String;
                LexSucc         : String;
                DynPreds        : String [];
                DynSuccs        : String [];
                mutable Instrs  : Instr [];
                ReturnsVoid     : Boolean;
              }

let nonoverlapping_stackLoc (size1,ofs1) (size2,ofs2) = ofs2 >= ofs1 + size1 || ofs1 >= ofs2 + size2




// auto-generate that
let try__M128sXds_of_XsXds = function I_XsXds.Mulps -> Some(I_M128sXds.Mulps)
                                    | I_XsXds.Mulpd -> Some(I_M128sXds.Mulpd)
                                    | I_XsXds.Xorpd -> Some(I_M128sXds.Xorpd)
                                    | I_XsXds.Xorps -> Some(I_M128sXds.Xorps)
                                    | I_XsXds.Addpd -> Some(I_M128sXds.Addpd)
                                    | I_XsXds.Addps -> Some(I_M128sXds.Addps)
                                    | I_XsXds.Subpd -> Some(I_M128sXds.Subpd)
                                    | I_XsXds.Subps -> Some(I_M128sXds.Subps)
                                    | _ -> None

let try__XsXds_of_M128sXds = function I_M128sXds.Mulps -> Some(I_XsXds.Mulps)
                                    | I_M128sXds.Mulpd -> Some(I_XsXds.Mulpd)
                                    | I_M128sXds.Xorpd -> Some(I_XsXds.Xorpd)
                                    | I_M128sXds.Xorps -> Some(I_XsXds.Xorps)
                                    | I_M128sXds.Addpd -> Some(I_XsXds.Addpd)
                                    | I_M128sXds.Addps -> Some(I_XsXds.Addps)
                                    | I_M128sXds.Subpd -> Some(I_XsXds.Subpd)
                                    | I_M128sXds.Subps -> Some(I_XsXds.Subps)
                                    | _ -> None


let try__M128sXd_of_XsXd = function I_XsXd.Movaps -> Some(I_M128sXd.Movaps)
                                  | I_XsXd.Movapd -> Some(I_M128sXd.Movapd)
                                  | I_XsXd.Pshufd(l) -> Some(I_M128sXd.Pshufd(l))
                                  | _ -> None


(* => generate from attribute: move to Generate_InstrIO *)
let XsXds_is_commutative = function
    (* add FP          *) | I_XsXds.Addpd  | I_XsXds.Addps -> true
    (* multiply FP     *) | I_XsXds.Mulpd  | I_XsXds.Mulps -> true 
    (* minimum FP      *) | I_XsXds.Minpd  | I_XsXds.Minps -> true
    (* maximum FP      *) | I_XsXds.Maxpd  | I_XsXds.Maxps -> true
    (* average int     *) | I_XsXds.Pavgb  | I_XsXds.Pavgw -> true
    (* bitwise AND     *) | I_XsXds.Andpd  | I_XsXds.Andps  | I_XsXds.Pand   -> true
    (* bitwise OR      *) | I_XsXds.Orpd   | I_XsXds.Orps   | I_XsXds.Por    -> true
    (* bitwise XOR     *) | I_XsXds.Xorpd  | I_XsXds.Xorps  | I_XsXds.Pxor   -> true
    (* bitwise AND-NOT *) | I_XsXds.Andnpd | I_XsXds.Andnps | I_XsXds.Pandn  -> true
    (* add int         *) | I_XsXds.Paddb  | I_XsXds.Paddd  | I_XsXds.Paddq   | I_XsXds.Paddw   -> true
    (* add int w/sat.  *) | I_XsXds.Paddsb | I_XsXds.Paddsw | I_XsXds.Paddusb | I_XsXds.Paddusw -> true
    (* minimum int     *) | I_XsXds.Pminsb | I_XsXds.Pminsd | I_XsXds.Pminsw  | I_XsXds.Pminub   | I_XsXds.Pminud | I_XsXds.Pminuw -> true
    (* maximum int     *) | I_XsXds.Pmaxsb | I_XsXds.Pmaxsd | I_XsXds.Pmaxsw  | I_XsXds.Pmaxub   | I_XsXds.Pmaxud | I_XsXds.Pmaxuw -> true
                          | _ -> false



let get_all_stacklocs (tDict:Dictionary<String,IBlock>) (cc:String []) =
    let s  = new HashSet<Int32 * Int32>()
    let ds = new HashSet<Int32 * Int32>()
    let d  = new HashSet<Int32 * Int32>()

    for lbl in cc do
        let b = tDict.[lbl]
        b.Instrs |> Array.iter (fun(i) -> Ms_of_Instr(i)  |> List.choose stackLocOption |> List.iter (s.Add  >> ignore)
                                          Mds_of_Instr(i) |> List.choose stackLocOption |> List.iter (ds.Add >> ignore)
                                          Md_of_Instr(i)  |> List.choose stackLocOption |> List.iter (d.Add  >> ignore))
    let map1(x) = x |> ArrayMisc.array_of_collection // |> Array.map (fun(KeyValue(k,v)) -> (k,v))
    (map1(s), map1(ds), map1(d))



let rec flags_overwritten_after (flags:Set<F>) (instrs:Instr []) (ix:Int32) =
                if   Set.isEmpty(flags)  then true
                elif ix >= instrs.Length then false
                else let instr = instrs.[ix]
                     let inflags  = Fs_of_Instr(instr)  |> Set.ofList
                     let outflags = Fd_of_Instr(instr) |> Set.ofList
                     if sets_intersect flags inflags then false
                                                     else flags_overwritten_after (Set.difference flags outflags) instrs (ix+1)
            
let flags_written_by_addq = Fd_of_Instr(M64sR64ds(I_M64sR64ds.Addq, {S=M64.S(EA64(Iexpr.Int 0I,Some{Base=R64.rsp},None)); DS=R64.rbx})) |> Set.ofList
let outFlags__addq = Fd_of_Instr(M64sR64ds(I_M64sR64ds.Addq, {S=M64.S(EA64(Iexpr.Int 0I,Some{Base=R64.rsp},None)); DS=R64.rbx})) |> Set.ofList





let attributes    = new Dictionary<String, list<String * list<Item>>>()

let isGlobal(x)   = let (found,lines) = attributes.TryGetValue(x) in found && (lines |> List.exists (fst >> ((=)"global")))
let isFunction(x) = let (found,lines) = attributes.TryGetValue(x) in found && (lines |> List.exists ((=) ("type",[Type "function"])))

let ninstrs_in_cc (tDict:Dictionary<String,IBlock>) (comp:String []) =
                  let n = ref 0
                  comp |> Array.iter (fun l -> let b = tDict.[l] in n := b.Instrs.Length + !n)
                  !n

let try_get_single_entrypoint (cc:String []) =
                              match Array.filter isFunction cc with
                                  | [| x |] -> Some(x)
                                  | _ -> None

let get_single_entrypoint (cc:String []) =
                          try_get_single_entrypoint(cc)
                          |> function None -> failwith "get_single_entrypoint"
                                    | Some(e) -> e

let get_exitpoints (tDict:Dictionary<String,IBlock>) (cc:String []) =
                   cc |> Array.filter (fun lbl -> let b = tDict.[lbl]
                                                  b.Instrs <> [||] &&
                                                  Instr_is_FunReturn(b.Instrs.[b.Instrs.Length - 1]))


let get_nonempty_exitpoints (tDict:Dictionary<String,IBlock>) (cc:String []) =
                            get_exitpoints tDict cc
                            |> function [||] -> failwith "get_nonempty_exitpoints"
                                      | xs -> xs
(***)

let getUsedLabels sections = let used = new HashSet<string>()
                             let rec loopArgs = function R8reg _ | R16reg _ | R32reg _ | R64reg _ -> ()
                                                       | NR64reg _ | XMMreg _ | YMMreg _ | Float _ | Unit -> ()
                                                       | Mem m  -> List.iter (used.Add >> ignore) m.Ids
                                                       | NMem m -> List.iter (used.Add >> ignore) m.S.Ids
                                                       | Type s | String s -> used.Add(s) |> ignore
                                                       | Literal e | Iexpr e -> List.iter (used.Add >> ignore) e.Ids

                             for KeyValue(_,pairs) in attributes do pairs |> List.iter (snd >> List.iter loopArgs)
                             let doLine = function ALabel(s,_) -> () // kept labels are either global or used elsewhere
                                                 | Instr(i) -> Id_of_Instr(i) |> List.iter (fun(i) -> used.Add(i.Name) |> ignore)
                                                 | Directive(_,args) -> args |> List.iter loopArgs
                             sections |> Array.iter (fun(_,_,lines) -> lines |> Array.iter doLine)
                             used
(***)

let labelGen = new LabelGenerator.t(".L",3)
let nErrors  = ref 0

(***)

let externABI = new HashSet<string>()
let voidDict  = new HashSet<string>()

let build_voidDict(nxynSections) = let doLine(s) = let p = (s:string).IndexOf("(")
                                                   if p > 0 then match s.Substring(0,p).Split([| ' '; '\t' |], StringSplitOptions.RemoveEmptyEntries) |> Array.toList 
                                                                  with | ["void"; name] | ["static"; "void"; name] -> voidDict.Add(name) |> ignore
                                                                       | _ -> ()
                                   nxynSections |> Array.iter (fun(_,_,lines) -> lines |> Array.iter (function Directive("ascii",[String(s)]) -> doLine(s)
                                                                                                             | Directive _ | Instr _ | ALabel _ -> ()))

let build_konstSizeDict(secs) = let doALabel(name) = match attributes.TryGetValue(name)
                                                      with | (false,_) -> ()
                                                           | (true,lines) -> match List.tryPick (function("size",[Item.Iexpr(IexprInt32(i))]) -> Some(i) | _ -> None) lines
                                                                              with | None -> ()
                                                                                   | Some(size) -> konstSizeDict.[name] <- size
                                let doLine = function ALabel(name,_) -> doALabel(name)
                                                    | Directive _ | Instr _ -> ()
                                secs |> Array.iter (fun(name:String,_,lines) -> if name.StartsWith(".rodata") then lines |> Array.iter doLine)

(***)

(* topological sort of nodes before liveness analysis reduces # of iterations until fixpoint is reached; algorithm 17.5 *)
let tsort_cc (tDict:Dictionary<String,IBlock>) (cc:String []) =
    let visited = new HashSet<String>()
    let sorted  = new ResizeArray<String>()
    let rec dfs(lbl) = if not(visited.Contains(lbl)) then visited.Add(lbl) |> ignore // consistent?
                                                          if tDict.ContainsKey(lbl) then
                                                              tDict.[lbl].DynSuccs |> Array.iter (fun(succ) -> if tDict.ContainsKey(succ) then dfs(succ))
                                                          sorted.Add(lbl)
    cc |> Array.iter dfs
    sorted.ToArray()

let get_ccs (tDict:Dictionary<String,IBlock>) =
                                let components = ResizeArray<String []>()
                                let visited = new HashSet<string>()
                                for l in tDict.Keys do
                                    if not(visited.Contains(l)) then
                                        (* start new component *)
                                        let comp = new ResizeArray<String>()
                                        let rec dfs(lbl:string) =
                                            if externABI.Contains(lbl) then ()
                                            elif not(visited.Contains(lbl)) then
                                                visited.Add(lbl) |> ignore
                                                comp.Add(lbl)
                                                if tDict.ContainsKey(lbl) then
                                                    let block = tDict.[lbl]
                                                    for lPred in block.DynPreds do dfs(lPred)
                                                    for lSucc in block.DynSuccs do dfs(lSucc)
                                        dfs(l)

                                        let arr = comp.ToArray()
                                        (* one "function" per component minimum *)
                                        if arr |> Array.exists isFunction then components.Add(arr)
                                components.ToArray()

(***)


