module FFTWschedule

open X64instr

open InstrIO
open MoreInstrIO



(* basic operations over instructions *)
type operand = OpV of int
             | OpR of int
             | OpM of M
             | OpF of F

(* FFTW2 style *)
let similar_operand a b =
    match (a,b) with
        | (OpM(_), OpM(_)) -> true
        | _ -> false
        (* 
        MemOp(AE(id1,idx1,_)), MemOp(AE(id2,idx2,_))) ->
            | (MemOp(AE(id1,idx1,_)), MemOp(AE(id2,idx2,_))) ->
                id1=id2 && idx1=idx2
                *)

(* FFTW2 style *)
let operand_clobbers a b =
    match (a,b) with
        | (OpM(m1), OpM(m2)) -> m1=m2
        | _ -> false
  //(MemOp(AE(id1,idx1,ofs1)), MemOp(AE(id2,idx2,ofs2))) -> 
//      id1=id2 && idx1=idx2 && ofs1=ofs2


let inputoperands_of_Instr(i) =
    let rs0 = (AllRs_of_Instr(i) @ AllRds_of_Instr(i))  |> List.map (fun(r:R) -> OpR(r.Int))
    let rs1 = (AllXs_of_Instr(i) @ AllXds_of_Instr(i))  |> List.map (fun(r:X) -> OpV(r.Int))
    let rs2 = (AllYs_of_Instr(i) @ AllYds_of_Instr(i))  |> List.map (fun(r:Y) -> OpV(r.Int))
    let rs3 = (Ms_of_Instr(i)    @ Mds_of_Instr(i))     |> List.map (fun(_,m) -> OpM(m))
    let rs4 = Fs_of_Instr(i) |> List.map OpF
    rs0@rs1@rs2@rs3@rs4


let outputoperands_of_Instr(i) =
    let ws0 = (AllRd_of_Instr(i) @ AllRds_of_Instr(i)) |> List.map (fun(r:R) -> OpR(r.Int))
    let ws1 = (AllXd_of_Instr(i) @ AllXds_of_Instr(i)) |> List.map (fun(r:X) -> OpV(r.Int))
    let ws2 = (AllYd_of_Instr(i) @ AllYds_of_Instr(i)) |> List.map (fun(r:Y) -> OpV(r.Int))
    let ws3 = (Md_of_Instr(i) @ Mds_of_Instr(i))       |> List.map (fun(_,m) -> OpM(m))
    let ws4 = Fd_of_Instr(i)                     |> List.map OpF
    ws0@ws1@ws2@ws3@ws4



  (* Here, we have functions to transform a sequence of assignments
     (variable = expression) into a DAG (a directed, acyclic graph).
     The nodes of the DAG are the assignments, and the edges indicate
     dependencies.  (The DAG is analyzed in the scheduler to find an
     efficient ordering of the assignments.)

     This module contains utilities to manipulate the DAG in various ways. *)

(* Dag structure *)
type color = RED | BLUE | BLACK | YELLOW

type dagnode = { instr         : Instr;
                 inputs        : operand list;
                 outputs       : operand list;
                 mutable succs : dagnode list;
                 mutable preds : dagnode list;
                 mutable label : int;
                 mutable color : color }

//type dag = dagnode list // Dag of dagnode list

type schedule = SDone
              | SInstr of X64instr.Instr
              | SSeq of schedule * schedule
              | SPar of schedule list



(* true if node uses one of vs *)
let node_uses vs node = List.exists (fun v -> List.exists ((=)v) node.inputs) vs

(* true if assignment of any of `vs' clobbers any input of node `n' *)
let node_clobbers n vs = List.exists (fun v -> List.exists (operand_clobbers v) n.inputs) vs

(* true if nodeb depends on nodea *)
let depends_on node_a node_b = node_uses node_a.outputs node_b || node_clobbers node_a node_b.outputs
      
let dagnode_of_Instr(i) = { instr = i; inputs = inputoperands_of_Instr(i); outputs = outputoperands_of_Instr(i); succs = []; preds = []; label = 0; color = BLACK }

(* transform a list of annotated instructions into a dag *)
let makedag(instrs) = let note_dep i j = if depends_on i j then (i.succs <- j::i.succs; j.preds <- i::j.preds)
                      let dag = List.map dagnode_of_Instr instrs in
                      List.iter (fun i -> List.iter (fun j -> note_dep i j) dag) dag
                      dag

//let map f (Dag dag) = Dag (List.map f dag)
//let for_all (Dag dag) f = List.iter f dag
//let to_list (Dag dag) = dag
//let find_node f (Dag dag) = Basics.find f dag

(* breadth-first search *)
let rec bfs dag node init_label = node.label <- init_label
                                  let rec loop = function [] -> ()
                                                        | node::rest -> let neighbors = node.preds @ node.succs
                                                                        let m = List.min (List.map (fun node -> node.label) neighbors)
                                                                        if node.label <= m + 1 then loop rest
                                                                                               else node.label <- m + 1
                                                                                                    loop (rest @ neighbors)
                                  loop (node.preds @ node.succs)

    (* This module contains the instruction scheduler, which finds an
     efficient ordering for a given list of instructions.

     The scheduler analyzes the DAG (Directed, Acyclic Graph) formed by
     the instruction dependencies, and recursively partitions it.  The
     resulting schedule data structure expresses a "good" ordering
     and structure for the computation.

     The scheduler makes use of utilties in Dag and other packages to
     manipulate the Dag and the instruction list. *)


(* Dag scheduler *)
let to_instr(node) = node.instr

(*
XXX  let return x = x
*)
let has_color c n = n.color = c
let set_color c n = n.color <- c
let has_either_color c1 c2 n = n.color = c1 || n.color = c2

let infinity = 100000

(* XXX more compact *)
let cc dag inputs = dag |> List.iter (fun node -> node.label <- infinity)
                    match inputs with [] -> failwith "connected"
                                    | a::_ -> bfs dag a 0 
                    let (lt,gte) = dag |> List.partition (fun n -> n.label < infinity)
                    (List.map to_instr lt, List.map to_instr gte)

let rec connected_components alist = let dag = makedag alist
                                     let inputs = dag |> List.filter (fun node -> node.preds = [])
                                     match cc dag inputs with (a, []) -> [a]
                                                            | (a, b) -> a :: connected_components b

(* XXX ??? *)
let loads_twiddle node = false 
(*
    match (node.input_variables, node.predecessors) with
      | ([x], []) -> Variable.is_twiddle x
      | _ -> false
*)

let loopi_cand n = has_color        BLACK        n && List.forall (has_either_color RED  YELLOW) n.preds
let loopo_cand n = has_either_color BLACK YELLOW n && List.forall (has_color        BLUE)        n.succs

let partition magic_loopo alist =
    let dag = makedag alist
    let dag' = dag // Dag.to_list dag
    let inputs  = List.filter (fun node -> node.preds = []) dag'
    let outputs = List.filter (fun node -> node.succs = []) dag'
    let special_inputs = List.filter loads_twiddle dag' 
    dag |> List.iter (set_color BLACK)

    List.iter (set_color RED) inputs

    (* The special inputs are input that read a twiddle factor.  They
    can end up either in the blue or in the red part.  If a red
    node needs a special input, the special input becomes red.  If
    all successors of a special input are blue, it becomes blue.
    Outputs are always blue.
    
    As a consequence, however, the final partition might be
    composed only of blue nodes (which is incorrect).  In this case
    we manually reset all inputs (whether special or not) to be red. *)
    
    List.iter (set_color YELLOW) special_inputs
    
    List.iter (set_color BLUE) outputs
    
    let rec loopi donep = 
        match List.filter loopi_cand dag' with
            | [] -> if not donep then loopo true
            | i ->
                List.iter (fun node ->
                              set_color RED node
                              List.iter (set_color RED) node.preds) i
                loopo false
    and loopo donep =
        match List.filter loopo_cand dag' with
            | [] -> if not donep then loopi true
            | o -> List.iter (set_color BLUE) o
                   loopi false

    (* among the magic parameters, this is the most obscure *)
    if magic_loopo then loopo false else loopi false
  
    (* fix the partition if it is incorrect *)
    if not (List.exists (has_color RED) dag') then List.iter (set_color RED) inputs
    
    ((List.map to_instr (List.filter (has_color RED) dag')), (List.map to_instr (List.filter (has_color BLUE) dag')))



//type aInstrs = X64instr.Instr list

let schedule magic_loopo =
    let rec schedule_alist = function
      | [] -> SDone
      | [a] -> SInstr(a)
      | alist ->
           match connected_components alist with
               | [a] -> schedule_connected a
               | l -> SPar (List.map schedule_alist l)
    and schedule_connected alist = 
      let (a,b) = partition magic_loopo alist 
      SSeq(schedule_alist a, schedule_alist b)
    fun alist -> schedule_alist alist







(* find all variables used inside a scheduling unit *)
let rec find_block_vars = function SDone     -> []
                                 | SInstr(i) -> inputoperands_of_Instr(i) @ outputoperands_of_Instr(i)
                                 | SPar(a)   -> List.concat (List.map find_block_vars a)
                                 | SSeq(a,b) -> (find_block_vars a) @ (find_block_vars b)

let overlap = let fold b n y = if List.exists (similar_operand y) b then n+1 else n
              fun a b -> List.fold (fold b) 0 a

// List.minBy
//let rec minimize f = function
//    | [] -> failwith "minimize"
//    | [n] -> n
//    | n :: rest ->
//        let x = minimize f rest 
//        if f x >= f n then n else x

(* reorder a list of schedules so as to maximize overlap of variables *)
let reorder =
    let rec loop = function
        | [] -> []
        | (a,va) :: b ->
            let c  = List.map (fun (a, x) -> ((a, x), overlap va x)) b
            let c' = List.sortWith (fun (_, a) (_, b) -> compare b a (* a > b *)) c
            let b' = List.map (fun (a, _) -> a) c' 
            a :: (loop b') 
    fun l -> 
      (* start with smallest block --- does this matter ? *)
      match List.map (fun x -> x, Set.toList(Set.ofList(find_block_vars(x)))) l with
          | [] -> []
          | l' ->  
              let m = List.minBy (fun(_,x) -> List.length x) l'
              let l'' = List.filter (fun i -> i <> m) l' 
              loop (m :: l'')

(* remove Par blocks *)
let rec linearize = function
    | SSeq(a,SDone) -> linearize a
    | SSeq(SDone,a) -> linearize a
    | SSeq(a,b) -> SSeq(linearize a,linearize b)
    | SPar [a] -> linearize a
    | SPar l -> (* try to balance nested Par blocks *)
        let n2 = (List.length l) / 2
        let rec loop n a b =
            if n = 0 then (List.rev b, a)
            else
                match a with
                    | [] -> failwith "loop"
                    | x :: y -> loop (n - 1) y (x :: b)
        let (a, b) = loop n2 (reorder l) []
        linearize (SSeq(SPar(a),SPar(b)))
    | x -> x

let rec instrs_of_seqschedule = function SDone     -> []
                                       | SInstr(i) -> [i]
                                       | SSeq(a,b) -> (instrs_of_seqschedule a) @ (instrs_of_seqschedule b)
                                       | SPar _    -> failwith "instrs_of_seqschedule"



(**************************************************************************************************************************************)

(* HC... a selection of genes should be considered for hillclimbing *)

type ATdecomp = Nary of (* gene positions *) list<int> * (* left *) ATdecomp * (* right *) ATdecomp
              | BINary of int (* pos of single gene *)
              | NOdecision
with static member Make(p0,howmany) = if   howmany < 1 then failwith "ATdecomp.Make"
                                      elif howmany = 1 then (p0,NOdecision)
                                      elif howmany = 2 then (p0+1,BINary(p0))
                                                       else let howmanyR = howmany / 2
                                                            let howmanyL = howmany - howmanyR
                                                            let p1 = p0 + howmany - 1
                                                            let positions = Array.init (howmany-1) (fun i -> i+p0) |> Array.toList
                                                            let (p2,left) = ATdecomp.Make(p1,howmanyL)
                                                            let (p,right) = ATdecomp.Make(p2,howmanyR)
                                                            (p,Nary(positions,left,right))

     static member Partition(genes,ls,rs,n,nL,nR,xs,gs) = match (gs,xs)
                                                           with | ([],[x]) -> if nL = 1 then (x::ls, rs) else (ls, x::rs)
                                                                | (g::gs0,x::xs0) -> if   n <= nL   then (genes:System.Collections.BitArray).[g] <- true
                                                                                                         ATdecomp.Partition(genes, x::ls, rs, n-1, nL-1, nR, xs0, gs0)
                                                                                     elif n <= nR   then genes.[g] <- false
                                                                                                         ATdecomp.Partition(genes, ls, x::rs, n-1, nL, nR-1, xs0, gs0)
                                                                                     elif genes.[g] then ATdecomp.Partition(genes, x::ls, rs, n-1, nL-1, nR, xs0, gs0)
                                                                                                    else ATdecomp.Partition(genes, ls, x::rs, n-1, nL, nR-1, xs0, gs0)
                                                                | _ -> failwith "ATdecomp.Partition"

     member x.GetHCpoints() = match x with Nary(ps,l,r) -> ps @ (l.GetHCpoints()) @ (r.GetHCpoints())
                                         | BINary(p)    -> [p]
                                         | NOdecision   -> []

     

(* schedule that includes annotations *)
type ATschedule = ATDone
                | ATInstr of Instr
                | ATSeq of ATschedule * ATschedule
                | ATPar of ATschedule list * ATdecomp
with static member FromSchedule(p0,s) = match s
                                         with | SDone -> (p0,ATDone)
                                              | SInstr(i) -> (p0,ATInstr(i))
                                              | SSeq(a,b) -> let (p1,a') = ATschedule.FromSchedule(p0,a)
                                                             let (p2,b') = ATschedule.FromSchedule(p1,b)
                                                             (p2,ATSeq(a',b'))
                                              | SPar(xs) -> let (p1,decomp) = ATdecomp.Make(p0, List.length(xs))
                                                            let (p2,xs) = List.fold (fun (p0,xs) x -> let (p,x') = ATschedule.FromSchedule(p0,x) in (p,x'::xs)) (p1,[]) xs
                                                            (p2,ATPar(List.rev xs,decomp))

     member x.GetHCpoints(l) = match x
                                with | ATDone | ATInstr _ -> []
                                     | ATSeq(a,b) -> (a.GetHCpoints(l)) @ (b.GetHCpoints(l))
                                     | ATPar(xs,decomp) -> List.concat ((decomp.GetHCpoints()) :: (if l > 0 then List.map (fun(x:ATschedule) -> x.GetHCpoints(l-1)) xs else []))

     member x.ToArray() = let arr = new ResizeArray<Instr>()
                          let rec loop = function ATDone -> ()
                                                | ATInstr(i) -> arr.Add(i)
                                                | ATSeq(l,r) -> loop(l)
                                                                loop(r)
                                                | ATPar _ -> failwith "ATschedule.ToArray"
                          loop(x)
                          arr.ToArray()

     static member Permute(genes,xs,decomp) = match decomp
                                               with | Nary(sel,l,r) -> let howmany  = 1 + List.length sel
                                                                       let howmanyL = (howmany + 1) / 2
                                                                       let howmanyR = howmany / 2
                                                                       let (l_xs,r_xs) = ATdecomp.Partition(genes,[],[],howmany,howmanyL,howmanyR,xs,sel)
                                                                       ATSeq(ATschedule.Permute(genes,l_xs,l),ATschedule.Permute(genes,r_xs,r))
                                                    | BINary(g) -> match xs with [l;r] -> if genes.[g] then ATSeq(l,r) else ATSeq(r,l)
                                                                               | _ -> failwith "ATschedule.Permute: BINary"
                                                    | NOdecision -> match xs with [x] -> x
                                                                                | _ -> failwith "ATschedule.Permute: NOdecision"

     member x.Linearize(genes) = match x with ATDone | ATInstr _ -> x
                                            | ATSeq(a,b) -> ATSeq(a.Linearize(genes), b.Linearize(genes))
                                            | ATPar(xs,decomp) -> ATschedule.Permute(genes, List.map (fun(x:ATschedule) -> x.Linearize(genes)) xs, decomp)

