module InstrSize

open System
open System.Collections.Generic


open X64instr
open Ast

open MoreInstrIO

open AstIO
open MoreAstIO


(* instruction sizes are obtained by invoking GNU assembler `as' and then cached in a dictionary *)
let sizes = new Dictionary<Instr,Int32>()

(* globally accessible read-only process start info for `as' *)
let psi = new Diagnostics.ProcessStartInfo("as",                      (* name of GNU assembler executable *)
                                           "--64 -a -o /dev/null")    (* parameters: 64-bit mode `--64', output listing `-a', discard object file `-o /dev/null' *)
psi.CreateNoWindow         <- true
psi.RedirectStandardInput  <- true
psi.RedirectStandardOutput <- true
psi.RedirectStandardError  <- true
psi.UseShellExecute        <- false

(* literal constants passed to `as' *)
let asmNOPline = ".byte 0x00090"    (* used as a separatator between two instructions passed to `as' *)
let asmPrelude = ".psize 0"         (* advise `as' not to generate any formfeeds *)



let invoke_AS (instrs0:Instr []) =
              (* process textual GNU assembler listing line-per-line *)
              let replies   = new ResizeArray<Int32>()
              let prevOfs = ref 0
              let processReply src (outLine:Diagnostics.DataReceivedEventArgs) =
                               let line = outLine.Data
                               if not(String.IsNullOrEmpty(line)) && line.Contains(asmNOPline) then
                                  let ofs = Int32.Parse(line.Split([| ' '; '\t' |], StringSplitOptions.RemoveEmptyEntries).[1], Globalization.NumberStyles.HexNumber)
                                  replies.Add(ofs - !prevOfs)
                                  prevOfs := ofs + 1

              (* start up auxiliary GNU assembler process and connect handler for asynchronous data input *)
              use ps = new Diagnostics.Process()
              ps.OutputDataReceived.AddHandler(new Diagnostics.DataReceivedEventHandler(processReply))
              ps.StartInfo <- psi
              ps.Start() |> ignore
              ps.BeginOutputReadLine()

              (* send prelude and instructions to process *)
              fprintfn ps.StandardInput "%s" asmPrelude

              
              instrs0
              |> Array.choose (fun(i) -> if sizes.ContainsKey(i) then None
                                                                 else sizes.[i] <- 0
                                                                      fprintfn ps.StandardInput "%t\n%s" (WriteInstr_to i) asmNOPline
                                                                      Some(i))
              |> fun instrs -> (* close channel and wait for process completion *)
                               ps.StandardInput.Close()
                               ps.WaitForExit()

                               (* assert that # of answers matches # of requests *)
                               if replies.Count <> instrs.Length then failwithf "InstrSize.invoke_AS: expected %d lines, got %d" instrs.Length replies.Count

                               (* insert replies into `size' table *)
                               for ix = 0 to instrs.Length - 1 do sizes.[instrs.[ix]] <- replies.[ix]

let prepareForInstrs (instrs0:Instr []) =
                     (* filter out instructions not already present in the dictionary *)
                     let instrs = instrs0 |> Array.filter (sizes.ContainsKey >> not)
                     if instrs <> [||] then invoke_AS(instrs)

(***)

let size_of_Instr (instr:Instr) =
                  prepareForInstrs([|instr|])
                  sizes.[instr]

let sizes_of_Instrs (instrs:Instr []) =
                    prepareForInstrs(instrs)
                    instrs |> Array.map (fun(i:Instr) -> sizes.[i])

let size_of_Instr08 (instr:Instr) = instr |> Map_XYto08_in_Instr |> size_of_Instr

let sizes_of_Instrs08 (instrs:Instr []) = instrs |> Array.map Map_XYto08_in_Instr |> sizes_of_Instrs
