
module OptPeephole

open System
open System.Collections.Generic

open Ast
open X64instr

open InstrIO
open MoreInstrIO

open AstIO
open MoreAstIO


open NXyn
open NXyn.Aux
open NXyn.Aux.Global

open NXyn.Aux.IntMisc
open NXyn.Aux.OtherMisc
open NXyn.Aux.SetMisc

open Basics

open LocalScheduling





// for most current targets (except for core i7)!!
// for some targets, not for others!
(* dispatch for target ! *)



let (|I_XsXd_SSEcopy|_|) (i:I_XsXd) =
                         match i
                          with | I_XsXd.Movapd | I_XsXd.Movaps | I_XsXd.Movdqa | I_XsXd.Movdqu | I_XsXd.Movupd | I_XsXd.Movups -> Some()
                               | _ -> None
    

let (|I_M128sXd_SSEloadAligned|_|) (x:I_M128sXd) =
                                   match x
                                     with | I_M128sXd.Movapd | I_M128sXd.Movaps | I_M128sXd.Movdqa -> Some()
                                          | _ -> None
let (|I_M128sXd_SSEloadUnaligned|_|) (x:I_M128sXd) =
                                     match x
                                      with | I_M128sXd.Movdqu | I_M128sXd.Movupd | I_M128sXd.Movups -> Some()
                                           | _ -> None
let (|I_XsM128d_SSEstoreAligned|_|) (x:I_XsM128d) =
                                    match x
                                     with | I_XsM128d.Movapd | I_XsM128d.Movaps | I_XsM128d.Movdqa -> Some()
                                          | _ -> None
let (|I_XsM128d_SSEstoreUnaligned|_|) (x:I_XsM128d) =
                                      match x
                                       with | I_XsM128d.Movupd | I_XsM128d.Movups | I_XsM128d.Movdqu -> Some()
                                            | _ -> None

let (|I_XsM64d_SSEstoreLo|_|) (x:I_XsM64d) =
                              match x
                               with | I_XsM64d.Movlpd | I_XsM64d.Movlps | I_XsM64d.Movsd | I_XsM64d.Movq -> Some()
                                    | _ -> None


(* peephole1: the optimization window includes exactly one instruction.
              if vector register contents may be assumed to be untyped (AMD K10, Intel Core 2),
              more compact instruction encodings can often be chosen. *)
let peephole1 (cc:IBlock []) =
              let map1_AMDk10 (i:Instr) =
                           match i
                            with | XsXd(I_XsXd_SSEcopy, args)                                  -> XsXd(I_XsXd.Movaps, args) 
                                 | M128sXd(I_M128sXd_SSEloadAligned, args)                     -> M128sXd(I_M128sXd.Movaps, args)
                                 | XsM128d(I_XsM128d_SSEstoreAligned, args)                    -> XsM128d(I_XsM128d.Movaps, args)
                                 | XsM64d(I_XsM64d_SSEstoreLo, args)                           -> XsM64d(I_XsM64d.Movlps, args)
                                 | XsXds((I_XsXds.Pxor | I_XsXds.Xorpd), args)               -> XsXds(I_XsXds.Xorps, args)
                                 | M128sXds((I_M128sXds.Pxor | I_M128sXds.Xorpd), args)      -> M128sXds(I_M128sXds.Xorps, args)
                                 | XsXds(I_XsXds.Shufpd(LitInt32(1)), {S=s;DS=ds}) when s=ds -> XsXds(I_XsXds.Shufps(Lit.Int32(0x4e)), {S=s;DS=ds})
                                 | I0(I_I0.Retq)                                             -> I0(I_I0.Ret(Lit.Int32(0)))                       (* possibly better branch prediction *)
                                 | XsXd(I_XsXd.Pshufd(l), {S=s;D=d}) when s=d                -> XsXds(I_XsXds.Shufps(l), {S=s;DS=d})
                                 | XsXds(I_XsXds.Unpckhps, {S=s;DS=ds}) when s=ds            -> XsXd(I_XsXd.Pshufd(Lit.Int32(0xfa)), {S=s;D=ds}) (* allows LOAD&USE *)
                                 | x -> x
              let map1_IntelCore2 (i:Instr) = (* `super-shuffle engine' not present, so single precision shuffles are quite expensive *)
                                  match i
                                   with | XsXd(I_XsXd_SSEcopy, args)                             -> XsXd(I_XsXd.Movaps, args) 
                                        | M128sXd(I_M128sXd_SSEloadAligned, args)                -> M128sXd(I_M128sXd.Movaps, args)
                                        | XsM128d(I_XsM128d_SSEstoreAligned, args)               -> XsM128d(I_XsM128d.Movaps, args)
                                        | XsM64d(I_XsM64d_SSEstoreLo, args)                      -> XsM64d(I_XsM64d.Movlps, args)
                                        | XsXds((I_XsXds.Pxor | I_XsXds.Xorpd), args)          -> XsXds(I_XsXds.Xorps, args)
                                        | M128sXds((I_M128sXds.Pxor | I_M128sXds.Xorpd), args) -> M128sXds(I_M128sXds.Xorps, args)
                                        | XsXds(I_XsXds.Shufps(l), {S=s;DS=ds}) when s=ds      -> XsXd(I_XsXd.Pshufd(l), {S=s;D=ds})               (* less compact *)
                                        | XsXds(I_XsXds.Unpckhps, {S=s;DS=ds}) when s=ds       -> XsXd(I_XsXd.Pshufd(Lit.Int32(0xfa)), {S=s;D=ds}) (* allows LOAD&USE *)
                                        | x -> x
              let map1_IntelCoreI7 (i:Instr) = i
    
              let map1 (i:Instr) =
                       match i
                        with | R64d(I_R64d.Movq(LitInt32(0)),    {D=Q(d)}) -> R32d(I_R32d.Movl(Lit.Int32(0)), {D=L(d)})
                             | R64d(I_R64d.Movabsq(LitInt32(0)), {D=Q(d)}) -> R32d(I_R32d.Movl(Lit.Int32(0)), {D=L(d)})
                             | x -> match Basics.target()
                                     with | AMDk10      -> map1_AMDk10(x)
                                          | IntelCore2  -> map1_IntelCore2(x)
                                          | IntelCoreI7 -> map1_IntelCoreI7(x)

              cc |> Array.iter (fun b -> b.Instrs <- Array.map map1 b.Instrs)


// EA64   of Iexpr * EAbase<R64> option * EAscaledIndex<R64> option
                                      
                                    
let peepholeN (cc:IBlock []) =
              let rec loop (bag:list<Instr>) (instrs:list<Instr>) =
                           match instrs
                            with | [] -> List.rev(bag)
                                 | (M128sXd(I_M128sXd_SSEloadAligned, {S=M128.S(EA64(IexprInt32(ofs0),base0,index0));D=d0}))::
                                   (XsXd(I_XsXd_SSEcopy, {S=s1;D=d1}))::
                                   (XsXds(I_XsXds.Unpcklpd, {S=s2;DS=ds2}))::
                                   (XsXds(I_XsXds.Unpckhpd, {S=s3;DS=ds3}))::instrs
                                   when d0 = s1
                                     && s1 <> d1
                                     && s2 = s1
                                     && ds2 = d1
                                     && s3 = s1
                                     && ds3 = s1
                                        -> loop ((M64sXd(I_M64sXd.Movddup, {S=M64.S(EA64(Iexpr.Int32(  ofs0),base0,index0));D=ds2}))::
                                                 (M64sXd(I_M64sXd.Movddup, {S=M64.S(EA64(Iexpr.Int32(8+ofs0),base0,index0));D=ds3})):: bag) instrs


                                 | (XsXd(I_XsXd_SSEcopy, {S=s1;D=d1}))::
                                   (XsXds(I_XsXds.Unpcklpd, {S=s2;DS=ds2})):: instrs
                                   when s1 <> d1
                                     && s2 = s1
                                     && ds2 = d1
                                        -> loop ((XsXd(I_XsXd.Movddup, {S=s1;D=ds2}))::bag) instrs


                                 | (XsXd(I_XsXd_SSEcopy, {S=s1;D=d1}))::
                                   (XsXds(I_XsXds.Unpcklps, {S=s2;DS=ds2})) :: instrs
                                   when s1 <> d1
                                     && s2 = s1
                                     && ds2 = d1
                                        -> loop ((XsXd(I_XsXd.Pshufd(Lit.Int32(0x50)), {S=s1;D=ds2}))::bag) instrs



                                 | (XsXd(I_XsXd_SSEcopy, {S=s1;D=d1}))::
                                   (XsXds(I_XsXds.Unpckhps, {S=s2;DS=ds2})) :: instrs
                                   when s1 <> d1
                                     && s2 = s1
                                     && ds2 = d1
                                        -> loop ((XsXd(I_XsXd.Pshufd(Lit.Int32(0xfa)), {S=s1;D=ds2}))::bag) instrs


                                 | i::instrs -> loop (i::bag) instrs


              cc |> Array.iter (fun b ->
                                    eprintfn "block %s" b.Name
                                    b.Instrs <- b.Instrs
                                                |> Array.toList
                                                |> moveDown_Instrs_like b (function XsXd(_) | M128sXd(_) -> true | i -> not(Instr_is_Vec(i)))  (* move integer instructions out of the way *)
                                                |> moveDown_Instrs_like b (fun (i:Instr) -> Fd_of_Instr(i) <> [] || Fs_of_Instr(i) <> [])      (* move int instrs altering flags farther still... *)

                                                |> moveUp_Instrs_like   b (fun (instr,aux,_,_,_) instrs ->
                                                                                     match instr
                                                                                      with | XsXds(_, {S=s;DS=ds}) -> s=ds
                                                                                           | _ -> false)

//                                                      |> fun instrs ->
//                                                             for i in instrs do eprintfn "%A" i
//                                                             instrs

                                                |> loop []
                                                |> Array.ofList)


(*
let extractLoopIndVar (cc:IBlock []) =
                      let candidates = new HashSet<R64>()
                      for b in cc do
                            b.Instrs.Length > 2 
*)
