(* nxcc ... C compiler driver *)

(*

icc -S -fno-verbose-asm

what about icc -ffunction-sections?
   *)

open System
open NXyn.Aux


(* ==> NXyn.Globals *)
let logfile = ref "/dev/null"

(***)

let ccoptsTakingArgument = [
                            "--param"
                            "-A"
                            "-D"
                            "-I"
                            "-MF"; "-MQ"; "-MT"
                            "-U"
                            "-V"
                            "-Xassembler"; "-Xlinker"; "-Xpreprocessor"
                            "-aux-info"
                            "-b"
                            "-idirafter"; "-imacros"; "-imultilib"; "-include"; "-iprefix"; "-iquote"; "-isysroot"; "-isystem"; "-iwithprefix"; "-iwithprefixbefore"
                            "-l"
                            "-o"
                            "-u"
                            "-x"
                           ]

(***) 

let ccopts_of_ccargs(x) = let rec loop bag = function [] -> List.rev(bag)
                                                    | opt::opts when not(ListMisc.member_of opt ccoptsTakingArgument) -> loop ((opt,None)::bag) opts
                                                    | [opt] -> failwithf "missing argument to option '%s'" opt
                                                    | opt::arg::opts -> loop ((opt,Some(arg))::bag) opts
                          loop [] x

let ccargs_of_ccopts(x) = x
                          |> List.map (fun(opt,argOpt) -> opt::(match argOpt with None -> [] | Some(arg) -> [arg]))
                          |> List.concat

(***)

let cFileExtensions = [| ".c"; ".i" |]
let sFileExtensions = [| ".S"; ".s"; ".sx" |]

let sourceFileExtensions = Array.concat [| cFileExtensions; sFileExtensions |]

let has_extension s extensions = Array.exists (fun(ending) -> (s:string).EndsWith(ending)) extensions

(***)

let is_sourceFileName(f:string) = not(f.StartsWith("-")) && Array.exists (fun(es) -> f.EndsWith(es)) sourceFileExtensions

let langOpt_of_string(s:string) = if s = "none" then None else Some(s)

let getSrcFiles(ccopts) = let rec do1 (langOpt,srcs,opts) = function ("-x",Some(l))                  -> (langOpt_of_string(l),   srcs,    opts)
                                                                   | ("-",None)                      -> (langOpt, ("-",langOpt)::srcs,    opts)
                                                                   | (f,_) when is_sourceFileName(f) -> (langOpt,   (f,langOpt)::srcs,    opts)
                                                                   | (opt,argOpt) as p               -> (langOpt,                srcs, p::opts)
                          let (_,srcs,opts) = List.fold do1 (None,[],[]) ccopts
                          (List.rev(srcs), List.rev(opts))

(***)

type xn = C2asm   of string * string
        | Asm2obj of string * string
        | Asm2asm of string * string
        | Obj2lib of string list * string

(***)

let isCsource(file,langOpt) = langOpt = Some("c")         || (langOpt = None && has_extension file cFileExtensions)
let isSsource(file,langOpt) = langOpt = Some("assembler") || (langOpt = None && has_extension file sFileExtensions)

(***)

let invoke_cc (cc:string) (nxynargs:list<String>) (ccargs:list<String>) =
    (* combine arguments that take additional arguments *)
    let ccopts = ccopts_of_ccargs(ccargs)

    (* filter out occurrences of "-c" option *)
    let (compileOnlies,ccopts) = List.partition ((=)("-c",None)) ccopts
    let compileOnly = compileOnlies <> []

    (* get all files that are possible source files *)
    let (srcFiles,ccopts) = getSrcFiles(ccopts)
    if ListMisc.member_of ("-",None) srcFiles then failwith("-E or -x required when input is from standard input")

    (* take last "-o" filename, if any *)
    let (outputFileNames,ccopts) = List.partition (fun(opt,_) -> opt = "-o") ccopts
    let outputFileNameOption = if outputFileNames = [] then None else Some(outputFileNames |> List.rev |> List.head |> snd |> Option.get)
    if compileOnly && outputFileNameOption <> None && srcFiles.Length > 1 then failwith("cannot specify -o with -c or -S with multiple files")
    let getObjFileName(filename) = match outputFileNameOption with Some(x) -> x | None -> IOMisc.RemoveFileExtension(filename) + ".o"

    (* create plan *)
    let plan = new ResizeArray<xn>()
    let compileTo cFile oFile = let sFile1 = IOMisc.GetNXCCtmpFileName ".S" cFile
                                let sFile2 = IOMisc.GetNXCCtmpFileName ".S" cFile
                                plan.Add(C2asm(cFile,sFile1))
                                plan.Add(Asm2asm(sFile1,sFile2))
                                plan.Add(Asm2obj(sFile2,oFile))
    let assembleTo sFile1 oFile = let sFile2 = IOMisc.GetNXCCtmpFileName ".S" sFile1
                                  plan.Add(Asm2asm(sFile1,sFile2))
                                  plan.Add(Asm2obj(sFile2,oFile))
                                  
    let dispatch1 do1 (file,langOpt) = if   isCsource(file,langOpt) then do1 compileTo  file
                                       elif isSsource(file,langOpt) then do1 assembleTo file
                                                                    else failwithf "nxcc: ignoring file %s (%A)" file langOpt
    srcFiles
    |> match compileOnly with true -> List.iter (dispatch1 (fun doWhat file -> getObjFileName(file) |> doWhat file))
                            | false -> List.map (dispatch1 (fun doWhat file -> let oFile = IOMisc.GetNXCCtmpFileName ".o" file in doWhat file oFile; oFile))
                                       >> fun objFiles -> plan.Add(Obj2lib(objFiles,defaultArg outputFileNameOption "a.out"))

    (* execute plan *)
    let ccargs = ccargs_of_ccopts(ccopts)

    let d_i_args = let cwd = Environment.CurrentDirectory
                   let w   = new IO.StringWriter()
                   let rec loop = function
                       | [] -> ()
                       | ["-I"] -> ()
                       | arg::args when arg.StartsWith("-D") -> fprintf w "%s " arg; loop(args)
                       | arg::args when arg.StartsWith("-U") -> fprintf w "%s " arg; loop(args)
                       | "-I"::dir::args                     -> fprintf w "-I%s " (IO.Path.Combine(cwd,dir.Trim())); loop args
                       | arg::args when arg.StartsWith("-I") -> fprintf w "-I%s " (IO.Path.Combine(cwd,arg.Substring("-I".Length).Trim())); loop args
                       | _::args -> loop(args)
                   loop(ccargs)
                   w.ToString().Trim()

    let assertEQ x y = if x <> y then failwith "assertEQ"
    
    let invokeAS       inFiles outFile = assertEQ 0 <| ProcessMisc.invoke_with_args !logfile cc       (["-x";"assembler";"-c"] @ inFiles @ ["-o";outFile] @ ccargs)
    let invokeCC  args inFiles outFile = assertEQ 0 <| ProcessMisc.invoke_with_args !logfile cc       (args @ inFiles @ ["-o"; outFile] @ ccargs)
    let invokeIndent   inFiles outFile = assertEQ 0 <| ProcessMisc.invoke_with_args !logfile "indent" (["-l";"5000000";"-bls";"-blf";"-npsl";"-nbc";"-fca"] @ inFiles @ ["-o";outFile])

    plan.ToArray()
    |> Array.iter (function Asm2asm(sFile1,sFile2)  -> let tFile1 = sFile1 + ".nxas.json1"
                                                       let tFile2 = sFile1 + ".nxas.json2"
                                                       let oFile1 = sFile1 + ".nxas.1.o"
                                                       let oFile2 = sFile1 + ".nxas.2.o"
                                                       let hdrFile1 = sFile1 + ".nxas.1.objdump_hdr"
                                                       let hdrFile2 = sFile2 + ".nxas.2.objdump_hdr"
                                                           
                                                       invokeAS [sFile1] oFile1
                                                       ProcessMisc.invoke_with_args !logfile "nxyn_objdumphdr"   [oFile1;hdrFile1] |> assertEQ 0
                                                           
                                                           
                                                       ProcessMisc.invoke_with_args !logfile "nxas_parse"   [!logfile; sFile1; tFile1] |> assertEQ 0
                                                       ProcessMisc.invoke_with_args !logfile "nxas_opt"     ([!logfile; tFile1; tFile2] @ nxynargs) |> assertEQ 0
                                                       ProcessMisc.invoke_with_args !logfile "nxas_unparse" [!logfile; tFile2; sFile2] |> assertEQ 0

                                                       invokeAS [sFile2] oFile2
                                                       ProcessMisc.invoke_with_args !logfile "nxyn_objdumphdr"   [oFile2;hdrFile2] |> assertEQ 0

                                                       let tryGetSize_of_objdumphdr(lines) =
                                                           match lines |> Array.filter (fun (s:String) -> s.Contains(".text")) with
                                                               | [| line |] ->
                                                                   match Array.toList(line.Split([| " "; "\t" |], StringSplitOptions.RemoveEmptyEntries)) with
                                                                       | "0"::".text"::sizeStr::_ -> ConvertMisc.base16bigintconv.TryParse(sizeStr)
                                                                       | _ -> None
                                                               | _ -> None

                                                       let hdrLines1 = IO.File.ReadAllLines(hdrFile1)
                                                       let hdrLines2 = IO.File.ReadAllLines(hdrFile2)

                                                       let size1 = tryGetSize_of_objdumphdr(hdrLines1)
                                                       let size2 = tryGetSize_of_objdumphdr(hdrLines2)

                                                       match (size1,size2) with
                                                           | (Some(size1),Some(size2)) ->
                                                               let size1 = int64 size1
                                                               let size2 = int64 size2
                                                               let str = sprintf "%s: codesize %d => %d (%.2f %%)" sFile1 size1 size2 ((float size2) * 100.0 / (float size1))
                                                               ignore (ProcessMisc.append_line_to_file str !logfile)
                                                           | _ -> let str = sprintf "%s: codesize UNKNOWN!" sFile1
                                                                  ignore (ProcessMisc.append_line_to_file str !logfile)

                                                           
                          | Asm2obj(sFile,oFile)    -> invokeAS    [sFile]  oFile
                          | Obj2lib(oFiles,outFile) -> invokeCC [] oFiles   outFile
                          | C2asm(cFile,sFile)      -> invokeCC ["-x";"c";"-S"]  [cFile] sFile 
                                                       let cFile1 = IOMisc.GetNXCCtmpFileName ".c" cFile
                                                       IO.File.Delete(cFile1) (* Intel C compiler 'icc' needs this *)
                                                       invokeCC ["-E"; "-Wp,-P"] [cFile]  cFile1 
                                                       let cFile2 = IOMisc.GetNXCCtmpFileName ".c" cFile
                                                       invokeIndent              [cFile1] cFile2
                                                         
                                                       let cLines = IO.File.ReadAllLines(cFile2)
                                                       let isInteresting(line:string) =
                                                            (line.StartsWith("void") || line.StartsWith("static void"))
                                                            && (not(line.EndsWith(";")))
                                                            && (match Array.toList(line.Split([| "(" |], StringSplitOptions.RemoveEmptyEntries)) with
                                                                    | [pre;post] -> not(pre.Contains("*"))
                                                                    | _ -> false)

                                                       IO.File.Delete(cFile1)
                                                       IO.File.Delete(cFile2)

                                                       let decls = cLines |> Array.filter isInteresting |> Array.map (fun (l:string) -> sprintf "\t.ascii \"%s\"" l)

                                                       [
                                                        IO.File.ReadAllLines(sFile)
                                                        [| ".section .rodata.NXyn" |]
                                                        [| "NXynCCargs:" |]
                                                        [| sprintf "\t.ascii \"%s\"" d_i_args |]
                                                        [| "NXynFunProtos:" |]
                                                        decls
                                                       ]
                                                       |> fun parts -> IO.File.WriteAllLines(sFile,parts |> Array.concat))
    0 (* all done => exit-code zero *)
        
(***)

let isWS(c:Char) = Char.IsWhiteSpace(c)

let split(r) = let args = new ResizeArray<string>()
               let rec loop() = let c = (r:IO.TextReader).Read()
                                if c >= 0 then let c = char c
                                               if isWS(c) then loop()
                                                          else let w = new IO.StringWriter()
                                                               let rec readArg(c) = w.Write(c:Char)
                                                                                    if c = '\"' then readQuoted()
                                                                                                else let c = r.Read()
                                                                                                     if c >= 0 then let c = char c
                                                                                                                    if not(isWS(c)) then readArg(c)
                                                               and readQuoted() = let c = r.Read()
                                                                                  if c >= 0 then let c = char c
                                                                                                 w.Write(c)
                                                                                                 if c = '\\' then if r.Peek() = int '\"' then ignore(r.Read())
                                                                                                                                              w.Write('\"')
                                                                                                                  readQuoted()
                                                                                                 elif c <> '\"' then readQuoted()
                                                               readArg(c)
                                                               args.Add(w.ToString())
                                                               loop()
               loop()
               args.ToArray()


let expandCommandLineArgs (ccargs:list<String>) =
    let all = new ResizeArray<String>()
    let rec loop preds args = args |> List.iter(fun(arg:String) -> if arg.StartsWith("@") then arg.Substring(1) |> traverse preds
                                                                                          else all.Add(arg))
    and traverse preds filename = if not(IO.File.Exists(filename)) then failwithf "nxcc: param file '%s' not found" filename
                                  if ListMisc.member_of filename preds then failwithf "nxcc: param file '%s' depends on %A (cycle)" filename preds
                                  split(new IO.StringReader(IO.File.ReadAllText(filename)))
                                  |> Array.toList
                                  |> loop (filename::preds)
    loop [] ccargs
    Array.toList(all.ToArray())

let doBypass(ccargs:list<String>) = ccargs = [] || List.exists (fun(arg) -> ListMisc.member_of arg ["-"; "--version"; "-E"; "-S"]) ccargs

let main = function me::cc::log::args -> logfile := log
                                         let (nxynargs,ccargs0) = List.partition (fun(s:String) -> s.StartsWith("--nxyn")) args
                                         let ccargs = expandCommandLineArgs(ccargs0)
                                         if doBypass(ccargs) then ProcessMisc.invoke_with_args !logfile cc ccargs0
                                                             else try  invoke_cc cc nxynargs ccargs
                                                                  with exn -> ProcessMisc.invoke_with_args !logfile cc ccargs0
                                         |> exit
                  | _ -> eprintfn "usage: nxcc CC LOGFILE [CCargs]"; exit 1


main(Environment.GetCommandLineArgs() |> Array.toList)
