module X64instr

open System
open System.Collections.Generic

(*
  instructions not supported:
    * privileged instructions: IO (in, out, ins[bwl], out[bwl]), rdmsr, wrmsr
    * MMX
    * x87
    * 3Dnow (except for prefetch, prefetchw, which are part of an separate ISA)
    * syscall
*)

(***)

(* AMD64 FLAG *)
type F =  | (* overflow   *) O
  (* F *) | (* direction  *) D
  (* F *) | (* interrupt  *) I
  (* F *) | (* trap       *) T
  (* F *) | (* sign       *) S
  (* F *) | (* zero       *) Z
  (* F *) | (* aux. carry *) A
  (* F *) | (* parity     *) P
  (* F *) | (* carry      *) C
with override x.ToString() = match x with O -> "o" | D -> "d" | I -> "i" | T -> "t" | S -> "s" | Z -> "z" | A -> "a" | P -> "p" | C -> "c"
     static member All = [O;D;I;T;S;Z;A;P;C]

(***)

let R8Lname  = [|  "%al"; "%cl"; "%dl"; "%bl";"%spl";"%bpl";"%sil";"%dil";"%r8b";"%r9b";"%r10b";"%r11b";"%r12b";"%r13b";"%r14b";"%r15b" |]
let R8Hname  = [|  "%ah"; "%ch"; "%dh"; "%bh" |]
let R16name  = [|  "%ax"; "%cx"; "%dx"; "%bx"; "%sp"; "%bp"; "%si"; "%di";"%r8w";"%r9w";"%r10w";"%r11w";"%r12w";"%r13w";"%r14w";"%r15w" |]
let R32name  = [| "%eax";"%ecx";"%edx";"%ebx";"%esp";"%ebp";"%esi";"%edi";"%r8d";"%r9d";"%r10d";"%r11d";"%r12d";"%r13d";"%r14d";"%r15d" |]
let R64name  = [| "%rax";"%rcx";"%rdx";"%rbx";"%rsp";"%rbp";"%rsi";"%rdi";"%r8"; "%r9"; "%r10"; "%r11"; "%r12"; "%r13"; "%r14"; "%r15"  |]

let Rnames = [R8Lname; R8Hname; R16name; R32name; R64name] |> Array.concat 

type R8    = Lo of int
  (* R8 *) | Hi of int
with override x.ToString() = match x with Lo(i) -> R8Lname.[i] | Hi(i) -> R8Hname.[i]
     member   x.Int        = match x with Lo(i) | Hi(i) -> i

     static member al   = Lo(0)
     static member cl   = Lo(1)
     static member dl   = Lo(2)
     static member bl   = Lo(3)
     static member spl  = Lo(4)
     static member bpl  = Lo(5)
     static member sil  = Lo(6)
     static member dil  = Lo(7)
     static member r8b  = Lo(8)
     static member r9b  = Lo(9)
     static member r10b = Lo(10)
     static member r11b = Lo(11)
     static member r12b = Lo(12)
     static member r13b = Lo(13)
     static member r14b = Lo(14)
     static member r15b = Lo(15)
     static member ah   = Hi(0)
     static member ch   = Hi(1)
     static member dh   = Hi(2)
     static member bh   = Hi(3)

     static member LoL  = R8.al
     static member HiL  = R8.r8b

                                        
type R16 = W of int
with override x.ToString() = match x with W(i) -> R16name.[i]
     member   x.Int        = match x with W(i) -> i

     static member ax   = W(0)
     static member cx   = W(1)
     static member dx   = W(2)
     static member bx   = W(3)
     static member sp   = W(4)
     static member bp   = W(5)
     static member si   = W(6)
     static member di   = W(7)
     static member r8w  = W(8)
     static member r9w  = W(9)
     static member r10w = W(10)
     static member r11w = W(11)
     static member r12w = W(12)
     static member r13w = W(13)
     static member r14w = W(14)
     static member r15w = W(15)

     static member Lo   = R16.ax
     static member Hi   = R16.r8w

type R32 = L of int
with override x.ToString() = match x with L(i) -> R32name.[i]
     member   x.Int        = match x with L(i) -> i

     static member eax  = L(0)
     static member ecx  = L(1)
     static member edx  = L(2)
     static member ebx  = L(3)
     static member esp  = L(4)
     static member mESP = L(-4)
     static member ebp  = L(5)
     static member esi  = L(6)
     static member edi  = L(7)
     static member r8d  = L(8)
     static member r9d  = L(9)
     static member r10d = L(10)
     static member r11d = L(11)
     static member r12d = L(12)
     static member r13d = L(13)
     static member r14d = L(14)
     static member r15d = L(15)

     static member Lo   = R32.eax
     static member Hi   = R32.r8d

type R64 = Q of int 
with override x.ToString() = match x with Q(i) -> if i < R64name.Length then R64name.[i] else sprintf "%%R%d" i
     member   x.Int        = match x with Q(i) -> i

     static member rax  = Q(0)
     static member rcx  = Q(1)
     static member rdx  = Q(2)
     static member rbx  = Q(3)
     static member rsp  = Q(4)
     static member mRSP = Q(-4)
     static member rbp  = Q(5)
     static member rsi  = Q(6)
     static member rdi  = Q(7)
     static member r8   = Q(8)
     static member r9   = Q(9)
     static member r10  = Q(10)
     static member r11  = Q(11)
     static member r12  = Q(12)
     static member r13  = Q(13)
     static member r14  = Q(14)
     static member r15  = Q(15)
     static member Lo   = R64.rax // bp // [was:] rax
     static member Hi   = R64.r8
     static member All       = [R64.rax; R64.rbx; R64.rcx; R64.rdx; R64.rsp; R64.rbp; R64.rsi; R64.rdi; R64.r8; R64.r9; R64.r10; R64.r11; R64.r12; R64.r13; R64.r14; R64.r15]

     static member CalleeSaved = [R64.rbx; R64.rbp; R64.r12; R64.r13; R64.r14; R64.r15]

     static member EncodedBest = [ R64.rax; R64.rbx; R64.rcx; R64.rdx; R64.rsi; R64.rdi ]



(* general-purpose scalar integer register *)
type R =  | Rbyte of R8
  (* R *) | Rword of R16
  (* R *) | Rlong of R32
  (* R *) | Rquad of R64
with override x.ToString() = match x with Rbyte(i) -> string(i) | Rword(i) -> string(i) | Rlong(i) -> string(i) | Rquad(i) -> string(i)
     member   x.Int        = match x with Rbyte(Lo(i) | Hi(i)) | Rword(W(i)) | Rlong(L(i)) | Rquad(Q(i)) -> i
     member   r1.Aliases (r2:R) = match (r1,r2) with Rbyte(x), Rbyte(y) -> x=y
                                                   | (Rbyte(Lo(x)) | Rbyte(Hi(x)) | Rword(W(x)) | Rlong(L(x)) | Rquad(Q(x))),   (Rword(W(y)) | Rlong(L(y)) | Rquad(Q(y))) -> x=y
                                                   | (Rword(W(x)) | Rlong(L(x)) | Rquad(Q(x))),   (Rbyte(Lo(y)) | Rbyte(Hi(y)) | Rword(W(y)) | Rlong(L(y)) | Rquad(Q(y))) -> x=y
     static member al   = Rbyte(R8.al)
     static member bl   = Rbyte(R8.bl)
     static member cl   = Rbyte(R8.cl)
     static member dl   = Rbyte(R8.dl)
     static member spl  = Rbyte(R8.spl)
     static member bpl  = Rbyte(R8.bpl)
     static member sil  = Rbyte(R8.sil)
     static member dil  = Rbyte(R8.dil)
     static member r8b  = Rbyte(R8.r8b)
     static member r9b  = Rbyte(R8.r9b)
     static member r10b = Rbyte(R8.r10b)
     static member r11b = Rbyte(R8.r11b)
     static member r12b = Rbyte(R8.r12b)
     static member r13b = Rbyte(R8.r13b)
     static member r14b = Rbyte(R8.r14b)
     static member r15b = Rbyte(R8.r15b)

     static member ah   = Rbyte(R8.ah)
     static member bh   = Rbyte(R8.bh)
     static member ch   = Rbyte(R8.ch)
     static member dh   = Rbyte(R8.dh)

     static member ax   = Rword(R16.ax)
     static member bx   = Rword(R16.bx)
     static member cx   = Rword(R16.cx)
     static member dx   = Rword(R16.dx)
     static member sp   = Rword(R16.sp)
     static member bp   = Rword(R16.bp)
     static member si   = Rword(R16.si)
     static member di   = Rword(R16.di)
     static member r8w  = Rword(R16.r8w)
     static member r9w  = Rword(R16.r9w)
     static member r10w = Rword(R16.r10w)
     static member r11w = Rword(R16.r11w)
     static member r12w = Rword(R16.r12w)
     static member r13w = Rword(R16.r13w)
     static member r14w = Rword(R16.r14w)
     static member r15w = Rword(R16.r15w)

     static member eax  = Rlong(R32.eax)
     static member ebx  = Rlong(R32.ebx)
     static member ecx  = Rlong(R32.ecx)
     static member edx  = Rlong(R32.edx)
     static member esp  = Rlong(R32.esp)
     static member ebp  = Rlong(R32.ebp)
     static member esi  = Rlong(R32.esi)
     static member edi  = Rlong(R32.edi)
     static member r8d  = Rlong(R32.r8d)
     static member r9d  = Rlong(R32.r9d)
     static member r10d = Rlong(R32.r10d)
     static member r11d = Rlong(R32.r11d)
     static member r12d = Rlong(R32.r12d)
     static member r13d = Rlong(R32.r13d)
     static member r14d = Rlong(R32.r14d)
     static member r15d = Rlong(R32.r15d)
  
     static member rax  = Rquad(R64.rax)
     static member rbx  = Rquad(R64.rbx)
     static member rcx  = Rquad(R64.rcx)
     static member rdx  = Rquad(R64.rdx)
     static member rsp  = Rquad(R64.rsp)
     static member rbp  = Rquad(R64.rbp)
     static member rsi  = Rquad(R64.rsi)
     static member rdi  = Rquad(R64.rdi)
     static member r8   = Rquad(R64.r8)
     static member r9   = Rquad(R64.r9)
     static member r10  = Rquad(R64.r10)
     static member r11  = Rquad(R64.r11)
     static member r12  = Rquad(R64.r12)
     static member r13  = Rquad(R64.r13)
     static member r14  = Rquad(R64.r14)
     static member r15  = Rquad(R64.r15)
     static member All  = [R.rax; R.rbx; R.rcx; R.rdx; R.rsp; R.rbp; R.rsi; R.rdi; R.r8; R.r9; R.r10; R.r11; R.r12; R.r13; R.r14; R.r15]

let size_of_R = function Rbyte _ -> 1 | Rword _ -> 2 | Rlong _ -> 4 | Rquad _ -> 8

let Int32_of_R8(r:R8) = r.Int
let Int32_of_R16(r:R16) = r.Int
let Int32_of_R32(r:R32) = r.Int
let Int32_of_R64(r:R64) = r.Int

(***)

let X128name = [| "%xmm0";"%xmm1";"%xmm2";"%xmm3";"%xmm4";"%xmm5";"%xmm6";"%xmm7";"%xmm8";"%xmm9";"%xmm10";"%xmm11";"%xmm12";"%xmm13";"%xmm14";"%xmm15" |]

(* SSE vector register *)
type X = X128 of int
with override x.ToString() = match x with X128(i) -> X128name.[i]
     member   x.Int        = match x with X128(i) -> i
     member   x.IsSmall    = let v = x.Int in v >= 0 && v < 8
     member   x.Bucket01   = let v = x.Int in if x.IsSmall then 0 else 1
     member   x.Normalize  = match x with X128(i) -> if x.IsSmall then X.xmm0 else X.xmm8

     static member xmm0  = X128(0)
     static member xmm1  = X128(1)
     static member xmm2  = X128(2)
     static member xmm3  = X128(3)
     static member xmm4  = X128(4)
     static member xmm5  = X128(5)
     static member xmm6  = X128(6)
     static member xmm7  = X128(7)
     static member xmm8  = X128(8)
     static member xmm9  = X128(9)
     static member xmm10 = X128(10)
     static member xmm11 = X128(11)
     static member xmm12 = X128(12)
     static member xmm13 = X128(13)
     static member xmm14 = X128(14)
     static member xmm15 = X128(15)
     static member Lo    = X.xmm0
     static member Hi    = X.xmm8
     static member All   = [X.xmm0;X.xmm1;X.xmm2;X.xmm3;X.xmm4;X.xmm5;X.xmm6;X.xmm7;X.xmm8;X.xmm9;X.xmm10;X.xmm11;X.xmm12;X.xmm13;X.xmm14;X.xmm15]

let Int32_of_X(x:X) = x.Int

(***)

let Y256name = [| "%ymm0";"%ymm1";"%ymm2";"%ymm3";"%ymm4";"%ymm5";"%ymm6";"%ymm7";"%ymm8";"%ymm9";"%ymm10";"%ymm11";"%ymm12";"%ymm13";"%ymm14";"%ymm15" |]

(* AVX vector register *)
type Y = Y256 of int
with override x.ToString() = match x with Y256(i) -> Y256name.[i]
     member   x.Int        = match x with Y256(i) -> i
     member   x.IsSmall    = let v = x.Int in v >= 0 && v < 8
     member   x.Bucket01   = let v = x.Int in if x.IsSmall then 0 else 1
     member   x.Normalize  = match x with Y256(i) -> if x.IsSmall then Y.ymm0 else Y.ymm8

     static member ymm0  = Y256(0)
     static member ymm1  = Y256(1)
     static member ymm2  = Y256(2)
     static member ymm3  = Y256(3)
     static member ymm4  = Y256(4)
     static member ymm5  = Y256(5)
     static member ymm6  = Y256(6)
     static member ymm7  = Y256(7)
     static member ymm8  = Y256(8)
     static member ymm9  = Y256(9)
     static member ymm10 = Y256(10)
     static member ymm11 = Y256(11)
     static member ymm12 = Y256(12)
     static member ymm13 = Y256(13)
     static member ymm14 = Y256(14)
     static member ymm15 = Y256(15)
     static member Lo    = Y.ymm0
     static member Hi    = Y.ymm8
     static member All   = [Y.ymm0;Y.ymm1;Y.ymm2;Y.ymm3;Y.ymm4;Y.ymm5;Y.ymm6;Y.ymm7;Y.ymm8;Y.ymm9;Y.ymm10;Y.ymm11;Y.ymm12;Y.ymm13;Y.ymm14;Y.ymm15]

let Int32_of_Y (y:Y) = y.Int

(***)

let Vnames = [X128name; Y256name] |> Array.concat

let AllRegisterNames = [Rnames; Vnames] |> Array.concat

(***)

type Operand =  | Scalar of Int32     (* scalar register *)
  (* Operand *) | Vector of Int32     (* vector register (SSE/AVX) *)
  (* Operand *) | Gmemory             (* global memory, unknown, possibly aliased *)
  (* Operand *) | Flag of F           (* flag *)
  (* Operand *) | PstackB of Int32    (* thread-local private stack byte *)
  (* Operand *) | Control             (* control flow *)
with static member rsp = Scalar(R.rsp.Int)
     override x.ToString() = match x with Scalar(i)  -> R64.Q(i).ToString()
                                        | Vector(i)  -> Y256name.[i]
                                        | Flag(f)    -> string(f)
                                        | PstackB(i) -> sprintf "rsp[%d]" i
                                        | Gmemory    -> "memory"
                                        | Control    -> "control"

let Operand_needs_ReadPort = function Scalar _ | Vector _ | Flag _ -> true
                                    | Gmemory | PstackB _ | Control -> false

let Operand_is_Scalar = function Scalar(_) -> true | Vector(_) | Gmemory | Flag(_) | PstackB(_) | Control -> false
let Operand_is_Vector = function Vector(_) -> true | Scalar(_) | Gmemory | Flag(_) | PstackB(_) | Control -> false

let Operand_is_Scalar_or_Vector = function Scalar(_) | Vector(_) -> true | Gmemory | Flag(_) | PstackB(_) | Control -> false

let Int32_of_VectorOperand (o:Operand) =
                           match o
                            with | Vector(v) -> v
                                 | Scalar(_) | PstackB(_) | Gmemory | Flag(_) | Control -> failwith "Int32_of_ScalarOperand"

let Int32_of_ScalarOperand (o:Operand) =
                           match o
                            with | Scalar(v) -> v
                                 | Vector(_) | PstackB(_) | Gmemory | Flag(_) | Control -> failwith "Int32_of_ScalarOperand"

let Int32_of_Operand (o:Operand) =
                     match o
                      with | Vector(v) | Scalar(v) | PstackB(v) -> v
                           | Gmemory | Flag(_) | Control -> failwith "Int32_of_Operand"


let Operand_of_F = function f -> Operand.Flag(f)
let Operand_of_R = function Rbyte(Lo(i))| Rbyte(Hi(i)) | Rword(W(i)) | Rlong(L(i)) | Rquad(Q(i)) -> Operand.Scalar(i)
let Operand_of_X = function X128(i) -> Operand.Vector(i)
let Operand_of_Y = function Y256(i) -> Operand.Vector(i)
//let Operand_of_V = function V64(i)  -> Operand.MMXvec(i)

let AllRegistersSet = Set.ofList(List.map Operand_of_R R.All @ List.map Operand_of_X X.All @ List.map Operand_of_Y Y.All @ List.map Operand_of_F F.All)

(* => move to R *)
let CalleeSavedRegisters = [R.rbx; R.rsp; R.rbp; R.r12; R.r13; R.r14; R.r15]
(*
   => move to R
   Inputs
     [R.rax; R.rcx; R.rdx; R.rsi; R.rdi; R.r8; R.r9; R.r10; R.rsp]
   Outputs
     [R.rax; R.rdx]
*)

let MinReturnRegisters = List.map Operand_of_R CalleeSavedRegisters 
//                         @List.map Operand_of_V V.All
let MaxReturnRegisters = MinReturnRegisters @ List.map Operand_of_R [R.rax; R.rdx] @ List.map Operand_of_Y [Y.ymm0; Y.ymm1]

(* XXX move in part to type R or R64 *)
let ABIParamRegistersSet = let rs = List.map Operand_of_Y (List.map Y256 [0..7]) @
                                    List.map Operand_of_R [R.rax (* # of varargs *); R.rcx; R.rdx; R.rsi; R.rdi; R.r8; R.r9; R.r10 (* static chain pointer *); R.rsp]
                           Set.ofList(rs)

(***)


type AKAAttribute(s:string)    = inherit Attribute(); member this.GetValue = s
let string_of_AKA(s:obj)       = (s :?> AKAAttribute).GetValue

type DAttribute(col:int)       = inherit Attribute(); member this.GetValue = col
type DSAttribute(col:int)      = inherit Attribute(); member this.GetValue = col
type SAttribute(col:int)       = inherit Attribute(); member this.GetValue = col

type VecAttribute()            = inherit Attribute()

type NopAttribute()            = inherit Attribute()

type MemAttribute()            = inherit Attribute() (* play safe when code is easily broken due to unknown aliasing with implicit operands *)
type ExnAttribute()            = inherit Attribute() 


type RdAttribute(s:string)     = inherit Attribute(); member this.GetValue = s
type RdsAttribute(s:string)    = inherit Attribute(); member this.GetValue = s
type RsAttribute(s:string)     = inherit Attribute(); member this.GetValue = s

type XdAttribute(s:string)     = inherit Attribute(); member this.GetValue = s
type XdsAttribute(s:string)    = inherit Attribute(); member this.GetValue = s
type XsAttribute(s:string)     = inherit Attribute(); member this.GetValue = s

type YdAttribute(s:string)     = inherit Attribute(); member this.GetValue = s
type YdsAttribute(s:string)    = inherit Attribute(); member this.GetValue = s
type YsAttribute(s:string)     = inherit Attribute(); member this.GetValue = s

type ISAAttribute(s:string)    = inherit Attribute(); member this.GetValue = s

type JfollowAttribute()        = inherit Attribute()
type JthruAttribute()          = inherit Attribute()
type RetAttribute()            = inherit Attribute() (* does a instruction return from a function? *)

type FsAttribute(s:string)     = inherit Attribute(); member this.GetValue = s
type FdAttribute(s:string)     = inherit Attribute(); member this.GetValue = s

type PfxAttribute()            = inherit Attribute() (* instruction prefix *)

type IdAttribute()             = inherit Attribute()

type AttPosAttribute(i:Int32)  = inherit Attribute(); member this.GetValue = i


type RepAttribute()            = inherit Attribute()

(***)

(* AMD64 ISA extension *)
type ISAextension =  | AMD_PF   = 1        (* AMD prefetch/prefetchW                 [AMD K7 Athlon] *) 
  (* ISAextension *) | SSE3     = 2        (* Intel SSE3                  [Intel Pentium 4 Prescott] *) 
  (* ISAextension *) | SSSE3    = 4        (* Intel Supplemental SSE3          [Intel Core 2 Conroe] *) 
  (* ISAextension *) | SSE4_1   = 8        (* Intel SSE4 part 1                [Intel Core 2 Penryn] *) 
  (* ISAextension *) | SSE4_2   = 16       (* Intel SSE4 part 2              [Intel Core i7 Nehalem] *) 
  (* ISAextension *) | SSE4a    = 32       (* AMD SSE4a                          [AMD K10 Barcelona] *) 
  (* ISAextension *) | CLMUL    = 64       (* Carry-less multiplication     [Intel Core i5 Westmere] *) 
  (* ISAextension *) | AES      = 128      (* Advanced Encryption Standard  [Intel Core i5 Westmere] *) 
  (* ISAextension *) | MOVBE    = 256      (* Move bytes and swap                       [Intel Atom] *) 
  (* ISAextension *) | POPC     = 512      (* Hamming weight                                         *) 
  (* ISAextension *) | AVX      = 1024     (* Advanced Vector eXtensions        <<future Intel/AMD>> *) 
  (* ISAextension *) | XOP      = 2048     (* XOP                                     <<future AMD>> *) 
  (* ISAextension *) | FMA3     = 4096     (* Fused Multiply-Add                    <<future Intel>> *) 
  (* ISAextension *) | FMA4     = 8192     (* Fused Multiply-Add                      <<future AMD>> *) 
  (* ISAextension *) | F16C     = 16384    (* FP16 support                          <<future Intel>> *) 
  (* ISAextension *) | FSGSBASE = 32768    (* FS/GS base access                     <<future Intel>> *) 
  (* ISAextension *) | RDRAND   = 65536    (* HW random number generator            <<future Intel>> *) 

(***)

let fits_signed_byte(i:Int64)  = i >= -128L && i <= 127L
let fits_signed_dword(i:Int64) = i >= -2147483648L && i <= 2147483647L


type IunaryOp =  | A_NEG    (* arithmetic *)
  (* IunaryOp *) | B_NEG    (* bitwise *)
with override x.ToString() = match x with A_NEG -> "-" | B_NEG -> "~"

type IbinaryOp =  | A_ADD | A_DIV | A_MUL | A_REM | A_SHL | A_SHR | A_SUB            (* arithmetic *) 
  (* IbinaryOp *) | B_AND | B_IOR | B_XOR | B_ORNOT                                  (* bitwise *) 
  (* IbinaryOp *) | L_AND | L_EQ  | L_GEQ | L_GT  | L_LEQ | L_LT  | L_NEQ  | L_OR    (* logical *) 
with override x.ToString() = match x with A_ADD -> "+"  | A_SUB -> "-"
                                        | A_MUL -> "*"  | A_DIV -> "/"  | A_REM -> "%"
                                        | A_SHL -> "<<" | A_SHR -> ">>"
                                        | B_IOR -> "|"  | B_AND -> "&"  | B_XOR -> "^"  | B_ORNOT -> "!"
                                        | L_LT  -> "<"  | L_GT  -> ">"  | L_LEQ -> "<=" | L_GEQ   -> ">="
                                        | L_AND -> "&&" | L_OR  -> "||"
                                        | L_EQ  -> "==" | L_NEQ -> "!="

type Id = { [<Id>] Name : string }

(* integer expression *)
type Iexpr =  | Unary  of IunaryOp * Iexpr
  (* Iexpr *) | Binary of IbinaryOp * Iexpr * Iexpr
  (* Iexpr *) | Id of Id (* TOO MANY ID's !!! *)
  (* Iexpr *) | Int of bigint
with static member Zero     = Int(0I)
     static member Int32(i) = Int(bigint(i:Int32))
     static member Int64(i) = Int(bigint(i:Int64))
     static member Named(s) = Id({Name=s})

     member x.FoldIds(acc:String list) = match x with Unary(_,e) -> e.FoldIds acc
                                                    | Binary(_,e1,e2) -> e2.FoldIds (e1.FoldIds acc)
                                                    | Id(i) -> i.Name::acc
                                                    | Int(_) -> acc
     member x.Ids = x.FoldIds []

     override x.ToString() = match x with Unary(op,e) -> sprintf "(%s %s)" (op.ToString()) (e.ToString())
                                        | Binary(op,e1,e2) -> sprintf "(%s %s %s)" (e1.ToString()) (op.ToString()) (e2.ToString())
                                        | Id(s) -> s.Name
                                        | Int(i) -> NXyn.Aux.ConvertMisc.base10bigintconv.Unparse(i)

     member x.CompactString = match x with Unary(op,e) -> sprintf "%s%s" (op.ToString()) e.CompactString
                                         | Binary(op,e1,e2) -> sprintf "%s%s%s" e1.CompactString (op.ToString()) e2.CompactString
                                         | Id(s) -> s.Name
                                         | Int(i) -> NXyn.Aux.ConvertMisc.base10bigintconv.Unparse(i)

     (* add [MiniSSA98] optimizations? *)
     member x.Simplify = match x with Unary(A_NEG, Int(i)) -> Int(-i)
                                    | Unary(op,z) -> Unary(op,z.Simplify)
                                    | Binary(op,z1,z2) -> Binary(op,z1.Simplify,z2.Simplify)
                                    | Id _ | Int _ -> x

(* TBD rename? *)
let (|Zero|_|) x = match x with Unary _ | Binary _ | Id _ -> None
                              | Int(i) -> if i = 0I then Some() else None

let (|IexprNamed|_|) = function Iexpr.Id({Id.Name=name}) -> Some(name) | _ -> None

let (|IexprInt32|_|) = function Iexpr.Int(v) -> try Some(int32(v)) with _ -> None
                              | _ -> None
let (|IexprInt64|_|) = function Iexpr.Int(v) -> try Some(int64(v)) with _ -> None
                              | _ -> None

(***)

type EAscaledIndex<'r> = { [<S(1)>] Index : 'r; Scale : Int64 }
type EAbase<'r>        = { [<S(2)>] Base  : 'r }

(* memory access *)
type M    = K      of Iexpr
  (* M *) | K32eip of Iexpr
  (* M *) | K64rip of Iexpr
  (* M *) | EA32   of Iexpr * EAbase<R32> option * EAscaledIndex<R32> option
  (* M *) | EA64   of Iexpr * EAbase<R64> option * EAscaledIndex<R64> option
with static member D     (                          ofs) = EA64(Iexpr.Int64(ofs), None,                   None)
     static member B_D   (baseReg,                  ofs) = EA64(Iexpr.Int64(ofs), Some({Base = baseReg}), None)
     static member B     (baseReg                      ) = EA64(Iexpr.Zero,       Some({Base = baseReg}), None)
     static member B_SI_D(baseReg, indexReg, scale, ofs) = EA64(Iexpr.Int64(ofs), Some({Base = baseReg}), Some({Index = indexReg; Scale = scale}))
     static member B_SI  (baseReg, indexReg, scale     ) = EA64(Iexpr.Zero,       Some({Base = baseReg}), Some({Index = indexReg; Scale = scale}))
     static member S     (         indexReg            ) = EA64(Iexpr.Zero,       None,                   Some({Index = indexReg; Scale = 1L}))
     static member SI    (         indexReg, scale     ) = EA64(Iexpr.Zero,       None,                   Some({Index = indexReg; Scale = scale}))
     static member SI_D  (         indexReg, scale, ofs) = EA64(Iexpr.Int64(ofs), None,                   Some({Index = indexReg; Scale = scale}))

     member x.Ids = match x with EA32(ofs,_,_) | EA64(ofs,_,_) | K(ofs) | K64rip(ofs) | K32eip(ofs) -> ofs.Ids

     override x.ToString() = match x with K(k)                                                 -> k.ToString()
                                        | K32eip(k)                                            -> k.ToString() + "(%eip)"
                                        | K64rip(k)                                            -> k.ToString() + "(%rip)"
                                        | EA32(ofs,  None,           None)                     -> string ofs
                                        | EA32(Zero, Some {Base=r},  None)                     -> sprintf "(%s)" (string r)
                                        | EA32(ofs,  Some {Base=r},  None)                     -> sprintf "%s(%s)" (string ofs) (string r)
                                        | EA32(Zero, None,           Some{Index=r;  Scale=i})  -> sprintf "(,%s,%d)" (string r) i
                                        | EA32(ofs,  None,           Some{Index=r;  Scale=i})  -> sprintf "%s(,%s,%d)" (string ofs) (string r) i
                                        | EA32(Zero, Some {Base=r1}, Some{Index=r2; Scale=1L}) -> sprintf "(%s,%s)" (string r1) (string r2)
                                        | EA32(Zero, Some {Base=r1}, Some{Index=r2; Scale=i})  -> sprintf "(%s,%s,%d)" (string r1) (string r2) i
                                        | EA32(ofs,  Some {Base=r1}, Some{Index=r2; Scale=1L}) -> sprintf "%s(%s,%s)" (string ofs) (string r1) (string r2)
                                        | EA32(ofs,  Some {Base=r1}, Some{Index=r2; Scale=i})  -> sprintf "%s(%s,%s,%d)" (string ofs) (string r1) (string r2) i
                                        | EA64(ofs,  None,           None)                     -> string ofs
                                        | EA64(Zero, Some {Base=r},  None)                     -> sprintf "(%s)" (string r)
                                        | EA64(ofs,  Some {Base=r},  None)                     -> sprintf "%s(%s)" (string ofs) (string r)
                                        | EA64(Zero, None,           Some{Index=r;  Scale=i})  -> sprintf "(,%s,%d)" (string r) i
                                        | EA64(ofs,  None,           Some{Index=r;  Scale=i})  -> sprintf "%s(,%s,%d)" (string ofs) (string r) i
                                        | EA64(Zero, Some {Base=r1}, Some{Index=r2; Scale=1L}) -> sprintf "(%s,%s)" (string r1) (string r2)
                                        | EA64(Zero, Some {Base=r1}, Some{Index=r2; Scale=i})  -> sprintf "(%s,%s,%d)" (string r1) (string r2) i
                                        | EA64(ofs,  Some {Base=r1}, Some{Index=r2; Scale=1L}) -> sprintf "%s(%s,%s)" (string ofs) (string r1) (string r2)
                                        | EA64(ofs,  Some {Base=r1}, Some{Index=r2; Scale=i})  -> sprintf "%s(%s,%s,%d)" (string ofs) (string r1) (string r2) i

(***)

type M8   = S of M with member x.Ids = match x with S(x) -> x.Ids
type M16  = S of M with member x.Ids = match x with S(x) -> x.Ids
type M32  = S of M with member x.Ids = match x with S(x) -> x.Ids
type M64  = S of M with member x.Ids = match x with S(x) -> x.Ids
type M128 = S of M with member x.Ids = match x with S(x) -> x.Ids
type M256 = S of M with member x.Ids = match x with S(x) -> x.Ids

(***)

type NM64 = { S : M64 }
type NR64 = { S : R64 }

(***)

type Iargs_D<'d>                   = { [<AttPos(1);D(1)>]  D   : 'd  }
type Iargs_DS<'ds>                 = { [<AttPos(1);DS(1)>] DS  : 'ds }
type Iargs_DS_DS<'ds1,'ds2>        = { [<AttPos(1);DS(1)>] DS1 : 'ds1; [<AttPos(2);DS(2)>] DS2 : 'ds2 }
type Iargs_S<'s>                   = { [<AttPos(1);S(1)>]  S   : 's  }
type Iargs_S_D<'s,'d>              = { [<AttPos(1);S(1)>]  S   : 's;   [<AttPos(2);D(1)>]  D   : 'd   }
type Iargs_S_DS<'s,'ds>            = { [<AttPos(1);S(1)>]  S   : 's;   [<AttPos(2);DS(1)>] DS  : 'ds  }
type Iargs_S_S<'s1,'s2>            = { [<AttPos(1);S(1)>]  S1  : 's1;  [<AttPos(2);S(2)>]  S2  : 's2  }
type Iargs_S_S_D<'s1,'s2,'d>       = { [<AttPos(1);S(1)>]  S1  : 's1;  [<AttPos(2);S(2)>]  S2  : 's2; [<AttPos(3);D(1)>]  D  : 'd  }
type Iargs_S_S_DS<'s1,'s2,'ds>     = { [<AttPos(1);S(1)>]  S1  : 's1;  [<AttPos(2);S(2)>]  S2  : 's2; [<AttPos(3);DS(1)>] DS : 'ds }
type Iargs_S_S_S_D<'s1,'s2,'s3,'d> = { [<AttPos(1);S(1)>]  S1  : 's1;  [<AttPos(2);S(2)>]  S2  : 's2; [<AttPos(3);S(3)>]  S3 : 's3; [<AttPos(4);D(1)>] D : 'd }

(***)

type Lit = { K : Iexpr }
with static member Int32(i) = { Lit.K = Iexpr.Int32(i) }
     static member Int64(i) = { Lit.K = Iexpr.Int64(i) }

let (|LitInt32|_|) = function {Lit.K=IexprInt32(v)} -> Some(v) | _ -> None
let (|LitInt64|_|) = function {Lit.K=IexprInt64(v)} -> Some(v) | _ -> None

(*****************************************************************************************************************************************************************************)

type I_C  = | [< AKA "call";                   Mem;  Jthru;               Rds "rsp"              >]  Callq     (* call procedure *)
  (* I_C *) | [<                               Mem;  Jfollow                                     >]  Jmp       (* unconditional jump *)
  (* I_C *) | [< AKA "jnbe";                         Jthru;                           Fs "Z,C"   >]  Ja        (* jump if above *)
  (* I_C *) | [< AKA "jb,jnae";                      Jthru;                           Fs "C"     >]  Jc        (* jump if carry *)
  (* I_C *) | [<                                     Jthru;    Rs "ecx"                          >]  Jecxz     (* jump if ecx is zero *)
  (* I_C *) | [< AKA "jnle";                         Jthru;                           Fs "Z,S,O" >]  Jg        (* jump if greater *)
  (* I_C *) | [< AKA "jnge";                         Jthru;                           Fs "S,O"   >]  Jl        (* jump if less *)
  (* I_C *) | [< AKA "jbe";                          Jthru;                           Fs "Z,C"   >]  Jna       (* jump if not above *)
  (* I_C *) | [< AKA "jnb,jae";                      Jthru;                           Fs "C"     >]  Jnc       (* jump if not carry *)
  (* I_C *) | [< AKA "jle";                          Jthru;                           Fs "Z,S,O" >]  Jng       (* jump if not greater *)
  (* I_C *) | [< AKA "jge";                          Jthru;                           Fs "S,O"   >]  Jnl       (* jump if not less *)
  (* I_C *) | [<                                     Jthru;                           Fs "O"     >]  Jno       (* jump if not overflow *)
  (* I_C *) | [< AKA "jpo";                          Jthru;                           Fs "P"     >]  Jnp       (* jump if parity odd *)
  (* I_C *) | [<                                     Jthru;                           Fs "S"     >]  Jns       (* jump if not sign *)
  (* I_C *) | [< AKA "jne";                          Jthru;                           Fs "Z"     >]  Jnz       (* jump if not zero *)
  (* I_C *) | [<                                     Jthru;                           Fs "O"     >]  Jo        (* jump if overflow *)
  (* I_C *) | [< AKA "jpe";                          Jthru;                           Fs "P"     >]  Jp        (* jump if parity even *)
  (* I_C *) | [<                                     Jthru;    Rs "rcx"                          >]  Jrcxz     (* jump if rcx is zero *)
  (* I_C *) | [<                                     Jthru;                           Fs "S"     >]  Js        (* jump if sign *)
  (* I_C *) | [< AKA "je";                           Jthru;                           Fs "Z"     >]  Jz        (* jump if zero *)
  (* I_C *) | [< AKA "loop";                         Jthru;               Rds "rcx"              >]  Loopq     (* loop according to rcx *)
  (* I_C *) | [< AKA "loope,loopz,loopzq";           Jthru;               Rds "rcx";  Fs "Z"     >]  Loopeq    (* loop according to rcx *)
  (* I_C *) | [< AKA "loopzl";                       Jthru;               Rds "ecx";  Fs "Z"     >]  Loopel    (* loop according to ecx *)
  (* I_C *) | [<                                     Jthru;               Rds "ecx"              >]  Loopl     (* loop according to ecx *)
  (* I_C *) | [< AKA "loopne,loopnz,loopnzq";        Jthru;               Rds "rcx";  Fs "Z"     >]  Loopne    (* loop according to rcx *)
  (* I_C *) | [< AKA "loopnzl";                      Jthru;               Rds "ecx";  Fs "Z"     >]  Loopnel   (* loop according to ecx *)

(*****************************************************************************************************************************************************************************)

type I_V0 =  | [< ISA "AVX";  Yd  "ymm0,ymm1,ymm2,ymm3,ymm4,ymm5,ymm6,ymm7,ymm8,ymm9,ymm10,ymm11,ymm12,ymm13,ymm14,ymm15"  >]  Vzeroall      (* clear ymm regs *)
  (* I_V0 *) | [< ISA "AVX";  Yds "ymm0,ymm1,ymm2,ymm3,ymm4,ymm5,ymm6,ymm7,ymm8,ymm9,ymm10,ymm11,ymm12,ymm13,ymm14,ymm15"  >]  Vzeroupper    (* clear upper halves of ymm regs *)

(*****************************************************************************************************************************************************************************)

type I_I0 =  | [< AKA "cdq";                          Rs "rax";                  Rd "rdx"                                                           >]  Cltd
  (* I_I0 *) | [< AKA "cdqe";                                    Rds "rax"                                                                          >]  Cltq
  (* I_I0 *) | [< AKA "cbtw";                                    Rds "rax"                                                                          >]  Cbw
  (* I_I0 *) | [< AKA "cmpsd";  Rep;  Mem;                       Rds "rsi,rdi";                     Fs "D";                  Fd "O,S,Z,A,P,C"       >]  Cmpsl
  (* I_I0 *) | [<               Rep;  Mem;                       Rds "rsi,rdi";                     Fs "D";                  Fd "O,S,Z,A,P,C"       >]  Cmpsb
  (* I_I0 *) | [<               Rep;  Mem;                       Rds "rsi,rdi";                     Fs "D";                  Fd "O,S,Z,A,P,C"       >]  Cmpsq
  (* I_I0 *) | [<               Rep;  Mem;                       Rds "rsi,rdi";                     Fs "D";                  Fd "O,S,Z,A,P,C"       >]  Cmpsw
  (* I_I0 *) | [<                                                                                                            Fd "C"                 >]  Clc
  (* I_I0 *) | [<                                                                                                            Fd "D"                 >]  Cld
  (* I_I0 *) | [<                                                                                                            Fd "I"                 >]  Cli
  (* I_I0 *) | [<                                                                                   Fs "C";                  Fd "C"                 >]  Cmc
  (* I_I0 *) | [<                                                Rds "rax,rcx";  Rd "rbx,rdx"                                                       >]  Cpuid
  (* I_I0 *) | [< AKA "cqo";                          Rs "rax";                  Rd "rdx"                                                           >]  Cqto
  (* I_I0 *) | [< AKA "cwd";                          Rs "rax";                  Rd "rdx"                                                           >]  Cwtd
  (* I_I0 *) | [< AKA "cwde";                                    Rds "rax"                                                                          >]  Cwtl
  (* I_I0 *) | [< AKA "enter";        Mem;                       Rds "rbp,rsp"                                                                      >]  Enterq of Lit * Lit
  (* I_I0 *) | [<                     Mem;  Jthru;                                                                           Fd "I,T"               >]  Int of Lit
  (* I_I0 *) | [<                     Mem;  Jthru;               Rds "rsp";                                                  Fd "I,T"               >]  Int3
  (* I_I0 *) | [< AKA "leave";        Mem;                       Rds "rsp,rbp"                                                                      >]  Leaveq
  (* I_I0 *) | [<                     Mem                                                                                                           >]  Lfence
  (* I_I0 *) | [<               Pfx                                                                                                                 >]  Lock
  (* I_I0 *) | [< AKA "lodsd";  Rep;  Mem;            Rs "eax";  Rds "rsi";                         Fs "D"                                          >]  Lodsl
  (* I_I0 *) | [<               Rep;  Mem;            Rs "al";   Rds "rsi";                         Fs "D"                                          >]  Lodsb
  (* I_I0 *) | [<               Rep;  Mem;            Rs "ax";   Rds "rsi";                         Fs "D"                                          >]  Lodsq
  (* I_I0 *) | [<               Rep;  Mem;            Rs "rax";  Rds "rsi";                         Fs "D"                                          >]  Lodsw
  (* I_I0 *) | [<                     Mem                                                                                                           >]  Mfence
  (* I_I0 *) | [< AKA "movsd";  Rep;  Mem;                       Rds "rdi,rsi";                     Fs "D"                                          >]  Movsl
  (* I_I0 *) | [<               Rep;  Mem;                       Rds "rdi,rsi";                     Fs "D"                                          >]  Movsb
  (* I_I0 *) | [<               Rep;  Mem;                       Rds "rdi,rsi";                     Fs "D"                                          >]  Movsq
  (* I_I0 *) | [<               Rep;  Mem;                       Rds "rdi,rsi";                     Fs "D"                                          >]  Movsw
  (* I_I0 *) | [<               Nop                                                                                                                 >]  Nop 
  (* I_I0 *) |                                                                                                                                          Pause
  (* I_I0 *) | [< AKA "popf";         Mem;                       Rds "rsp";                                                  Fd "O,D,I,T,S,Z,A,P,C" >]  Popfq
  (* I_I0 *) | [< AKA "pushf";        Mem;                       Rds "rsp";                         Fs "O,D,I,T,S,Z,A,P,C"                          >]  Pushfq
  (* I_I0 *) | [< AKA "push";         Mem;                       Rds "rsp"                                                                          >]  Pushq of Lit
  (* I_I0 *) | [<                                     Rs "rcx";                  Rd "rax,rdx"                                                       >]  Rdpmc
  (* I_I0 *) | [<                                                                Rd "rax,rdx"                                                       >]  Rdtsc
  (* I_I0 *) | [<                                                                Rd "rax,rcx,rdx"                                                   >]  Rdtscp
  (* I_I0 *) | [<               Pfx;                             Rds "rcx"                                                                          >]  Rep
  (* I_I0 *) | [< AKA "repz";   Pfx;                             Rds "rcx";                         Fs "Z"                                          >]  Repe
  (* I_I0 *) | [< AKA "repnz";  Pfx;                             Rds "rcx";                         Fs "Z"                                          >]  Repne
  (* I_I0 *) | [< AKA "ret";    Ret;  Mem;  Jfollow;             Rds "rsp"                                                                          >]  Retq
  (* I_I0 *) | [< AKA "retq";   Ret;  Mem;  Jfollow;             Rds "rsp"                                                                          >]  Ret of Lit
  (* I_I0 *) | [< AKA "scasd";  Rep;  Mem;            Rs "eax";  Rds "rdi";                         Fs "D";                  Fd "O,S,Z,A,P,C"       >]  Scasl
  (* I_I0 *) | [<               Rep;  Mem;            Rs "al";   Rds "rdi";                         Fs "D";                  Fd "O,S,Z,A,P,C"       >]  Scasb
  (* I_I0 *) | [<               Rep;  Mem;            Rs "ax";   Rds "rdi";                         Fs "D";                  Fd "O,S,Z,A,P,C"       >]  Scasq
  (* I_I0 *) | [<               Rep;  Mem;            Rs "rax";  Rds "rdi";                         Fs "D";                  Fd "O,S,Z,A,P,C"       >]  Scasw
  (* I_I0 *) | [<                     Mem                                                                                                           >]  Sfence
  (* I_I0 *) | [<                                                                                                            Fd "C"                 >]  Stc
  (* I_I0 *) | [<                                                                                                            Fd "D"                 >]  Std
  (* I_I0 *) | [<                                                                                                            Fd "I"                 >]  Sti
  (* I_I0 *) | [< AKA "stosd";  Rep;  Mem;            Rs "eax";  Rds "rdi";                         Fs "D"                                          >]  Stosl
  (* I_I0 *) | [<               Rep;  Mem;            Rs "al";   Rds "rdi";                         Fs "D"                                          >]  Stosb
  (* I_I0 *) | [<               Rep;  Mem;            Rs "ax";   Rds "rdi";                         Fs "D"                                          >]  Stosq
  (* I_I0 *) | [<               Rep;  Mem;            Rs "rax";  Rds "rdi";                         Fs "D"                                          >]  Stosw
  (* I_I0 *) | [<               Exn                                                                                                                 >]  Ud2
  (* I_I0 *) | [< AKA "xlat";         Mem;            Rs "rbx";  Rds "al"                                                                           >]  Xlatb
  (* I_I0 *) | [< AKA "fwait"                                                                                                                       >]  Wait
  (* I_I0 *) | [<                                     Rs "ah";                                      Fs "S,Z,A,P,C"                                  >]  Sahf
  (* I_I0 *) | [<                                                                Rd "ah";           Fs "S,Z,A,P,C"                                  >]  Lahf

(*****************************************************************************************************************************************************************************)

type I_NR64s =  | [< AKA "call";  Mem;  Jthru;    Rds "rsp" >]  Callq    (* indirect call subroutine; TBD state + flags *)
  (* I_NR64s *) | [< AKA "jmp";         Jfollow             >]  Jmpq     (* indirect branch; TBD state + flags *)

type I_NM64s =  | [< AKA "call";  Mem;  Jthru;    Rds "rsp" >]  Callq    (* indirect call subroutine; TBD state + flags *)
  (* I_NM64s *) | [< AKA "jmp";         Jfollow             >]  Jmpq     (* indirect branch; TBD state + flags *)

(*****************************************************************************************************************************************************************************)

type I_R8s =   | [< AKA "cmp";                                     Fd "O,S,Z,A,P,C" >]  Cmpb of Lit     (* arithmetic compare byte *)
  (* I_R8s *)  | [< AKA "div";    Exn;  Rds "ax";                  Fd "O,S,Z,A,P,C" >]  Divb            (* unsigned divide byte *)
  (* I_R8s *)  | [< AKA "idiv";   Exn;  Rds "ax";                  Fd "O,S,Z,A,P,C" >]  Idivb           (* signed divide byte *)
  (* I_R8s *)  | [< AKA "imul";         Rds "ax";                  Fd "O,S,Z,A,P,C" >]  Imulb           (* signed multiply byte *)
  (* I_R8s *)  | [< AKA "mul";          Rds "ax";                  Fd "O,S,Z,A,P,C" >]  Mulb            (* unsigned multiply byte *)
  (* I_R8s *)  | [< AKA "test";                                    Fd "O,S,Z,A,P,C" >]  Testb of Lit    (* logical compare byte *)
type I_R16s =  | [< AKA "bt";                                      Fd "C"           >]  Btw of Lit      (* test bit in word *)
  (* I_R16s *) | [< AKA "cmp";                                     Fd "O,S,Z,A,P,C" >]  Cmpw of Lit     (* arithmetic compare word *)
  (* I_R16s *) | [< AKA "div";    Exn;  Rds "ax,dx";               Fd "O,S,Z,A,P,C" >]  Divw            (* unsigned divide word *)
  (* I_R16s *) | [< AKA "idiv";   Exn;  Rds "ax,dx";               Fd "O,S,Z,A,P,C" >]  Idivw           (* signed divide word *)
  (* I_R16s *) | [< AKA "imul";         Rds "ax";       Rd "dx";   Fd "O,S,Z,A,P,C" >]  Imulw           (* signed multiply word *)
  (* I_R16s *) | [< AKA "mul";          Rds "ax";       Rd "dx";   Fd "O,S,Z,A,P,C" >]  Mulw            (* unsigned multiply word *)
  (* I_R16s *) | [< AKA "push";   Mem;  Rds "rsp"                                   >]  Pushw           (* push word onto stack *)
  (* I_R16s *) | [< AKA "test";                                    Fd "O,S,Z,A,P,C" >]  Testw of Lit    (* logical compare word *)
  (* I_R16s *) | [< AKA "nop";    Nop                                               >]  Nopw            (* no operation *)
type I_R32s =  | [< AKA "bt";                                      Fd "C";          >]  Btl of Lit      (* test bit in long *)
  (* I_R32s *) | [< AKA "cmp";                                     Fd "O,S,Z,A,P,C" >]  Cmpl of Lit     (* arithmetic compare long *)
  (* I_R32s *) | [< AKA "div";    Exn;  Rds "eax,edx";             Fd "O,S,Z,A,P,C" >]  Divl            (* unsigned divide long *)
  (* I_R32s *) | [< AKA "idiv";   Exn;  Rds "eax,edx";             Fd "O,S,Z,A,P,C" >]  Idivl           (* signed divide long *)
  (* I_R32s *) | [< AKA "imul";         Rds "eax";      Rd "edx";  Fd "O,S,Z,A,P,C" >]  Imull           (* signed multiply long *)
  (* I_R32s *) | [< AKA "mul";          Rds "eax";      Rd "edx";  Fd "O,S,Z,A,P,C" >]  Mull            (* unsigned multiply long *)
  (* I_R32s *) | [< AKA "test";                                    Fd "O,S,Z,A,P,C" >]  Testl of Lit    (* logical compare long *) 
  (* I_R32s *) | [< AKA "nop";    Nop                                               >]  Nopl            (* no operation *)
type I_R64s =  | [< AKA "bt";                                      Fd "C";          >]  Btq of Lit      (* test bit in quad *)
  (* I_R64s *) | [< AKA "cmp";                                     Fd "O,S,Z,A,P,C" >]  Cmpq of Lit     (* arithmetic compare quad *)
  (* I_R64s *) | [< AKA "div";    Exn;  Rds "rax,rdx";             Fd "O,S,Z,A,P,C" >]  Divq            (* unsigned divide quad *)
  (* I_R64s *) | [< AKA "idiv";   Exn;  Rds "rax,rdx";             Fd "O,S,Z,A,P,C" >]  Idivq           (* signed divide quad *)
  (* I_R64s *) | [< AKA "imul";         Rds "rax";      Rd "rdx";  Fd "O,S,Z,A,P,C" >]  Imulq           (* signed multiply quad *)
  (* I_R64s *) | [< AKA "mul";          Rds "rax";      Rd "rdx";  Fd "O,S,Z,A,P,C" >]  Mulq            (* unsigned multiply quad  *)
  (* I_R64s *) | [< AKA "push";   Mem;  Rds "rsp"                                   >]  Pushq           (* push quad onto stack *)
  (* I_R64s *) | [< AKA "test";                                    Fd "O,S,Z,A,P,C" >]  Testq of Lit    (* logical compare quad *)
  (* I_R64s *) | [< AKA "nop";    Nop                                               >]  Nopq            (* no operation *)

(*****************************************************************************************************************************************************************************)

type I_M8s  =  |                                                                                     Clflush         (* flush cache line *)
  (* I_M8s *)  | [<                                                             Fd "O,S,Z,A,P,C" >]  Cmpb of Lit     (* arithmetic compare bytes *)
  (* I_M8s *)  | [<                            Exn;  Rds "ax";                  Fd "O,S,Z,A,P,C" >]  Divb            (* unsigned divide bytes *)
  (* I_M8s *)  | [<                            Exn;  Rds "ax";                  Fd "O,S,Z,A,P,C" >]  Idivb           (* signed divide bytes *)
  (* I_M8s *)  | [<                                  Rds "ax";                  Fd "O,S,Z,A,P,C" >]  Imulb           (* signed multiply bytes *)
  (* I_M8s *)  | [<                                  Rds "ax";                  Fd "O,S,Z,A,P,C" >]  Mulb            (* unsigned multiply bytes *)
  (* I_M8s *)  | [< ISA "AMD_PF"                                                                 >]  Prefetch        (* prefetch to L1 cache *)
  (* I_M8s *)  |                                                                                     Prefetchnta     (* prefetch into caches w/NTA-hint *)
  (* I_M8s *)  |                                                                                     Prefetcht0      (* prefetch into caches w/T0-hint *)
  (* I_M8s *)  |                                                                                     Prefetcht1      (* prefetch into caches w/T1-hint *)
  (* I_M8s *)  |                                                                                     Prefetcht2      (* prefetch into caches w/T2-hint *)
  (* I_M8s *)  | [< ISA "AMD_PF"                                                                 >]  Prefetchw       (* prefetch into L1 w/write-hint *)
  (* I_M8s *)  | [<                                                             Fd "O,S,Z,A,P,C" >]  Testb of Lit    (* logical compare byte *)
type I_M16s =  | [<                            Nop                                               >]  Nopw            (* no operation *)
  (* I_M16s *) | [<                                                             Fd "C"           >]  Btw of Lit      (* test bit in word *)
  (* I_M16s *) | [<                                                             Fd "O,S,Z,A,P,C" >]  Cmpw of Lit     (* arithmetic compare word *)
  (* I_M16s *) | [<                            Exn;  Rds "ax,dx";               Fd "O,S,Z,A,P,C" >]  Divw            (* unsigned divide word *)
  (* I_M16s *) | [<                            Exn;  Rds "ax,dx";               Fd "O,S,Z,A,P,C" >]  Idivw           (* signed divide word *)
  (* I_M16s *) | [<                                  Rds "ax";       Rd "dx";   Fd "O,S,Z,A,P,C" >]  Imulw           (* signed multiply word *)
  (* I_M16s *) | [<                                  Rds "ax";       Rd "dx";   Fd "O,S,Z,A,P,C" >]  Mulw            (* unsigned multiply word *)
  (* I_M16s *) | [<                            Mem;  Rds "rsp"                                   >]  Pushw           (* push word onto the stack *)
  (* I_M16s *) | [<                                                             Fd "O,S,Z,A,P,C" >]  Testw of Lit    (* logical compare word *)
type I_M32s =  | [<                AKA "nop";  Nop                                               >]  Nopl            (* no operation *)
  (* I_M32s *) | [<                                                             Fd "C"           >]  Btl of Lit      (* test bit in long *)
  (* I_M32s *) | [<                                                             Fd "O,S,Z,A,P,C" >]  Cmpl of Lit     (* arithmetic compare long *)
  (* I_M32s *) | [<                            Exn;  Rds "eax,edx";             Fd "O,S,Z,A,P,C" >]  Divl            (* unsigned divide word *)
  (* I_M32s *) | [<                            Exn;  Rds "eax,edx";             Fd "O,S,Z,A,P,C" >]  Idivl           (* signed divide word *)
  (* I_M32s *) | [<                                  Rds "eax";      Rd "edx";  Fd "O,S,Z,A,P,C" >]  Imull           (* signed multiply long *)
  (* I_M32s *) | [<                            Mem                                               >]  Ldmxcsr         (* load MXCSR register *)
  (* I_M32s *) | [< ISA "AVX";                 Mem                                               >]  Vldmxcsr        (* load MXCSR register *)
  (* I_M32s *) | [<                                  Rds "eax";      Rd "edx";  Fd "O,S,Z,A,P,C" >]  Mull            (* unsigned multiply long *)
  (* I_M32s *) | [<                                                             Fd "O,S,Z,A,P,C" >]  Testl of Lit    (* logical compare long *)
type I_M64s =  | [<                            Nop                                               >]  Nopq            (* no operation *)
  (* I_M64s *) | [<                                                             Fd "C"           >]  Btq of Lit      (* test bit in quad *)
  (* I_M64s *) | [<                                                             Fd "O,S,Z,A,P,C" >]  Cmpq of Lit     (* arithmetic compare quad *)
  (* I_M64s *) | [<                            Exn;  Rds "rax,rdx";             Fd "O,S,Z,A,P,C" >]  Divq            (* unsigned divide quad *)
  (* I_M64s *) | [<                            Exn;  Rds "rax,rdx";             Fd "O,S,Z,A,P,C" >]  Idivq           (* signed divide quad *)
  (* I_M64s *) | [<                                  Rds "rax";      Rd "rdx";  Fd "O,S,Z,A,P,C" >]  Imulq           (* signed multiply quad *)
  (* I_M64s *) | [<                                  Rds "rax";      Rd "rdx";  Fd "O,S,Z,A,P,C" >]  Mulq            (* unsigned multiply quad *)
  (* I_M64s *) | [<                            Mem;  Rds "rsp"                                   >]  Pushq           (* push quad onto the stack *)
  (* I_M64s *) | [<                                                             Fd "O,S,Z,A,P,C" >]  Testq of Lit    (* logical compare quad *)

(*****************************************************************************************************************************************************************************)

type I_R8d =   | [< AKA "mov"                                                            >]  Movb of Lit       (* move immediate byte *)
  (* I_R8d *)  | [< AKA "seta,setnbe,setnbeb";                                Fs "Z,C"   >]  Setab             (* set byte if above *)
  (* I_R8d *)  | [< AKA "setc,setb,setnae,setbb,setnaeb";                     Fs "C"     >]  Setcb             (* set byte if carry *)
  (* I_R8d *)  | [< AKA "setg,setnle,setnleb";                                Fs "Z,S,O" >]  Setgb             (* set byte if greater *)
  (* I_R8d *)  | [< AKA "setl,setnge,setngeb";                                Fs "S,O"   >]  Setlb             (* set byte if less *)
  (* I_R8d *)  | [< AKA "setna,setbe,setbeb";                                 Fs "Z,C"   >]  Setnab            (* set byte if not above *)
  (* I_R8d *)  | [< AKA "setnc,setnb,setae,setnbb,setaeb";                    Fs "C"     >]  Setncb            (* set byte if not carry *)
  (* I_R8d *)  | [< AKA "setng,setle,setleb";                                 Fs "Z,S,O" >]  Setngb            (* set byte if not greater *)
  (* I_R8d *)  | [< AKA "setnl,setge,setgeb";                                 Fs "S,O"   >]  Setnlb            (* set byte if not less *)
  (* I_R8d *)  | [< AKA "setnp,setpo,setpob";                                 Fs "P"     >]  Setnpb            (* set byte if no parity *)
  (* I_R8d *)  | [< AKA "setp,setpe,setpeb";                                  Fs "P"     >]  Setpb             (* set byte if parity *)
  (* I_R8d *)  | [< AKA "setz,sete,seteb";                                    Fs "Z"     >]  Setzb             (* set byte if zero *)
  (* I_R8d *)  | [< AKA "setnz,setne,setneb";                                 Fs "Z"     >]  Setnzb            (* set byte if not zero *)
  (* I_R8d *)  | [< AKA "setno";                                              Fs "O"     >]  Setnob            (* set byte if not overflow *)
  (* I_R8d *)  | [< AKA "setns";                                              Fs "S"     >]  Setnsb            (* set byte if not sign *)
  (* I_R8d *)  | [< AKA "seto";                                               Fs "O"     >]  Setob             (* set byte if overflow *)
  (* I_R8d *)  | [< AKA "sets";                                               Fs "S"     >]  Setsb             (* set byte if sign *)
type I_R16d =  | [< AKA "mov"                                                            >]  Movw of Lit       (* move immediate word *)
  (* I_R16d *) | [< AKA "pop";                              Mem;  Rds "rsp"              >]  Popw              (* pop word from stack *)
type I_R32d =  | [< AKA "lea"                                                            >]  Leal of M         (* load effective address *)
  (* I_R32d *) | [< AKA "mov"                                                            >]  Movl of Lit       (* move immediate long *)
type I_R64d =  | [< AKA "lea"                                                            >]  Leaq of M         (* load effective address *)
  (* I_R64d *) | [< AKA "mov"                                                            >]  Movq of Lit       (* move immediate quad *)
  (* I_R64d *) | [< AKA "movabs"                                                         >]  Movabsq of Lit    (* move 64-bit immediate *)
  (* I_R64d *) | [< AKA "pop";                              Mem;  Rds "rsp"              >]  Popq              (* pop quad from stack *)

(*****************************************************************************************************************************************************************************)

type I_R8ds =   | [< AKA "adc";           Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcb of Lit    (* add byte w/carry *)
  (* I_R8ds *)  | [< AKA "add";                    Fd "O,S,Z,A,P,C" >]  Addb of Lit    (* add byte *)
  (* I_R8ds *)  | [< AKA "and";                    Fd "O,S,Z,A,P,C" >]  Andb of Lit    (* logical AND byte *)
  (* I_R8ds *)  | [< AKA "dec";                    Fd "O,S,Z,A,P"   >]  Decb           (* decrement unsigned byte *)
  (* I_R8ds *)  | [< AKA "inc";                    Fd "O,S,Z,A,P"   >]  Incb           (* increment unsigned byte *)
  (* I_R8ds *)  | [< AKA "neg";                    Fd "O,S,Z,A,P,C" >]  Negb           (* two's complement negate byte *)
  (* I_R8ds *)  | [< AKA "not"                                      >]  Notb           (* one's complement negate byte *)
  (* I_R8ds *)  | [< AKA "or";                     Fd "O,S,Z,A,P,C" >]  Orb of Lit     (* logical inclusive-OR byte *)
  (* I_R8ds *)  | [< AKA "rcl";           Fs "C";  Fd "O,C"         >]  Rclb of Lit    (* rotate byte right thru carry *)
  (* I_R8ds *)  | [< AKA "rcr";           Fs "C";  Fd "O,C"         >]  Rcrb of Lit    (* rotate byte right thru carry *)
  (* I_R8ds *)  | [< AKA "rol";                    Fd "O,C"         >]  Rolb of Lit    (* rotate byte left *)
  (* I_R8ds *)  | [< AKA "ror";                    Fd "O,C"         >]  Rorb of Lit    (* rotate byte right *)
  (* I_R8ds *)  | [< AKA "sar";                    Fd "O,S,Z,A,P,C" >]  Sarb of Lit    (* shift byte right arithmetic *)
  (* I_R8ds *)  | [< AKA "sbb";           Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbb of Lit    (* subtract byte w/borrow *)
  (* I_R8ds *)  | [< AKA "shl,sal,salb";           Fd "O,S,Z,A,P,C" >]  Shlb of Lit    (* shift byte left logical *)
  (* I_R8ds *)  | [< AKA "shr";                    Fd "O,S,Z,A,P,C" >]  Shrb of Lit    (* shift byte right logical *)
  (* I_R8ds *)  | [< AKA "sub";                    Fd "O,S,Z,A,P,C" >]  Subb of Lit    (* subtract byte *)
  (* I_R8ds *)  | [< AKA "xor";                    Fd "O,S,Z,A,P,C" >]  Xorb of Lit    (* logical exclusive-OR byte *)
type I_R16ds =  | [< AKA "adc";           Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcw of Lit    (* add word w/carry *)
  (* I_R16ds *) | [< AKA "add";                    Fd "O,S,Z,A,P,C" >]  Addw of Lit    (* add word *)
  (* I_R16ds *) | [< AKA "and";                    Fd "O,S,Z,A,P,C" >]  Andw of Lit    (* logical AND word *)
  (* I_R16ds *) | [< AKA "btc";                    Fd "C"           >]  Btcw of Lit    (* test bit in word and complement *)
  (* I_R16ds *) | [< AKA "btr";                    Fd "C"           >]  Btrw of Lit    (* test bit in word and reset *)
  (* I_R16ds *) | [< AKA "bts";                    Fd "C"           >]  Btsw of Lit    (* test bit in word and set *)
  (* I_R16ds *) | [< AKA "dec";                    Fd "O,S,Z,A,P"   >]  Decw           (* decrement unsigned word *)
  (* I_R16ds *) | [< AKA "imul";                   Fd "O,S,Z,A,P,C" >]  Imulw of Lit   (* multiply signed word *)
  (* I_R16ds *) | [< AKA "inc";                    Fd "O,S,Z,A,P"   >]  Incw           (* increment unsigned word *)
  (* I_R16ds *) | [< AKA "neg";                    Fd "O,S,Z,A,P,C" >]  Negw           (* two's complement negate word *)
  (* I_R16ds *) | [< AKA "not"                                      >]  Notw           (* one's complement negate word *)
  (* I_R16ds *) | [< AKA "or";                     Fd "O,S,Z,A,P,C" >]  Orw of Lit     (* logical inclusive-OR word *)
  (* I_R16ds *) | [< AKA "rcl";           Fs "C";  Fd "O,C"         >]  Rclw of Lit    (* rotate word right thru carry *)
  (* I_R16ds *) | [< AKA "rcr";           Fs "C";  Fd "O,C"         >]  Rcrw of Lit    (* rotate word right thru carry *)
  (* I_R16ds *) | [< AKA "rol";                    Fd "O,C"         >]  Rolw of Lit    (* rotate word left *)
  (* I_R16ds *) | [< AKA "ror";                    Fd "O,C"         >]  Rorw of Lit    (* rotate word right *)
  (* I_R16ds *) | [< AKA "sar";                    Fd "O,S,Z,A,P,C" >]  Sarw of Lit    (* shift word right arithmetic *)
  (* I_R16ds *) | [< AKA "sbb";           Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbw of Lit    (* subtract word w/borrow *)
  (* I_R16ds *) | [< AKA "shl,sal,salw";           Fd "O,S,Z,A,P,C" >]  Shlw of Lit    (* shift word left logical *)
  (* I_R16ds *) | [< AKA "shr";                    Fd "O,S,Z,A,P,C" >]  Shrw of Lit    (* shift word right logical *)
  (* I_R16ds *) | [< AKA "sub";                    Fd "O,S,Z,A,P,C" >]  Subw of Lit    (* subtract word *)
  (* I_R16ds *) | [< AKA "xor";                    Fd "O,S,Z,A,P,C" >]  Xorw of Lit    (* logical exclusive-OR word *)
type I_R32ds =  | [< AKA "adc";           Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcl of Lit    (* add long w/carry *)
  (* I_R32ds *) | [< AKA "add";                    Fd "O,S,Z,A,P,C" >]  Addl of Lit    (* add long *)
  (* I_R32ds *) | [< AKA "and";                    Fd "O,S,Z,A,P,C" >]  Andl of Lit    (* logical AND long *)
  (* I_R32ds *) | [< AKA "bswap"                                    >]  Bswapl         (* swap bytes in long *)
  (* I_R32ds *) | [< AKA "btc";                    Fd "C";          >]  Btcl of Lit    (* test bit in long and complement *)
  (* I_R32ds *) | [< AKA "btr";                    Fd "C"           >]  Btrl of Lit    (* test bit in long and reset *)
  (* I_R32ds *) | [< AKA "bts";                    Fd "C"           >]  Btsl of Lit    (* test bit in long and set *)
  (* I_R32ds *) | [< AKA "dec";                    Fd "O,S,Z,A,P"   >]  Decl           (* decrement unsigned long *)
  (* I_R32ds *) | [< AKA "imul";                   Fd "O,S,Z,A,P,C" >]  Imull of Lit   (* multiply signed long *)
  (* I_R32ds *) | [< AKA "inc";                    Fd "O,S,Z,A,P"   >]  Incl           (* increment unsigned long *)
  (* I_R32ds *) | [< AKA "neg";                    Fd "O,S,Z,A,P,C" >]  Negl           (* two's complement negate long *)
  (* I_R32ds *) | [< AKA "not"                                      >]  Notl           (* one's complement negate long *)
  (* I_R32ds *) | [< AKA "or";                     Fd "O,S,Z,A,P,C" >]  Orl of Lit     (* logical inclusive-OR long *)
  (* I_R32ds *) | [< AKA "rcl";           Fs "C";  Fd "O,C"         >]  Rcll of Lit    (* rotate long right thru carry *)
  (* I_R32ds *) | [< AKA "rcr";           Fs "C";  Fd "O,C"         >]  Rcrl of Lit    (* rotate long right thru carry *)
  (* I_R32ds *) | [< AKA "rol";                    Fd "O,C"         >]  Roll of Lit    (* rotate long left *)
  (* I_R32ds *) | [< AKA "ror";                    Fd "O,C"         >]  Rorl of Lit    (* rotate long right *)
  (* I_R32ds *) | [< AKA "sar";                    Fd "O,S,Z,A,P,C" >]  Sarl of Lit    (* shift long right arithmetic *)
  (* I_R32ds *) | [< AKA "sbb";           Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbl of Lit    (* subtract long w/borrow *)
  (* I_R32ds *) | [< AKA "shl,sal,sall";           Fd "O,S,Z,A,P,C" >]  Shll of Lit    (* shift long left logical *)
  (* I_R32ds *) | [< AKA "shr";                    Fd "O,S,Z,A,P,C" >]  Shrl of Lit    (* shift long right logical *)
  (* I_R32ds *) | [< AKA "sub";                    Fd "O,S,Z,A,P,C" >]  Subl of Lit    (* subtract long *)
  (* I_R32ds *) | [< AKA "xor";                    Fd "O,S,Z,A,P,C" >]  Xorl of Lit    (* logical exclusive-OR long *)
type I_R64ds =  | [< AKA "adc";           Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcq of Lit    (* add quad w/carry *)
  (* I_R64ds *) | [< AKA "add";                    Fd "O,S,Z,A,P,C" >]  Addq of Lit    (* add quad *)
  (* I_R64ds *) | [< AKA "and";                    Fd "O,S,Z,A,P,C" >]  Andq of Lit    (* logical AND quad *)
  (* I_R64ds *) | [< AKA "bswap"                                    >]  Bswapq         (* swap bytes in quad *)
  (* I_R64ds *) | [< AKA "btc";                    Fd "C"           >]  Btcq of Lit    (* test bit in quad and complement *)
  (* I_R64ds *) | [< AKA "btr";                    Fd "C"           >]  Btrq of Lit    (* test bit in quad and reset *)
  (* I_R64ds *) | [< AKA "bts";                    Fd "C"           >]  Btsq of Lit    (* test bit in quad and set *)
  (* I_R64ds *) | [< AKA "dec";                    Fd "O,S,Z,A,P"   >]  Decq           (* decrement unsigned quad *)
  (* I_R64ds *) | [< AKA "imul";                   Fd "O,S,Z,A,P,C" >]  Imulq of Lit   (* signed multiply quad *)
  (* I_R64ds *) | [< AKA "inc";                    Fd "O,S,Z,A,P"   >]  Incq           (* increment unsigned quad *)
  (* I_R64ds *) | [< AKA "neg";                    Fd "O,S,Z,A,P,C" >]  Negq           (* two's complement negate quad *)
  (* I_R64ds *) | [< AKA "not";                                     >]  Notq           (* one's complement negate quad *)
  (* I_R64ds *) | [< AKA "or";                     Fd "O,S,Z,A,P,C" >]  Orq of Lit     (* logical inclusive-OR quad *)
  (* I_R64ds *) | [< AKA "rcl";           Fs "C";  Fd "O,C"         >]  Rclq of Lit    (* rotate quad right thru carry *)
  (* I_R64ds *) | [< AKA "rcr";           Fs "C";  Fd "O,C"         >]  Rcrq of Lit    (* rotate quad right thru carry *)
  (* I_R64ds *) | [< AKA "rol";                    Fd "O,C"         >]  Rolq of Lit    (* rotate quad left *)
  (* I_R64ds *) | [< AKA "ror";                    Fd "O,C"         >]  Rorq of Lit    (* rotate quad right *)
  (* I_R64ds *) | [< AKA "sar";                    Fd "O,S,Z,A,P,C" >]  Sarq of Lit    (* shift quad right arithmetic *)
  (* I_R64ds *) | [< AKA "sbb";           Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbq of Lit    (* subtract quad w/borrow *) 
  (* I_R64ds *) | [< AKA "shl,sal,salq";           Fd "O,S,Z,A,P,C" >]  Shlq of Lit    (* shift quad left logical *)
  (* I_R64ds *) | [< AKA "shr";                    Fd "O,S,Z,A,P,C" >]  Shrq of Lit    (* shift quad right logical *)
  (* I_R64ds *) | [< AKA "sub";                    Fd "O,S,Z,A,P,C" >]  Subq of Lit    (* subtract quad *)
  (* I_R64ds *) | [< AKA "xor";                    Fd "O,S,Z,A,P,C" >]  Xorq of Lit    (* logical exclusive-OR quad *)

(*****************************************************************************************************************************************************************************)

type I_M8d =   |                                                                                         Movb of Lit    (* store immediate byte *)
  (* I_M8d *)  | [<             AKA "seta,setnbe,setnbeb";                                Fs "Z,C"   >]  Setab          (* set byte if above *)
  (* I_M8d *)  | [<             AKA "setc,setb,setnae,setbb,setnaeb";                     Fs "C"     >]  Setcb          (* set byte if carry *)
  (* I_M8d *)  | [<             AKA "setg,setnle,setnleb";                                Fs "Z,S,O" >]  Setgb          (* set byte if greater *)
  (* I_M8d *)  | [<             AKA "setl,setnge,setngeb";                                Fs "S,O"   >]  Setlb          (* set byte if less *)
  (* I_M8d *)  | [<             AKA "setna,setbe,setbeb";                                 Fs "Z,C"   >]  Setnab         (* set byte if not above *)
  (* I_M8d *)  | [<             AKA "setnc,setnb,setae,setnbb,setaeb";                    Fs "C"     >]  Setncb         (* set byte if not carry *)
  (* I_M8d *)  | [<             AKA "setng,setle,setleb";                                 Fs "Z,S,O" >]  Setngb         (* set byte if not greater *)
  (* I_M8d *)  | [<             AKA "setnl,setge,setgeb";                                 Fs "S,O"   >]  Setnlb         (* set byte if not less*)
  (* I_M8d *)  | [<             AKA "setnp,setpo,setpob";                                 Fs "P"     >]  Setnpb         (* set byte if no parity *)
  (* I_M8d *)  | [<             AKA "setp,setpe,setpeb";                                  Fs "P"     >]  Setpb          (* set byte if parity *)
  (* I_M8d *)  | [<             AKA "setz,sete,seteb";                                    Fs "Z"     >]  Setzb          (* set byte if zero *)
  (* I_M8d *)  | [<             AKA "setnz,setne,setneb";                                 Fs "Z"     >]  Setnzb         (* set byte if not zero *)
  (* I_M8d *)  | [<             AKA "setno";                                              Fs "O"     >]  Setnob         (* set byte if not overflow *)
  (* I_M8d *)  | [<             AKA "setns";                                              Fs "S"     >]  Setnsb         (* set byte if not sign *)
  (* I_M8d *)  | [<             AKA "seto";                                               Fs "O"     >]  Setob          (* set byte if overflow *)
  (* I_M8d *)  | [<             AKA "sets";                                               Fs "S"     >]  Setsb          (* set byte if sign *)
type I_M16d =  |                                                                                         Movw of Lit    (* store immediate word *)
  (* I_M16d *) | [<                                                     Mem;  Rds "rsp"              >]  Popw           (* pop word from stack *)
type I_M32d =  | [< ISA "AVX";                                          Mem                          >]  Vstmxcsr       (* store MXCSR register *)
  (* I_M32d *) |                                                                                         Movl of Lit    (* store immediate long *)
  (* I_M32d *) | [<                                                     Mem                          >]  Stmxcsr        (* store MXCSR register *)
type I_M64d =  |                                                                                         Movq of Lit    (* store immediate quad *)
  (* I_M64d *) | [<                                                     Mem;  Rds "rsp"              >]  Popq           (* pop quad from stack *)

(*****************************************************************************************************************************************************************************)

type I_M8ds =    | [<                                             Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcb of Lit    (* add bytes w/carry *)
  (* I_M8ds *)   | [<                                                      Fd "O,S,Z,A,P,C" >]  Addb of Lit    (* add bytes *)
  (* I_M8ds *)   | [<                                                      Fd "O,S,Z,A,P,C" >]  Andb of Lit    (* logical AND bytes *)
  (* I_M8ds *)   | [<                                                      Fd "O,S,Z,A,P"   >]  Decb           (* decrement unsigned byte *)
  (* I_M8ds *)   | [<                                                      Fd "O,S,Z,A,P"   >]  Incb           (* increment unsigned byte *)
  (* I_M8ds *)   | [<                                                      Fd "O,S,Z,A,P,C" >]  Negb           (* two's complement negate byte *)
  (* I_M8ds *)   |                                                                              Notb           (* one's complement negate byte *)
  (* I_M8ds *)   | [<                                                      Fd "O,S,Z,A,P,C" >]  Orb of Lit     (* logical inclusive-OR bytes *)
  (* I_M8ds *)   | [<                                             Fs "C";  Fd "O,C"         >]  Rclb of Lit    (* rotate left thru carry byte *)
  (* I_M8ds *)   | [<                                             Fs "C";  Fd "O,C"         >]  Rcrb of Lit    (* rotate right thru carry byte *)
  (* I_M8ds *)   | [<                                                      Fd "O,C"         >]  Rolb of Lit    (* rotate left byte *)
  (* I_M8ds *)   | [<                                                      Fd "O,C"         >]  Rorb of Lit    (* rotate right byte *)
  (* I_M8ds *)   | [<                                                      Fd "O,S,Z,A,P,C" >]  Sarb of Lit    (* shift right arithmetic byte *)
  (* I_M8ds *)   | [<                                             Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbb of Lit    (* subtract with borrow byte *)
  (* I_M8ds *)   | [< AKA "salb";                                          Fd "O,S,Z,A,P,C" >]  Shlb of Lit    (* shift left logical byte *)
  (* I_M8ds *)   | [<                                                      Fd "O,S,Z,A,P,C" >]  Shrb of Lit    (* shift right logical byte *)
  (* I_M8ds *)   | [<                                                      Fd "O,S,Z,A,P,C" >]  Subb of Lit    (* subtract bytes *)
  (* I_M8ds *)   | [<                                                      Fd "O,S,Z,A,P,C" >]  Xorb of Lit    (* logical exclusive-OR bytes *)
type I_M16ds =   | [<                                             Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcw of Lit    (* add words w/carry *)
  (* I_M16ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Addw of Lit    (* add words *)
  (* I_M16ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Andw of Lit    (* logical AND words *)
  (* I_M16ds *)  | [<                                                      Fd "C"           >]  Btcw of Lit    (* test bit in word and complement *)
  (* I_M16ds *)  | [<                                                      Fd "C"           >]  Btrw of Lit    (* test bit in word and reset *)
  (* I_M16ds *)  | [<                                                      Fd "C"           >]  Btsw of Lit    (* test bit in word and set *)
  (* I_M16ds *)  | [<                                                      Fd "O,S,Z,A,P"   >]  Decw           (* decrement unsigned word *)
  (* I_M16ds *)  | [<                                                      Fd "O,S,Z,A,P"   >]  Incw           (* increment unsigned word *)
  (* I_M16ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Negw           (* two's complement negate word *)
  (* I_M16ds *)  |                                                                              Notw           (* one's complement negate word *)
  (* I_M16ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Orw of Lit     (* logical inclusive-OR words *)
  (* I_M16ds *)  | [<                                             Fs "C";  Fd "O,C"         >]  Rclw of Lit    (* rotate word left thru carry *)
  (* I_M16ds *)  | [<                                             Fs "C";  Fd "O,C"         >]  Rcrw of Lit    (* rotate word right thru carry *)
  (* I_M16ds *)  | [<                                                      Fd "O,C"         >]  Rolw of Lit    (* rotate word left *)
  (* I_M16ds *)  | [<                                                      Fd "O,C"         >]  Rorw of Lit    (* rotate word right *)
  (* I_M16ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Sarw of Lit    (* shift word right arithmetic *)
  (* I_M16ds *)  | [<                                             Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbw of Lit    (* subtract word w/borrow *)
  (* I_M16ds *)  | [< AKA "salw";                                          Fd "O,S,Z,A,P,C" >]  Shlw of Lit    (* shift word left logical *)
  (* I_M16ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Shrw of Lit    (* shift word right logical *)
  (* I_M16ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Subw of Lit    (* subtract words *)
  (* I_M16ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Xorw of Lit    (* logical exclusive-OR words *)
type I_M32ds =   | [<                                             Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcl of Lit    (* add longs w/carry *)
  (* I_M32ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Addl of Lit    (* add longs *)
  (* I_M32ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Andl of Lit    (* logical AND longs *)
  (* I_M32ds *)  | [<                                                      Fd "C"           >]  Btcl of Lit    (* test bit in long and complement *)
  (* I_M32ds *)  | [<                                                      Fd "C"           >]  Btrl of Lit    (* test bit in long and reset *)
  (* I_M32ds *)  | [<                                                      Fd "C"           >]  Btsl of Lit    (* test bit in long and set *)
  (* I_M32ds *)  | [<                                                      Fd "O,S,Z,A,P"   >]  Decl           (* decrement unsigned long *)
  (* I_M32ds *)  | [<                                                      Fd "O,S,Z,A,P"   >]  Incl           (* increment unsigned long *)
  (* I_M32ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Negl           (* two's complement negate long *)
  (* I_M32ds *)  |                                                                              Notl           (* one's complement negate long *)
  (* I_M32ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Orl of Lit     (* logical inclusive-OR longs *)
  (* I_M32ds *)  | [<                                             Fs "C";  Fd "O,C"         >]  Rcll of Lit    (* rotate left thru carry long *)
  (* I_M32ds *)  | [<                                             Fs "C";  Fd "O,C"         >]  Rcrl of Lit    (* rotate right thru carry long *)
  (* I_M32ds *)  | [<                                                      Fd "O,C"         >]  Roll of Lit    (* rotate left long *)
  (* I_M32ds *)  | [<                                                      Fd "O,C"         >]  Rorl of Lit    (* rotate right long *)
  (* I_M32ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Sarl of Lit    (* shift right arithmetic long *)
  (* I_M32ds *)  | [<                                             Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbl of Lit    (* subtract long w/borrow *)
  (* I_M32ds *)  | [< AKA "sall";                                          Fd "O,S,Z,A,P,C" >]  Shll of Lit    (* shift left logical long *)
  (* I_M32ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Shrl of Lit    (* shift right logical long *)
  (* I_M32ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Subl of Lit    (* subtract longs *)
  (* I_M32ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Xorl of Lit    (* logical exclusive-OR longs *)
type I_M64ds =   | [<                                             Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcq of Lit    (* add quads with carry *)
  (* I_M64ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Addq of Lit    (* add quads *)
  (* I_M64ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Andq of Lit    (* logical AND quads *)
  (* I_M64ds *)  | [<                                                      Fd "C"           >]  Btcq of Lit    (* test bit in quad and complement *)
  (* I_M64ds *)  | [<                                                      Fd "C"           >]  Btrq of Lit    (* test bit in quad and reset *)
  (* I_M64ds *)  | [<                                                      Fd "C"           >]  Btsq of Lit    (* test bit in word and set *)
  (* I_M64ds *)  | [<              Rs "rbx,rcx";  Rds "rax,rdx";           Fd "Z"           >]  Cmpxchg8b      (* compare and exchange 8 bytes *)
  (* I_M64ds *)  | [<                                                      Fd "O,S,Z,A,P"   >]  Decq           (* decrement unsigned quad *)
  (* I_M64ds *)  | [<                                                      Fd "O,S,Z,A,P"   >]  Incq           (* increment unsigned quad *)
  (* I_M64ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Negq           (* two's complement negate quad *)
  (* I_M64ds *)  |                                                                              Notq           (* one's complement negate quad *)
  (* I_M64ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Orq of Lit     (* logical inclusive-OR quads *)
  (* I_M64ds *)  | [<                                             Fs "C";  Fd "O,C"         >]  Rclq of Lit    (* rotate left thru carry quad *)
  (* I_M64ds *)  | [<                                             Fs "C";  Fd "O,C"         >]  Rcrq of Lit    (* rotate right thru carry quad *)
  (* I_M64ds *)  | [<                                                      Fd "O,C"         >]  Rolq of Lit    (* rotate left quad *)
  (* I_M64ds *)  | [<                                                      Fd "O,C"         >]  Rorq of Lit    (* rotate right quad *)
  (* I_M64ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Sarq of Lit    (* shift right arithmetic quad *)
  (* I_M64ds *)  | [<                                             Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbq of Lit    (* subtract with borrow quad *)
  (* I_M64ds *)  | [< AKA "salq";                                          Fd "O,S,Z,A,P,C" >]  Shlq of Lit    (* shift left logical quad *)
  (* I_M64ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Shrq of Lit    (* shift right logical quad *)
  (* I_M64ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Subq of Lit    (* subtract quads *)
  (* I_M64ds *)  | [<                                                      Fd "O,S,Z,A,P,C" >]  Xorq of Lit    (* logical exclusive-OR quads *)
type I_M128ds =  | [<              Rs "rbx,rcx";  Rds "rax,rdx";           Fd "Z"           >]  Cmpxchg16b     (* compare and exchange 16 bytes *)

(*****************************************************************************************************************************************************************************)

type I_R8sR8s =    | [< AKA "cmp";   Fd "O,S,Z,A,P,C" >]  Cmpb     (* arithmetic compare byte *)
  (* I_R8sR8s *)   | [< AKA "test";  Fd "O,S,Z,A,P,C" >]  Testb    (* logical compare byte *)
type I_R16sR16s =  | [< AKA "bt";    Fd "C"           >]  Btw      (* test bit in word *)
  (* I_R16sR16s *) | [< AKA "cmp";   Fd "O,S,Z,A,P,C" >]  Cmpw     (* arithmetic compare word *)
  (* I_R16sR16s *) | [< AKA "test";  Fd "O,S,Z,A,P,C" >]  Testw    (* logical compare word *)
type I_R32sR32s =  | [< AKA "bt";    Fd "C"           >]  Btl      (* test bit in long *)
  (* I_R32sR32s *) | [< AKA "cmp";   Fd "O,S,Z,A,P,C" >]  Cmpl     (* arithmetic compare long *)
  (* I_R32sR32s *) | [< AKA "test";  Fd "O,S,Z,A,P,C" >]  Testl    (* logical compare long *)
type I_R64sR64s =  | [< AKA "bt";    Fd "C"           >]  Btq      (* test bit in quad *)
  (* I_R64sR64s *) | [< AKA "cmp";   Fd "O,S,Z,A,P,C" >]  Cmpq     (* arithmetic compare quad *)
  (* I_R64sR64s *) | [< AKA "test";  Fd "O,S,Z,A,P,C" >]  Testq    (* logical compare quad *)

(*****************************************************************************************************************************************************************************)

type I_R8sM8s =    | [< AKA "cmp";    Fd "O,S,Z,A,P,C" >]  Cmpb     (* arithmetic compare byte *)
  (* I_R8sM8s *)   | [< AKA "test";   Fd "O,S,Z,A,P,C" >]  Testb    (* logical compare byte *)
type I_R16sM16s =  | [< AKA "bt";     Fd "C"           >]  Btw      (* test bit in word *)
  (* I_R16sM16s *) | [< AKA "cmp";    Fd "O,S,Z,A,P,C" >]  Cmpw     (* arithmetic compare word *)
  (* I_R16sM16s *) | [< AKA "test";   Fd "O,S,Z,A,P,C" >]  Testw    (* logical compare word *)
type I_R32sM32s =  | [< AKA "bt";     Fd "C"           >]  Btl      (* test bit in long *)
  (* I_R32sM32s *) | [< AKA "cmp";    Fd "O,S,Z,A,P,C" >]  Cmpl     (* arithmetic compare long *)
  (* I_R32sM32s *) | [< AKA "test";   Fd "O,S,Z,A,P,C" >]  Testl    (* logical compare long *)
type I_R64sM64s =  | [< AKA "bt";     Fd "C"           >]  Btq      (* test bit in quad *)
  (* I_R64sM64s *) | [< AKA "cmp";    Fd "O,S,Z,A,P,C" >]  Cmpq     (* arithmetic compare quad *)
  (* I_R64sM64s *) | [< AKA "test";   Fd "O,S,Z,A,P,C" >]  Testq    (* logical compare quad *)

(*****************************************************************************************************************************************************************************)

type I_M8sR8s =    | [< AKA "cmp";   Fd "O,S,Z,A,P,C" >]  Cmpb     (* arithmetic compare byte *)
  (* I_M8sR8s *)   | [< AKA "test";  Fd "O,S,Z,A,P,C" >]  Testb    (* logical compare byte *)
type I_M16sR16s =  | [< AKA "cmp";   Fd "O,S,Z,A,P,C" >]  Cmpw     (* arithmetic compare word *)
  (* I_M16sR16s *) | [< AKA "test";  Fd "O,S,Z,A,P,C" >]  Testw    (* logical compare word *)
type I_M32sR32s =  | [< AKA "cmp";   Fd "O,S,Z,A,P,C" >]  Cmpl     (* arithmetic compare long *)
  (* I_M32sR32s *) | [< AKA "test";  Fd "O,S,Z,A,P,C" >]  Testl    (* logical compare long *)
type I_M64sR64s =  | [< AKA "cmp";   Fd "O,S,Z,A,P,C" >]  Cmpq     (* arithmetic compare quad *)
  (* I_M64sR64s *) | [< AKA "test";  Fd "O,S,Z,A,P,C" >]  Testq    (* logical compare quad *)

(*****************************************************************************************************************************************************************************)

type I_R8sR8d =    | [<               AKA "mov"                       >]  Movb            (* copy byte *)
type I_R8sR16d =   | [<               AKA "movsx"                     >]  Movsbw          (* move byte w/sign-extension *)
  (* I_R8sR16d *)  | [<               AKA "movzx"                     >]  Movzbw          (* move byte w/zero-extension *)
type I_R8sR32d =   | [<               AKA "movsx"                     >]  Movsbl          (* move byte w/sign-extension *)
  (* I_R8sR32d *)  | [<               AKA "movzx"                     >]  Movzbl          (* move byte w/zero-extension *)
type I_R8sR64d =   | [<               AKA "movsx"                     >]  Movsbq          (* move byte w/sign-extension *)
  (* I_R8sR64d *)  | [<               AKA "movzx"                     >]  Movzbq          (* move byte w/zero-extension *)
type I_R16sR16d =  | [<               AKA "bsf";     Fd "Z"           >]  Bsfw            (* bit scan forward word *)
  (* I_R16sR16d *) | [<               AKA "bsr";     Fd "Z"           >]  Bsrw            (* bit scan reverse word *)
  (* I_R16sR16d *) | [<               AKA "imul";    Fd "O,S,Z,A,P,C" >]  Imulw of Lit    (* signed multiply word *)
  (* I_R16sR16d *) | [< ISA "SSE4a";  AKA "lzcnt";   Fd "C,Z"         >]  Lzcntw          (* count leading zeros in word *)
  (* I_R16sR16d *) | [<               AKA "mov"                       >]  Movw            (* copy word *)
  (* I_R16sR16d *) | [< ISA "POPC";   AKA "popcnt"                    >]  Popcntw         (* count number of set bits in word *)
type I_R16sR32d =  | [<               AKA "movsx"                     >]  Movswl          (* move word w/sign-extension *)
  (* I_R16sR32d *) | [<               AKA "movzx"                     >]  Movzwl          (* move word w/zero-extension *)
type I_R16sR64d =  | [<               AKA "movsx"                     >]  Movswq          (* move word w/sign-extension *)
  (* I_R16sR64d *) | [<               AKA "movzx"                     >]  Movzwq          (* move word w/zero-extension *)
type I_R32sR32d =  | [<               AKA "bsf";     Fd "Z"           >]  Bsfl            (* bit scan forward long *)
  (* I_R32sR32d *) | [<               AKA "bsr";     Fd "Z"           >]  Bsrl            (* bit scan reverse long *)
  (* I_R32sR32d *) | [<               AKA "imul";    Fd "O,S,Z,A,P,C" >]  Imull of Lit    (* signed multiply long *)
  (* I_R32sR32d *) | [< ISA "SSE4a";  AKA "lzcnt";   Fd "C,Z"         >]  Lzcntl          (* count leading zeros in long *)
  (* I_R32sR32d *) | [<               AKA "mov"                       >]  Movl            (* copy long *)
  (* I_R32sR32d *) | [< ISA "POPC";   AKA "popcnt"                    >]  Popcntl         (* count number of set bits in long *)
type I_R32sR64d =  | [<               AKA "movsx"                     >]  Movslq          (* move long w/sign-extension *)
type I_R64sR64d =  | [<               AKA "bsf";     Fd "Z"           >]  Bsfq            (* bit scan forward quad *)
  (* I_R64sR64d *) | [<               AKA "bsr";     Fd "Z"           >]  Bsrq            (* bit scan reverse quad *)
  (* I_R64sR64d *) | [<               AKA "imul";    Fd "O,S,Z,A,P,C" >]  Imulq of Lit    (* signed multiply quad *)
  (* I_R64sR64d *) | [< ISA "SSE4a";  AKA "lzcnt";   Fd "C,Z"         >]  Lzcntq          (* count leading zeros in quad *)
  (* I_R64sR64d *) | [<               AKA "mov"                       >]  Movq            (* copy quad *)
  (* I_R64sR64d *) | [< ISA "POPC";   AKA "popcnt"                    >]  Popcntq         (* count number of set bits in quad *)

(*****************************************************************************************************************************************************************************)

type I_M8sR8d =    | [<               AKA "mov"                          >]  Movb            (* load byte *)
type I_M8sR16d =   |                                                         Movsbw          (* load byte into word w/sign-extension *)
  (* I_M8sR16d *)  |                                                         Movzbw          (* load byte into word w/zero-extension *)
type I_M8sR32d =   |                                                         Movsbl          (* load byte into long w/sign-extension *)
  (* I_M8sR32d *)  |                                                         Movzbl          (* load byte into long w/zero-extension *)
type I_M8sR64d =   |                                                         Movsbq          (* load byte into quad w/sign-extension *)
  (* I_M8sR64d *)  |                                                         Movzbq          (* load byte into quad w/zero-extension *)
type I_M16sR16d =  | [<               AKA "bsf";        Fd "Z"           >]  Bsfw            (* bit scan forward *)
  (* I_M16sR16d *) | [<               AKA "bsr";        Fd "Z"           >]  Bsrw            (* bit scan reverse *)
  (* I_M16sR16d *) | [<               AKA "imul";       Fd "O,S,Z,A,P,C" >]  Imulw of Lit    (* multiply signed word *)
  (* I_M16sR16d *) | [< ISA "SSE4a";  AKA "lzcnt";      Fd "C,Z"         >]  Lzcntw          (* count leading zeros of word *)
  (* I_M16sR16d *) | [<               AKA "mov"                          >]  Movw            (* load word *)
  (* I_M16sR16d *) | [< ISA "MOVBE";  AKA "movbe"                        >]  Movbew          (* load word and swap bytes *)
  (* I_M16sR16d *) | [< ISA "POPC";   AKA "popcnt"                       >]  Popcntw         (* return the count of number of set bits *)
type I_M16sR32d =  |                                                         Movswl          (* load word into long w/sign-extension *)
  (* I_M16sR32d *) |                                                         Movzwl          (* load word into long w/zero-extension *)
type I_M16sR64d =  |                                                         Movswq          (* load word into quad w/sign-extension *)
  (* I_M16sR64d *) |                                                         Movzwq          (* load word into quad w/zero-extension *)
type I_M32sR32d =  | [<               AKA "cvtsd2si"                     >]  Cvtsd2sil       (* convert float64 to long *)
  (* I_M32sR32d *) | [<               AKA "cvttsd2si"                    >]  Cvttsd2sil      (* convert w/truncation float64 to long *)
  (* I_M32sR32d *) | [<               AKA "cvtss2si"                     >]  Cvtss2sil       (* convert float32 to long *)
  (* I_M32sR32d *) | [<               AKA "cvttss2si"                    >]  Cvttss2sil      (* convert w/truncation float32 to long *)
  (* I_M32sR32d *) | [<               AKA "bsf";        Fd "Z"           >]  Bsfl            (* bit scan forward in long *)
  (* I_M32sR32d *) | [<               AKA "bsr";        Fd "Z"           >]  Bsrl            (* bit scan reverse in long *)
  (* I_M32sR32d *) | [<               AKA "imul";       Fd "O,S,Z,A,P,C" >]  Imull of Lit    (* multiply signed long *)
  (* I_M32sR32d *) | [< ISA "SSE4a";  AKA "lzcnt";      Fd "C,Z"         >]  Lzcntl          (* count leading zeros of long *)
  (* I_M32sR32d *) | [<               AKA "mov"                          >]  Movl            (* load long *)
  (* I_M32sR32d *) | [< ISA "MOVBE";  AKA "movbe"                        >]  Movbel          (* load long and swap bytes *)
  (* I_M32sR32d *) | [< ISA "POPC";   AKA "popcnt"                       >]  Popcntl         (* return the count of number of set bits *)
type I_M32sR64d =  |                                                         Movslq          (* load long into quad w/sign-extension *)
type I_M64sR64d =  | [<               AKA "cvtsd2si"                     >]  Cvtsd2siq       (* convert float64 to quad *)
  (* I_M64sR64d *) | [<               AKA "cvttsd2si"                    >]  Cvttsd2siq      (* convert w/truncation float64 to quad *)
  (* I_M64sR64d *) | [<               AKA "cvtss2si"                     >]  Cvtss2siq       (* convert float32 to quad *)
  (* I_M64sR64d *) | [<               AKA "cvttss2si"                    >]  Cvttss2siq      (* convert w/truncation float32 to quad *)
  (* I_M64sR64d *) | [<               AKA "bsf";        Fd "Z"           >]  Bsfq            (* bit scan forward in quad *)
  (* I_M64sR64d *) | [<               AKA "bsr";        Fd "Z"           >]  Bsrq            (* bit scan reverse in quad *)
  (* I_M64sR64d *) | [<               AKA "imul";       Fd "O,S,Z,A,P,C" >]  Imulq of Lit    (* multiply signed quad *)
  (* I_M64sR64d *) | [< ISA "SSE4a";  AKA "lzcnt";      Fd "C,Z"         >]  Lzcntq          (* count leading zeros of quad *)
  (* I_M64sR64d *) | [<               AKA "mov"                          >]  Movq            (* load quad *)
  (* I_M64sR64d *) | [< ISA "MOVBE";  AKA "movbe"                        >]  Movbeq          (* load quad and swap bytes *)
  (* I_M64sR64d *) | [< ISA "POPC";   AKA "popcnt"                       >]  Popcntq         (* return the count of number of set bits *)

(*****************************************************************************************************************************************************************************)

type I_R8sM8d =    | [<               AKA "mov"    >]  Movb       (* store byte *)
type I_R16sM16d =  | [<               AKA "mov"    >]  Movw       (* store word *)
  (* I_R16sM16d *) | [< ISA "MOVBE";  AKA "movbe"  >]  Movbew     (* store word after swapping bytes *)
type I_R32sM32d =  | [<               AKA "mov"    >]  Movl       (* store long *)
  (* I_R32sM32d *) | [< ISA "MOVBE";  AKA "movbe"  >]  Movbel     (* store long after swapping bytes *)
  (* I_R32sM32d *) | [<               AKA "movnti" >]  Movntil    (* store long w/NT hint *)
type I_R64sM64d =  | [<               AKA "mov"    >]  Movq       (* store quad *)
  (* I_R64sM64d *) | [< ISA "MOVBE";  AKA "movbe"  >]  Movbeq     (* store quad after swapping bytes *)
  (* I_R64sM64d *) | [<               AKA "movnti" >]  Movntiq    (* store quad w/NT-hint *)

(*****************************************************************************************************************************************************************************)

type I_R8dsR8ds =    | [< AKA "xadd";  Fd "O,S,Z,A,P,C" >]  Xaddb    (* exchange and add byte *)
  (* I_R8dsR8ds *)   | [< AKA "xchg"                    >]  Xchgb    (* exchange byte *)
type I_R16dsR16ds =  | [< AKA "xadd";  Fd "O,S,Z,A,P,C" >]  Xaddw    (* exchange and add word *) 
  (* I_R16dsR16ds *) | [< AKA "xchg"                    >]  Xchgw    (* exchange word *) 
type I_R32dsR32ds =  | [< AKA "xadd";  Fd "O,S,Z,A,P,C" >]  Xaddl    (* exchange and add long *)
  (* I_R32dsR32ds *) | [< AKA "xchg"                    >]  Xchgl    (* exchange long *)
type I_R64dsR64ds =  | [< AKA "xadd";  Fd "O,S,Z,A,P,C" >]  Xaddq    (* exchange and add quad *)
  (* I_R64dsR64ds *) | [< AKA "xchg"                    >]  Xchgq    (* exchange quad *)

(*****************************************************************************************************************************************************************************)

type I_R8dsM8ds =    | [< AKA "xadd";  Fd "O,S,Z,A,P,C" >]  Xaddb    (* exchange and add byte *)
  (* I_R8dsM8ds *)   | [< AKA "xchg"                    >]  Xchgb    (* exchange byte *)
type I_R16dsM16ds =  | [< AKA "xadd";  Fd "O,S,Z,A,P,C" >]  Xaddw    (* exchange and add word *)
  (* I_R16dsM16ds *) | [< AKA "xchg"                    >]  Xchgw    (* exchange word *)
type I_R32dsM32ds =  | [< AKA "xadd";  Fd "O,S,Z,A,P,C" >]  Xaddl    (* exchange and add long *)
  (* I_R32dsM32ds *) | [< AKA "xchg"                    >]  Xchgl    (* exchange long *)
type I_R64dsM64ds =  | [< AKA "xadd";  Fd "O,S,Z,A,P,C" >]  Xaddq    (* exchange and add quad *)
  (* I_R64dsM64ds *) | [< AKA "xchg"                    >]  Xchgq    (* exchange quad *)

(*****************************************************************************************************************************************************************************)

type I_M8dsR8ds =    | [< AKA "xchg" >]  Xchgb    (* exchange bytes *) 
type I_M16dsR16ds =  | [< AKA "xchg" >]  Xchgw    (* exchange words *) 
type I_M32dsR32ds =  | [< AKA "xchg" >]  Xchgl    (* exchange longs *) 
type I_M64dsR64ds =  | [< AKA "xchg" >]  Xchgq    (* exchange quads *) 

(*****************************************************************************************************************************************************************************)

type I_R8sR16sR16ds =  | [< AKA "shld";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shldw    (* shift left double-width word *)
  (* I_R8sR16sR16ds *) | [< AKA "shrd";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shrdw    (* shift right double-width word *)
type I_R8sR32sR32ds =  | [< AKA "shld";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shldl    (* shift left double-width long *)
  (* I_R8sR32sR32ds *) | [< AKA "shrd";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shrdl    (* shift right double-width long *)
type I_R8sR64sR64ds =  | [< AKA "shld";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shldq    (* shift left double-width quad *)
  (* I_R8sR64sR64ds *) | [< AKA "shrd";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shrdq    (* shift right double-width quad *)

(*****************************************************************************************************************************************************************************)

type I_R8sR16sM16ds =  | [< AKA "shld";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shldw    (* shift left double-width word *)
  (* I_R8sR16sM16ds *) | [< AKA "shrd";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shrdw    (* shift right double-width word *)
type I_R8sR32sM32ds =  | [< AKA "shld";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shldl    (* shift left double-width long *)
  (* I_R8sR32sM32ds *) | [< AKA "shrd";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shrdl    (* shift right double-width long *)
type I_R8sR64sM64ds =  | [< AKA "shld";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shldq    (* shift left double-width quad *)
  (* I_R8sR64sM64ds *) | [< AKA "shrd";  Rs "cl";  Fd "O,S,Z,A,P,C" >]  Shrdq    (* shift right double-width quad *)

(*****************************************************************************************************************************************************************************)

type I_R8sR8ds =    | [<                AKA "adc";                                            Fs "C";      Fd "O,S,Z,A,P,C" >]  Adcb            (* add byte w/carry *)
  (* I_R8sR8ds *)   | [<                AKA "add";                                                         Fd "O,S,Z,A,P,C" >]  Addb            (* add byte *)
  (* I_R8sR8ds *)   | [<                AKA "and";                                                         Fd "O,S,Z,A,P,C" >]  Andb            (* logical AND byte *)
  (* I_R8sR8ds *)   | [< ISA "SSE4_2";  AKA "crc32"                                                                         >]  Crc32b          (* accumulate CRC32 value *)
  (* I_R8sR8ds *)   | [<                AKA "or";                                                          Fd "O,S,Z,A,P,C" >]  Orb             (* logical inclusive-OR byte *)
  (* I_R8sR8ds *)   | [<                AKA "rcl";                                  Rs "cl";  Fs "C";      Fd "O,C"         >]  Rclb            (* rotate byte left thru carry *)
  (* I_R8sR8ds *)   | [<                AKA "rcr";                                  Rs "cl";  Fs "C";      Fd "O,C"         >]  Rcrb            (* rotate byte right thru carry *)
  (* I_R8sR8ds *)   | [<                AKA "rol";                                  Rs "cl";               Fd "O,C"         >]  Rolb            (* rotate byte left *)
  (* I_R8sR8ds *)   | [<                AKA "ror";                                  Rs "cl";               Fd "O,C"         >]  Rorb            (* rotate byte right *)
  (* I_R8sR8ds *)   | [<                AKA "sar";                                  Rs "cl";               Fd "O,S,Z,A,P,C" >]  Sarb            (* shift byte right arithmetic *)
  (* I_R8sR8ds *)   | [<                AKA "sbb";                                            Fs "C";      Fd "O,S,Z,A,P,C" >]  Sbbb            (* subtract byte w/borrow *)
  (* I_R8sR8ds *)   | [<                AKA "shl,sal,salb";                         Rs "cl";               Fd "O,S,Z,A,P,C" >]  Shlb            (* shift byte left logical *)
  (* I_R8sR8ds *)   | [<                AKA "shr";                                  Rs "cl";               Fd "O,S,Z,A,P,C" >]  Shr             (* shift byte right logical *)
  (* I_R8sR8ds *)   | [<                AKA "sub";                                                         Fd "O,S,Z,A,P,C" >]  Sub             (* subtract byte *)
  (* I_R8sR8ds *)   | [<                AKA "xor";                                                         Fd "O,S,Z,A,P,C" >]  Xor             (* logical exclusive-OR byte *)
type I_R8sR16ds =   | [<                AKA "rcl";                                  Rs "cl";  Fs "C";      Fd "O,C"         >]  Rclw            (* rotate word left thru carry *)
  (* I_R8sR16ds *)  | [<                AKA "rcr";                                  Rs "cl";  Fs "C";      Fd "O,C"         >]  Rcrw            (* rotate word right thru carry *)
  (* I_R8sR16ds *)  | [<                AKA "rol";                                  Rs "cl";               Fd "O,C"         >]  Rolw            (* rotate word left *)
  (* I_R8sR16ds *)  | [<                AKA "ror";                                  Rs "cl";               Fd "O,C"         >]  Rorw            (* rotate word right *)
  (* I_R8sR16ds *)  | [<                AKA "sar";                                  Rs "cl";               Fd "O,S,Z,A,P,C" >]  Sarw            (* shift word right arithmetic *)
  (* I_R8sR16ds *)  | [<                AKA "shl,sal,salw";                         Rs "cl";               Fd "O,S,Z,A,P,C" >]  Shlw            (* shift word left logical *)
  (* I_R8sR16ds *)  | [<                AKA "shr";                                  Rs "cl";               Fd "O,S,Z,A,P,C" >]  Shrw            (* shift word right logical *)
type I_R8sR32ds =   | [<                AKA "rcl";                                  Rs "cl";  Fs "C";      Fd "O,C"         >]  Rcll            (* rotate long left thru carry *)
  (* I_R8sR32ds *)  | [<                AKA "rcr";                                  Rs "cl";  Fs "C";      Fd "O,C"         >]  Rcrl            (* rotate long right thru carry *)
  (* I_R8sR32ds *)  | [<                AKA "rol";                                  Rs "cl";               Fd "O,C"         >]  Roll            (* rotate long left *)
  (* I_R8sR32ds *)  | [<                AKA "ror";                                  Rs "cl";               Fd "O,C"         >]  Rorl            (* rotate long right *)
  (* I_R8sR32ds *)  | [<                AKA "sar";                                  Rs "cl";               Fd "O,S,Z,A,P,C" >]  Sarl            (* shift long right arithmetic *)
  (* I_R8sR32ds *)  | [<                AKA "shl,sal,sall";                         Rs "cl";               Fd "O,S,Z,A,P,C" >]  Shll            (* shift long left logical *)
  (* I_R8sR32ds *)  | [<                AKA "shr";                                  Rs "cl";               Fd "O,S,Z,A,P,C" >]  Shrl            (* shift long right logical *)
type I_R8sR64ds =   | [<                AKA "rcl";                                  Rs "cl";  Fs "C";      Fd "O,C"         >]  Rclq            (* rotate quad left thru carry *)
  (* I_R8sR64ds *)  | [<                AKA "rcr";                                  Rs "cl";  Fs "C";      Fd "O,C"         >]  Rcrq            (* rotate quad right thru carry *)
  (* I_R8sR64ds *)  | [<                AKA "rol";                                  Rs "cl";               Fd "O,C"         >]  Rolq            (* rotate quad left *)
  (* I_R8sR64ds *)  | [<                AKA "ror";                                  Rs "cl";               Fd "O,C"         >]  Rorq            (* rotate quad right *)
  (* I_R8sR64ds *)  | [<                AKA "sar";                                  Rs "cl";               Fd "O,S,Z,A,P,C" >]  Sarq            (* shift quad right arithmetic *)
  (* I_R8sR64ds *)  | [<                AKA "shl,sal,salq";                         Rs "cl";               Fd "O,S,Z,A,P,C" >]  Shlq            (* shift quad left logical *)
  (* I_R8sR64ds *)  | [<                AKA "shr";                                  Rs "cl";               Fd "O,S,Z,A,P,C" >]  Shrq            (* shift quad right logical *)
type I_R16sR16ds =  | [<                AKA "adc";                                            Fs "C";      Fd "O,S,Z,A,P,C" >]  Adcw            (* add word w/carry *)
  (* I_R16sR16ds *) | [<                AKA "add";                                                         Fd "O,S,Z,A,P,C" >]  Addw            (* add word *)
  (* I_R16sR16ds *) | [<                AKA "and";                                                         Fd "O,S,Z,A,P,C" >]  Andw            (* logical AND word *)
  (* I_R16sR16ds *) | [<                AKA "btc";                                                         Fd "C"           >]  Btcw            (* test bit in word and complement *)
  (* I_R16sR16ds *) | [<                AKA "btr";                                                         Fd "C"           >]  Btrw            (* test bit in word and reset *)
  (* I_R16sR16ds *) | [<                AKA "bts";                                                         Fd "C"           >]  Btsw            (* test bit in word and set *)
  (* I_R16sR16ds *) | [< ISA "SSE4_2";  AKA "crc32"                                                                         >]  Crc32w          (* wordwise accumulate CRC32 value *)
  (* I_R16sR16ds *) | [<                AKA "imul";                                                        Fd "O,S,Z,A,P,C" >]  Imulw           (* signed multiply word *)
  (* I_R16sR16ds *) | [<                AKA "or";                                                          Fd "O,S,Z,A,P,C" >]  Orw             (* logical inclusive-OR word *)
  (* I_R16sR16ds *) | [<                AKA "sbb";                                            Fs "C";      Fd "O,S,Z,A,P,C" >]  Sbbw            (* subtract word w/borrow *)
  (* I_R16sR16ds *) | [<                AKA "shld";                                                        Fd "O,S,Z,A,P,C" >]  Shldw of Lit    (* shift left double-width word *)
  (* I_R16sR16ds *) | [<                AKA "shrd";                                                        Fd "O,S,Z,A,P,C" >]  Shrdw of Lit    (* shift right double-width word *)
  (* I_R16sR16ds *) | [<                AKA "sub";                                                         Fd "O,S,Z,A,P,C" >]  Subw            (* subtract word *)
  (* I_R16sR16ds *) | [<                AKA "xor";                                                         Fd "O,S,Z,A,P,C" >]  Xorw            (* logical exclusive-OR word *)
type I_R32sR32ds =  | [<                AKA "adc";                                            Fs "C";      Fd "O,S,Z,A,P,C" >]  Adcl            (* add long w/carry *)
  (* I_R32sR32ds *) | [<                AKA "add";                                                         Fd "O,S,Z,A,P,C" >]  Addl            (* add long *)
  (* I_R32sR32ds *) | [<                AKA "and";                                                         Fd "O,S,Z,A,P,C" >]  Andl            (* logical AND long *)
  (* I_R32sR32ds *) | [<                AKA "btc";                                                         Fd "C"           >]  Btcl            (* test bit in long and complement *)
  (* I_R32sR32ds *) | [<                AKA "btr";                                                         Fd "C"           >]  Btrl            (* test bit in long and reset *)
  (* I_R32sR32ds *) | [<                AKA "bts";                                                         Fd "C"           >]  Btsl            (* test bit in long and set *)
  (* I_R32sR32ds *) | [< ISA "SSE4_2";  AKA "crc32"                                                                         >]  Crc32l          (* longwise accumulate CRC32 value *)
  (* I_R32sR32ds *) | [<                AKA "imul";                                                        Fd "O,S,Z,A,P,C" >]  Imull           (* signed multiply long *)
  (* I_R32sR32ds *) | [<                AKA "or";                                                          Fd "O,S,Z,A,P,C" >]  Orl             (* logical inclusive-OR long *)
  (* I_R32sR32ds *) | [<                AKA "sbb";                                            Fs "C";      Fd "O,S,Z,A,P,C" >]  Sbbl            (* subtract long w/borrow *)
  (* I_R32sR32ds *) | [<                AKA "shld";                                                        Fd "O,S,Z,A,P,C" >]  Shldl of Lit    (* shift left double-width long *)
  (* I_R32sR32ds *) | [<                AKA "shrd";                                                        Fd "O,S,Z,A,P,C" >]  Shrdl of Lit    (* shift right double-width long *)
  (* I_R32sR32ds *) | [<                AKA "sub";                                                         Fd "O,S,Z,A,P,C" >]  Subl            (* subtract long *)
  (* I_R32sR32ds *) | [<                AKA "xor";                                                         Fd "O,S,Z,A,P,C" >]  Xorl            (* logical exclusive-OR long *)
  (* I_R32sR32ds *) | [<                AKA "cmova,cmovnbe,cmovnbel";                         Fs "Z,C"                      >]  Cmoval          (* copy long if above *)
  (* I_R32sR32ds *) | [<                AKA "cmovc,cmovb,cmovnae,cmovbl,cmovnael";            Fs "C"                        >]  Cmovcl          (* copy long if carry *)
  (* I_R32sR32ds *) | [<                AKA "cmovg,cmovnle,cmovnlel";                         Fs "Z,S,O"                    >]  Cmovgl          (* copy long if greater *)
  (* I_R32sR32ds *) | [<                AKA "cmovl,cmovnge,cmovngel";                         Fs "S,O"                      >]  Cmovll          (* copy long if less *)
  (* I_R32sR32ds *) | [<                AKA "cmovna,cmovbe,cmovbel";                          Fs "Z,C"                      >]  Cmovnal         (* copy long if not above *)
  (* I_R32sR32ds *) | [<                AKA "cmovnc,cmovnb,cmovae,cmovnbl,cmovael";           Fs "C"                        >]  Cmovncl         (* copy long if not carry *)
  (* I_R32sR32ds *) | [<                AKA "cmovng,cmovle,cmovlel";                          Fs "Z,S,O"                    >]  Cmovngl         (* copy long if not greater *)
  (* I_R32sR32ds *) | [<                AKA "cmovnl,cmovge,cmovgel";                          Fs "S,O"                      >]  Cmovnll         (* copy long if not less *)
  (* I_R32sR32ds *) | [<                AKA "cmovnp,cmovpo,cmovpol";                          Fs "P"                        >]  Cmovnpl         (* copy long if parity odd *)
  (* I_R32sR32ds *) | [<                AKA "cmovp,cmovpe,cmovpel";                           Fs "P"                        >]  Cmovpl          (* copy long if parity even *)
  (* I_R32sR32ds *) | [<                AKA "cmovz,cmove,cmovel";                             Fs "Z"                        >]  Cmovzl          (* copy long if zero *)
  (* I_R32sR32ds *) | [<                AKA "cmovnz,cmovne,cmovnel";                          Fs "Z"                        >]  Cmovnzl         (* copy long if not zero *)
  (* I_R32sR32ds *) | [<                AKA "cmovno";                                         Fs "O"                        >]  Cmovnol         (* copy long if not overflow *)
  (* I_R32sR32ds *) | [<                AKA "cmovns";                                         Fs "S"                        >]  Cmovnsl         (* copy long if not sign *)
  (* I_R32sR32ds *) | [<                AKA "cmovo";                                          Fs "O"                        >]  Cmovol          (* copy long if overflow *)
  (* I_R32sR32ds *) | [<                AKA "cmovs";                                          Fs "S"                        >]  Cmovsl          (* copy long if sign *)
type I_R64sR64ds =  | [<                AKA "adc";                                            Fs "C";      Fd "O,S,Z,A,P,C" >]  Adcq            (* add quad w/carry *)
  (* I_R64sR64ds *) | [<                AKA "add";                                                         Fd "O,S,Z,A,P,C" >]  Addq            (* add quad *)
  (* I_R64sR64ds *) | [<                AKA "and";                                                         Fd "O,S,Z,A,P,C" >]  Andq            (* logical AND quad *)
  (* I_R64sR64ds *) | [<                AKA "btc";                                                         Fd "C"           >]  Btcq            (* test bit in quad and complement *)
  (* I_R64sR64ds *) | [<                AKA "btr";                                                         Fd "C"           >]  Btrq            (* test bit in quad and reset *)
  (* I_R64sR64ds *) | [<                AKA "bts";                                                         Fd "C"           >]  Btsq            (* test bit in quad and set *)
  (* I_R64sR64ds *) | [< ISA "SSE4_2";  AKA "crc32"                                                                         >]  Crc32q          (* accumulate CRC32 value *)
  (* I_R64sR64ds *) | [<                AKA "imul";                                                        Fd "O,S,Z,A,P,C" >]  Imulq           (* signed multiply quad *)
  (* I_R64sR64ds *) | [<                AKA "or";                                                          Fd "O,S,Z,A,P,C" >]  Orq             (* logical inclusive-OR quad *)
  (* I_R64sR64ds *) | [<                AKA "sbb";                                            Fs "C";      Fd "O,S,Z,A,P,C" >]  Sbbq            (* subtract quad w/borrow *)
  (* I_R64sR64ds *) | [<                AKA "shld";                                                        Fd "O,S,Z,A,P,C" >]  Shldq of Lit    (* shift left double-width quad *)
  (* I_R64sR64ds *) | [<                AKA "shrd";                                                        Fd "O,S,Z,A,P,C" >]  Shrdq of Lit    (* shift right double-width quad *)
  (* I_R64sR64ds *) | [<                AKA "sub";                                                         Fd "O,S,Z,A,P,C" >]  Subq            (* subtract quad *)
  (* I_R64sR64ds *) | [<                AKA "xor";                                                         Fd "O,S,Z,A,P,C" >]  Xorq            (* logical exclusive-OR quad *)
  (* I_R64sR64ds *) | [<                AKA "cmova,cmovnbe,cmovnbeq";                         Fs "Z,C"                      >]  Cmovaq          (* copy quad if above *)
  (* I_R64sR64ds *) | [<                AKA "cmovc,cmovb,cmovnae,cmovbq,cmovnaeq";            Fs "C"                        >]  Cmovcq          (* copy quad if carry *)
  (* I_R64sR64ds *) | [<                AKA "cmovg,cmovnle,cmovnleq";                         Fs "Z,S,O"                    >]  Cmovgq          (* copy quad if greater *)
  (* I_R64sR64ds *) | [<                AKA "cmovl,cmovnge,cmovngeq";                         Fs "S,O"                      >]  Cmovlq          (* copy quad if less *)
  (* I_R64sR64ds *) | [<                AKA "cmovna,cmovbe,cmovbeq";                          Fs "Z,C"                      >]  Cmovnaq         (* copy quad if not above *)
  (* I_R64sR64ds *) | [<                AKA "cmovnc,cmovnb,cmovae,cmovnbq,cmovaeq";           Fs "C"                        >]  Cmovncq         (* copy quad if not carry *)
  (* I_R64sR64ds *) | [<                AKA "cmovng,cmovle,cmovleq";                          Fs "Z,S,O"                    >]  Cmovngq         (* copy quad if not greater *)
  (* I_R64sR64ds *) | [<                AKA "cmovnl,cmovge,cmovgeq";                          Fs "S,O"                      >]  Cmovnlq         (* copy quad if not less *)
  (* I_R64sR64ds *) | [<                AKA "cmovnp,cmovpo,cmovpoq";                          Fs "P"                        >]  Cmovnpq         (* copy quad if parity odd *)
  (* I_R64sR64ds *) | [<                AKA "cmovp,cmovpe,cmovpeq";                           Fs "P"                        >]  Cmovpq          (* copy quad if parity even *)
  (* I_R64sR64ds *) | [<                AKA "cmovz,cmove,cmoveq";                             Fs "Z"                        >]  Cmovzq          (* copy quad if zero *)
  (* I_R64sR64ds *) | [<                AKA "cmovnz,cmovne,cmovneq";                          Fs "Z"                        >]  Cmovnzq         (* copy quad if not zero *)
  (* I_R64sR64ds *) | [<                AKA "cmovno";                                         Fs "O"                        >]  Cmovnoq         (* copy quad if not overflow *)
  (* I_R64sR64ds *) | [<                AKA "cmovns";                                         Fs "S"                        >]  Cmovnsq         (* copy quad if not sign *)
  (* I_R64sR64ds *) | [<                AKA "cmovo";                                          Fs "O"                        >]  Cmovoq          (* copy quad if overflow *)
  (* I_R64sR64ds *) | [<                AKA "cmovs";                                          Fs "S"                        >]  Cmovsq          (* copy quad if sign *)

(*****************************************************************************************************************************************************************************)

type I_M8sR8ds =    | [<                AKA "adc";                                   Fs "C";      Fd "O,S,Z,A,P,C" >]  Adcb       (* add byte w/carry *)
  (* I_M8sR8ds *)   | [<                AKA "add";                                                Fd "O,S,Z,A,P,C" >]  Addb       (* add byte *)
  (* I_M8sR8ds *)   | [<                AKA "and";                                                Fd "O,S,Z,A,P,C" >]  Andb       (* logical AND byte *)
  (* I_M8sR8ds *)   | [<                AKA "or";                                                 Fd "O,S,Z,A,P,C" >]  Orb        (* logical inclusive-OR byte *)
  (* I_M8sR8ds *)   | [<                AKA "sbb";                                   Fs "C";      Fd "O,S,Z,A,P,C" >]  Sbbb       (* subtract byte w/borrow *)
  (* I_M8sR8ds *)   | [<                AKA "sub";                                                Fd "O,S,Z,A,P,C" >]  Subb       (* subtract byte *)
  (* I_M8sR8ds *)   | [<                AKA "xor";                                                Fd "O,S,Z,A,P,C" >]  Xorb       (* logical exclusive-OR byte *)
  (* I_M8sR8ds *)   | [< ISA "SSE4_2"                                                                              >]  Crc32b     (* accumulate CRC32 value *)
type I_M16sR16ds =  | [<                AKA "adc";                                   Fs "C";      Fd "O,S,Z,A,P,C" >]  Adcw       (* add word w/carry *)
  (* I_M16sR16ds *) | [<                AKA "add";                                                Fd "O,S,Z,A,P,C" >]  Addw       (* add word *)
  (* I_M16sR16ds *) | [<                AKA "and";                                                Fd "O,S,Z,A,P,C" >]  Andw       (* logical AND word *)
  (* I_M16sR16ds *) | [<                AKA "imul";                                               Fd "O,S,Z,A,P,C" >]  Imulw      (* signed multiply word *)
  (* I_M16sR16ds *) | [<                AKA "or";                                                 Fd "O,S,Z,A,P,C" >]  Orw        (* logical inclusive-OR word *)
  (* I_M16sR16ds *) | [<                AKA "sbb";                                   Fs "C";      Fd "O,S,Z,A,P,C" >]  Sbbw       (* subtract word w/borrow *)
  (* I_M16sR16ds *) | [<                AKA "sub";                                                Fd "O,S,Z,A,P,C" >]  Subw       (* subtract word *)
  (* I_M16sR16ds *) | [<                AKA "xor";                                                Fd "O,S,Z,A,P,C" >]  Xorw       (* logical exclusive-OR word *)
  (* I_M16sR16ds *) | [< ISA "SSE4_2"                                                                              >]  Crc32w     (* accumulate CRC32 value *)
type I_M32sR32ds =  | [<                AKA "adc";                                   Fs "C";      Fd "O,S,Z,A,P,C" >]  Adcl       (* add long w/carry *)
  (* I_M32sR32ds *) | [<                AKA "add";                                                Fd "O,S,Z,A,P,C" >]  Addl       (* add long *)
  (* I_M32sR32ds *) | [<                AKA "and";                                                Fd "O,S,Z,A,P,C" >]  Andl       (* logical AND long *)
  (* I_M32sR32ds *) | [<                AKA "cmova,cmovnbe,cmovnbel";                Fs "Z,C"                      >]  Cmoval     (* load long if above *)
  (* I_M32sR32ds *) | [<                AKA "cmovc,cmovb,cmovnae,cmovbl,cmovnael";   Fs "C"                        >]  Cmovcl     (* load long if carry *)
  (* I_M32sR32ds *) | [<                AKA "cmovg,cmovnle,cmovnlel";                Fs "Z,S,O"                    >]  Cmovgl     (* load long if greater *)
  (* I_M32sR32ds *) | [<                AKA "cmovl,cmovnge,cmovngel";                Fs "S,O"                      >]  Cmovll     (* load long if less *)
  (* I_M32sR32ds *) | [<                AKA "cmovna,cmovbe,cmovbel";                 Fs "Z,C"                      >]  Cmovnal    (* load long if not above *)
  (* I_M32sR32ds *) | [<                AKA "cmovnc,cmovnb,cmovae,cmovnbl,cmovael";  Fs "C"                        >]  Cmovncl    (* load long if not carry *)
  (* I_M32sR32ds *) | [<                AKA "cmovng,cmovle,cmovlel";                 Fs "Z,S,O"                    >]  Cmovngl    (* load long if not greater *)
  (* I_M32sR32ds *) | [<                AKA "cmovnl,cmovge,cmovgel";                 Fs "S,O"                      >]  Cmovnll    (* load long if not less *)
  (* I_M32sR32ds *) | [<                AKA "cmovno";                                Fs "O"                        >]  Cmovnol    (* load long if not overflow *)
  (* I_M32sR32ds *) | [<                AKA "cmovnp,cmovpo,cmovpol";                 Fs "P"                        >]  Cmovnpl    (* load long if parity odd  *)
  (* I_M32sR32ds *) | [<                AKA "cmovns";                                Fs "S"                        >]  Cmovnsl    (* load long if not sign *)
  (* I_M32sR32ds *) | [<                AKA "cmovnz,cmovne,cmovnel";                 Fs "Z"                        >]  Cmovnzl    (* load long if not zero *)
  (* I_M32sR32ds *) | [<                AKA "cmovo";                                 Fs "O"                        >]  Cmovol     (* load long if overflow *)
  (* I_M32sR32ds *) | [<                AKA "cmovp,cmovpe,cmovpel";                  Fs "P"                        >]  Cmovpl     (* load long if parity even *)
  (* I_M32sR32ds *) | [<                AKA "cmovs";                                 Fs "S"                        >]  Cmovsl     (* load long if sign *)
  (* I_M32sR32ds *) | [<                AKA "cmovz,cmove,cmovel";                    Fs "Z"                        >]  Cmovzl     (* load long if zero *)
  (* I_M32sR32ds *) | [<                AKA "imul";                                               Fd "O,S,Z,A,P,C" >]  Imull      (* signed multiply long *)
  (* I_M32sR32ds *) | [<                AKA "or";                                                 Fd "O,S,Z,A,P,C" >]  Orl        (* logical inclusive-OR long *)
  (* I_M32sR32ds *) | [<                AKA "sbb";                                   Fs "C";      Fd "O,S,Z,A,P,C" >]  Sbbl       (* subtract long w/borrow *)
  (* I_M32sR32ds *) | [<                AKA "sub";                                                Fd "O,S,Z,A,P,C" >]  Subl       (* subtract long *)
  (* I_M32sR32ds *) | [<                AKA "xor";                                                Fd "O,S,Z,A,P,C" >]  Xorl       (* logical exclusive-OR long *)
  (* I_M32sR32ds *) | [< ISA "SSE4_2"                                                                              >]  Crc32l     (* accumulate CRC32 value *)
type I_M64sR64ds =  | [<                AKA "adc";                                   Fs "C";      Fd "O,S,Z,A,P,C" >]  Adcq       (* add quad w/carry *)
  (* I_M64sR64ds *) | [<                AKA "add";                                                Fd "O,S,Z,A,P,C" >]  Addq       (* add quad *)
  (* I_M64sR64ds *) | [<                AKA "and";                                                Fd "O,S,Z,A,P,C" >]  Andq       (* logical AND quad *)
  (* I_M64sR64ds *) | [<                AKA "cmova,cmovnbe,cmovnbeq";                Fs "Z,C"                      >]  Cmovaq     (* load quad if above *)
  (* I_M64sR64ds *) | [<                AKA "cmovc,cmovb,cmovnae,cmovbq,cmovnaeq";   Fs "C"                        >]  Cmovcq     (* load quad if carry *)
  (* I_M64sR64ds *) | [<                AKA "cmovg,cmovnle,cmovnleq";                Fs "Z,S,O"                    >]  Cmovgq     (* load quad if greater *)
  (* I_M64sR64ds *) | [<                AKA "cmovl,cmovnge,cmovngeq";                Fs "S,O"                      >]  Cmovlq     (* load quad if less *)
  (* I_M64sR64ds *) | [<                AKA "cmovna,cmovbe,cmovbeq";                 Fs "Z,C"                      >]  Cmovnaq    (* load quad if not above *)
  (* I_M64sR64ds *) | [<                AKA "cmovnc,cmovnb,cmovae,cmovnbq,cmovaeq";  Fs "C"                        >]  Cmovncq    (* load quad if not carry *)
  (* I_M64sR64ds *) | [<                AKA "cmovng,cmovle,cmovleq";                 Fs "Z,S,O"                    >]  Cmovngq    (* load quad if not greater *)
  (* I_M64sR64ds *) | [<                AKA "cmovnl,cmovge,cmovgeq";                 Fs "S,O"                      >]  Cmovnlq    (* load quad if not less *) 
  (* I_M64sR64ds *) | [<                AKA "cmovno";                                Fs "O"                        >]  Cmovnoq    (* load quad if not overflow *)
  (* I_M64sR64ds *) | [<                AKA "cmovnp,cmovpo,cmovpoq";                 Fs "P"                        >]  Cmovnpq    (* load quad if parity odd *)
  (* I_M64sR64ds *) | [<                AKA "cmovns";                                Fs "S"                        >]  Cmovnsq    (* load quad if not sign *)
  (* I_M64sR64ds *) | [<                AKA "cmovnz,cmovne,cmovneq";                 Fs "Z"                        >]  Cmovnzq    (* load quad if not zero *)
  (* I_M64sR64ds *) | [<                AKA "cmovo";                                 Fs "O"                        >]  Cmovoq     (* load quad if overflow *)
  (* I_M64sR64ds *) | [<                AKA "cmovp,cmovpe,cmovpeq";                  Fs "P"                        >]  Cmovpq     (* load quad if parity even *)
  (* I_M64sR64ds *) | [<                AKA "cmovs";                                 Fs "S"                        >]  Cmovsq     (* load quad if sign *)
  (* I_M64sR64ds *) | [<                AKA "cmovz,cmove,cmoveq";                    Fs "Z"                        >]  Cmovzq     (* load quad if zero *)
  (* I_M64sR64ds *) | [<                AKA "imul";                                               Fd "O,S,Z,A,P,C" >]  Imulq      (* signed multiply quad *)
  (* I_M64sR64ds *) | [<                AKA "or";                                                 Fd "O,S,Z,A,P,C" >]  Orq        (* logical inclusive-OR quad *)
  (* I_M64sR64ds *) | [<                AKA "sbb";                                   Fs "C";      Fd "O,S,Z,A,P,C" >]  Sbbq       (* subtract quad w/borrow *)
  (* I_M64sR64ds *) | [<                AKA "sub";                                                Fd "O,S,Z,A,P,C" >]  Subq       (* subtract quad *)
  (* I_M64sR64ds *) | [<                AKA "xor";                                                Fd "O,S,Z,A,P,C" >]  Xorq       (* logical exclusive-OR quad *)
  (* I_M64sR64ds *) | [< ISA "SSE4_2"                                                                              >]  Crc32q     (* accumulate CRC32 value *)

(*****************************************************************************************************************************************************************************)

type I_R8sM8ds =    | [< AKA "adc";                                 Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcb            (* add byte w/carry *)
  (* I_R8sM8ds *)   | [< AKA "add";                                          Fd "O,S,Z,A,P,C" >]  Addb            (* add byte *)
  (* I_R8sM8ds *)   | [< AKA "and";                                          Fd "O,S,Z,A,P,C" >]  Andb            (* logical AND byte *)
  (* I_R8sM8ds *)   | [< AKA "or";                                           Fd "O,S,Z,A,P,C" >]  Orb             (* logical inclusive-OR byte *)
  (* I_R8sM8ds *)   | [< AKA "rcl";           Rs "cl";              Fs "C";  Fd "O,C"         >]  Rclb            (* rotate byte left thru carry *)
  (* I_R8sM8ds *)   | [< AKA "rcr";           Rs "cl";              Fs "C";  Fd "O,C"         >]  Rcrb            (* rotate byte right thru carry *)
  (* I_R8sM8ds *)   | [< AKA "rol";           Rs "cl";                       Fd "O,C"         >]  Rolb            (* rotate byte left *)
  (* I_R8sM8ds *)   | [< AKA "ror";           Rs "cl";                       Fd "O,C"         >]  Rorb            (* rotate byte right *)
  (* I_R8sM8ds *)   | [< AKA "sar";           Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Sarb            (* shift byte right arithmetic *)
  (* I_R8sM8ds *)   | [< AKA "sbb";                                 Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbb            (* subtract byte w/borrow *)
  (* I_R8sM8ds *)   | [< AKA "shl,sal,salb";  Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Shlb            (* shift byte left logical *)
  (* I_R8sM8ds *)   | [< AKA "shr";           Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Shrb            (* shift byte right logical *)
  (* I_R8sM8ds *)   | [< AKA "sub";                                          Fd "O,S,Z,A,P,C" >]  Subb            (* subtract byte *)
  (* I_R8sM8ds *)   | [< AKA "xor";                                          Fd "O,S,Z,A,P,C" >]  Xorb            (* logical exclusive-OR byte *)
type I_R16sM16ds =  | [< AKA "adc";                                 Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcw            (* add word w/carry *)
  (* I_R16sM16ds *) | [< AKA "add";                                          Fd "O,S,Z,A,P,C" >]  Addw            (* add word *)
  (* I_R16sM16ds *) | [< AKA "and";                                          Fd "O,S,Z,A,P,C" >]  Andw            (* logical AND word *)
  (* I_R16sM16ds *) | [< AKA "btc";                                          Fd "C"           >]  Btcw            (* test bit and complement *)
  (* I_R16sM16ds *) | [< AKA "btr";                                          Fd "C"           >]  Btrw            (* test bit and reset *)
  (* I_R16sM16ds *) | [< AKA "bts";                                          Fd "C"           >]  Btsw            (* test bit and set *)
  (* I_R16sM16ds *) | [< AKA "or";                                           Fd "O,S,Z,A,P,C" >]  Orw             (* logical inclusive-OR word *)
  (* I_R16sM16ds *) | [< AKA "rcl";           Rs "cl";              Fs "C";  Fd "O,C"         >]  Rclw            (* rotate word left thru carry *)
  (* I_R16sM16ds *) | [< AKA "rcr";           Rs "cl";              Fs "C";  Fd "O,C"         >]  Rcrw            (* rotate word right thru carry *)
  (* I_R16sM16ds *) | [< AKA "rol";           Rs "cl";                       Fd "O,C"         >]  Rolw            (* rotate word left *)
  (* I_R16sM16ds *) | [< AKA "ror";           Rs "cl";                       Fd "O,C"         >]  Rorw            (* rotate word right *)
  (* I_R16sM16ds *) | [< AKA "sar";           Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Sarw            (* shift word right arithmetic *)
  (* I_R16sM16ds *) | [< AKA "sbb";                                 Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbw            (* subtract word w/borrow *)
  (* I_R16sM16ds *) | [< AKA "shl,sal,salw";  Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Shlw            (* shift word left logical *)
  (* I_R16sM16ds *) | [< AKA "shld";                                         Fd "O,S,Z,A,P,C" >]  Shldw of Lit    (* shift left double-width word *)
  (* I_R16sM16ds *) | [< AKA "shrd";                                         Fd "O,S,Z,A,P,C" >]  Shrdw of Lit    (* shift right double-width word *)
  (* I_R16sM16ds *) | [< AKA "shr";           Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Shrw            (* shift word right logical *)
  (* I_R16sM16ds *) | [< AKA "sub";                                          Fd "O,S,Z,A,P,C" >]  Subw            (* subtract word *)
  (* I_R16sM16ds *) | [< AKA "xor";                                          Fd "O,S,Z,A,P,C" >]  Xorw            (* logical exclusive-OR word *)
type I_R32sM32ds =  | [< AKA "adc";                                 Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcl            (* add long w/carry *)
  (* I_R32sM32ds *) | [< AKA "add";                                          Fd "O,S,Z,A,P,C" >]  Addl            (* add long *)
  (* I_R32sM32ds *) | [< AKA "and";                                          Fd "O,S,Z,A,P,C" >]  Andl            (* logical AND long *)
  (* I_R32sM32ds *) | [< AKA "btc";                                          Fd "C"           >]  Btcl            (* test bit and complement *)
  (* I_R32sM32ds *) | [< AKA "btr";                                          Fd "C"           >]  Btrl            (* test bit and reset *)
  (* I_R32sM32ds *) | [< AKA "bts";                                          Fd "C"           >]  Btsl            (* test bit and set *)
  (* I_R32sM32ds *) | [< AKA "cmpxchg";                 Rds "eax";           Fd "O,S,Z,A,P,C" >]  Cmpxchgl        (* compare and exchange long *)
  (* I_R32sM32ds *) | [< AKA "or";                                           Fd "O,S,Z,A,P,C" >]  Orl             (* logical inclusive-OR long *)
  (* I_R32sM32ds *) | [< AKA "rcl";           Rs "cl";              Fs "C";  Fd "O,C"         >]  Rcll            (* rotate long left thru carry *)
  (* I_R32sM32ds *) | [< AKA "rcr";           Rs "cl";              Fs "C";  Fd "O,C"         >]  Rcrl            (* rotate long right thru carry *)
  (* I_R32sM32ds *) | [< AKA "rol";           Rs "cl";                       Fd "O,C"         >]  Roll            (* rotate long left *)
  (* I_R32sM32ds *) | [< AKA "ror";           Rs "cl";                       Fd "O,C"         >]  Rorl            (* rotate long right *)
  (* I_R32sM32ds *) | [< AKA "sar";           Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Sarl            (* shift long right arithmetic *)
  (* I_R32sM32ds *) | [< AKA "sbb";                                 Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbl            (* subtract long w/borrow *)
  (* I_R32sM32ds *) | [< AKA "shl,sal,sall";  Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Shll            (* shift long left logical *)
  (* I_R32sM32ds *) | [< AKA "shld";                                         Fd "O,S,Z,A,P,C" >]  Shldl of Lit    (* shift left double-width long *)
  (* I_R32sM32ds *) | [< AKA "shrd";                                         Fd "O,S,Z,A,P,C" >]  Shrdl of Lit    (* shift right double-width long *)
  (* I_R32sM32ds *) | [< AKA "shr";           Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Shrl            (* shift long right logical *)
  (* I_R32sM32ds *) | [< AKA "sub";                                          Fd "O,S,Z,A,P,C" >]  Subl            (* subtract long *)
  (* I_R32sM32ds *) | [< AKA "xor";                                          Fd "O,S,Z,A,P,C" >]  Xorl            (* logical exclusive-OR long *)
type I_R64sM64ds =  | [< AKA "adc";                                 Fs "C";  Fd "O,S,Z,A,P,C" >]  Adcq            (* add quad w/carry *)
  (* I_R64sM64ds *) | [< AKA "add";                                          Fd "O,S,Z,A,P,C" >]  Addq            (* add quad *)
  (* I_R64sM64ds *) | [< AKA "and";                                          Fd "O,S,Z,A,P,C" >]  Andq            (* logical AND quad *)
  (* I_R64sM64ds *) | [< AKA "btc";                                          Fd "C"           >]  Btcq            (* test bit and complement *)
  (* I_R64sM64ds *) | [< AKA "btr";                                          Fd "C"           >]  Btrq            (* test bit and reset *)
  (* I_R64sM64ds *) | [< AKA "bts";                                          Fd "C"           >]  Btsq            (* test bit and set *)
  (* I_R64sM64ds *) | [< AKA "cmpxchg";                 Rds "rax";           Fd "O,S,Z,A,P,C" >]  Cmpxchgq        (* compare and exchange quad *) 
  (* I_R64sM64ds *) | [< AKA "or";                                           Fd "O,S,Z,A,P,C" >]  Orq             (* logical inclusive-OR quad *)
  (* I_R64sM64ds *) | [< AKA "rcl";           Rs "cl";              Fs "C";  Fd "O,C"         >]  Rclq            (* rotate quad left thru carry *)
  (* I_R64sM64ds *) | [< AKA "rcr";           Rs "cl";              Fs "C";  Fd "O,C"         >]  Rcrq            (* rotate quad right thru carry *)
  (* I_R64sM64ds *) | [< AKA "rol";           Rs "cl";                       Fd "O,C"         >]  Rolq            (* rotate quad left *)
  (* I_R64sM64ds *) | [< AKA "ror";           Rs "cl";                       Fd "O,C"         >]  Rorq            (* rotate quad right *)
  (* I_R64sM64ds *) | [< AKA "sar";           Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Sarq            (* shift quad right arithmetic *)
  (* I_R64sM64ds *) | [< AKA "sbb";                                 Fs "C";  Fd "O,S,Z,A,P,C" >]  Sbbq            (* subtract quad w/borrow *)
  (* I_R64sM64ds *) | [< AKA "shl,sal,salq";  Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Shlq            (* shift quad left logical *)
  (* I_R64sM64ds *) | [< AKA "shld";                                         Fd "O,S,Z,A,P,C" >]  Shldq of Lit    (* shift left double-width quad *)
  (* I_R64sM64ds *) | [< AKA "shrd";                                         Fd "O,S,Z,A,P,C" >]  Shrdq of Lit    (* shift right double-width quad *)
  (* I_R64sM64ds *) | [< AKA "shr";           Rs "cl";                       Fd "O,S,Z,A,P,C" >]  Shrq            (* shift quad right logical *)
  (* I_R64sM64ds *) | [< AKA "sub";                                          Fd "O,S,Z,A,P,C" >]  Subq            (* subtract quad *)
  (* I_R64sM64ds *) | [< AKA "xor";                                          Fd "O,S,Z,A,P,C" >]  Xorq            (* logical exclusive-OR quad *)

(*****************************************************************************************************************************************************************************)

(* 128-bit SSE/AVX vector instructions *)
type I_Xds =  | [< ISA "SSE4a" >]  Extrq of Lit * Lit    (* extract field from register *)
  (* I_Xds *) |                    Pslld of Lit          (* shift packed data left logical by <N> bits *)
  (* I_Xds *) |                    Pslldq of Lit         (* shift packed data left logical by <N> bytes *)
  (* I_Xds *) |                    Psllq of Lit          (* shift packed data left logical by <N> bits *)
  (* I_Xds *) |                    Psllw of Lit          (* shift packed data left logical by <N> bits *)
  (* I_Xds *) |                    Psrad of Lit          (* shift packed data right arithmetic by <N> bits *)
  (* I_Xds *) |                    Psraw of Lit          (* shift packed data right arithmetic by <N> bits *)
  (* I_Xds *) |                    Psrld of Lit          (* shift packed data right logical by <N> bits *)
  (* I_Xds *) |                    Psrldq of Lit         (* shift packed data right logical by <N> bytes *)
  (* I_Xds *) |                    Psrlq of Lit          (* shift packed data right logical by <N> bits *)
  (* I_Xds *) |                    Psrlw of Lit          (* shift packed data right logical by <N> bits *)

(*****************************************************************************************************************************************************************************)

type I_XsR32d =  |                                        Movd                  (* move long *)
  (* I_XsR32d *) | [< ISA "AVX"                       >]  Vmovd                 (* move long *)
  (* I_XsR32d *) | [<                AKA "cvtsd2si"   >]  Cvtsd2sil             (* convert float64 to long *)
  (* I_XsR32d *) | [<                AKA "cvttsd2si"  >]  Cvttsd2sil            (* convert w/truncation float64 to long *)
  (* I_XsR32d *) | [<                AKA "cvtss2si"   >]  Cvtss2sil             (* convert float32 to long *)
  (* I_XsR32d *) | [<                AKA "cvttss2si"  >]  Cvttss2sil            (* convert w/truncation float32 to long *)
  (* I_XsR32d *) | [< ISA "AVX";     AKA "vcvtsd2si"  >]  Vcvtsd2sil            (* convert float64 to long *)
  (* I_XsR32d *) | [< ISA "AVX";     AKA "vcvttsd2si" >]  Vcvttsd2sil           (* convert w/truncation float64 to long *)
  (* I_XsR32d *) | [< ISA "AVX";     AKA "vcvtss2si"  >]  Vcvtss2sil            (* convert float32 to long *)
  (* I_XsR32d *) | [< ISA "AVX";     AKA "vcvttss2si" >]  Vcvttss2sil           (* convert w/truncation float32 to long *)
  (* I_XsR32d *) | [<                AKA "extractps"  >]  Extractpsl of Lit     (* extract packed float32 *)
  (* I_XsR32d *) | [< ISA "AVX";     AKA "vextractps" >]  Vextractpsl of Lit    (* extract packed float32 *)
  (* I_XsR32d *) | [<                AKA "movmskpd"   >]  Movmskpdl             (* extract packed float64 sign mask *)
  (* I_XsR32d *) | [<                AKA "movmskps"   >]  Movmskpsl             (* extract packed float32 sign mask *)
  (* I_XsR32d *) | [< ISA "AVX";     AKA "vmovmskpd"  >]  Vmovmskpdl            (* extract packed float64 sign mask *)
  (* I_XsR32d *) | [< ISA "AVX";     AKA "vmovmskps"  >]  Vmovmskpsl            (* extract packed float32 sign mask *)
  (* I_XsR32d *) | [< ISA "SSE4_1";                   >]  Pextrb of Lit         (* extract byte *)
  (* I_XsR32d *) | [< ISA "SSE4_1"                    >]  Pextrd of Lit         (* extract long *)
  (* I_XsR32d *) |                                        Pextrw of Lit         (* extract word *)
  (* I_XsR32d *) | [< ISA "AVX"                       >]  Vpextrb of Lit        (* extract byte *)
  (* I_XsR32d *) | [< ISA "AVX"                       >]  Vpextrd of Lit        (* extract long *)
  (* I_XsR32d *) | [< ISA "AVX"                       >]  Vpextrw of Lit        (* extract word *)
  (* I_XsR32d *) | [<                AKA "pmovmskb"   >]  Pmovmskbl             (* move byte mask *)
  (* I_XsR32d *) | [< ISA "AVX";     AKA "vpmovmskb"  >]  Vpmovmskbl            (* move byte mask *)
type I_XsR64d =  |                                        Movd                  (* move long *)
  (* I_XsR64d *) |                                        Movq                  (* move quad *)
  (* I_XsR64d *) | [< ISA "AVX"                       >]  Vmovd                 (* move long *)
  (* I_XsR64d *) | [< ISA "AVX"                       >]  Vmovq                 (* move quad *)
  (* I_XsR64d *) | [<                AKA "cvtsd2si"   >]  Cvtsd2siq             (* convert float64 to quad *)
  (* I_XsR64d *) | [<                AKA "cvttsd2si"  >]  Cvttsd2siq            (* convert w/truncation float64 to quad *)
  (* I_XsR64d *) | [<                AKA "cvtss2si"   >]  Cvtss2siq             (* convert float32 to quad *)
  (* I_XsR64d *) | [<                AKA "cvttss2si"  >]  Cvttss2siq            (* convert w/truncation float32 to quad *)
  (* I_XsR64d *) | [< ISA "AVX";     AKA "vcvtsd2si"  >]  Vcvtsd2siq            (* convert float64 to quad *)
  (* I_XsR64d *) | [< ISA "AVX";     AKA "vcvttsd2si" >]  Vcvttsd2siq           (* convert w/truncation float64 to quad *)
  (* I_XsR64d *) | [< ISA "AVX";     AKA "vcvtss2si"  >]  Vcvtss2siq            (* convert float32 to quad *)
  (* I_XsR64d *) | [< ISA "AVX";     AKA "vcvttss2si" >]  Vcvttss2siq           (* convert w/truncation float32 to quad *)
  (* I_XsR64d *) | [<                AKA "extractps"  >]  Extractpsq of Lit     (* extract packed float32 *)
  (* I_XsR64d *) | [< ISA "AVX";     AKA "vextractps" >]  Vextractpsq of Lit    (* extract packed float32 *)
  (* I_XsR64d *) | [<                AKA "movmskpd"   >]  Movmskpdq             (* extract packed float64 sign mask *)
  (* I_XsR64d *) | [<                AKA "movmskps"   >]  Movmskpsq             (* extract packed float32 sign mask *)
  (* I_XsR64d *) | [< ISA "AVX";     AKA "vmovmskpd"  >]  Vmovmskpdq            (* extract packed float64 sign mask *)
  (* I_XsR64d *) | [< ISA "AVX";     AKA "vmovmskps"  >]  Vmovmskpsq            (* extract packed float32 sign mask *)
  (* I_XsR64d *) | [< ISA "SSE4_1"                    >]  Pextrb of Lit         (* extract byte *)
  (* I_XsR64d *) | [< ISA "SSE4_1"                    >]  Pextrq of Lit         (* extract quad *)
  (* I_XsR64d *) |                                        Pextrw of Lit         (* extract word *)
  (* I_XsR64d *) | [< ISA "AVX"                       >]  Vpextrb of Lit        (* extract byte *)
  (* I_XsR64d *) | [< ISA "AVX"                       >]  Vpextrq of Lit        (* extract quad *)
  (* I_XsR64d *) | [< ISA "AVX"                       >]  Vpextrw of Lit        (* extract word *)
  (* I_XsR64d *) | [<                AKA "pmovmskb"   >]  Pmovmskbq             (* move byte mask *)
  (* I_XsR64d *) | [< ISA "AVX";     AKA "vpmovmskb"  >]  Vpmovmskbq            (* move byte mask *)

(*****************************************************************************************************************************************************************************)

type I_R8sXsXd =   | [< ISA "AVX"                   >]  Vpinsrb of Lit    (* insert byte into vector *)
type I_R16sXsXd =  | [< ISA "AVX"                   >]  Vpinsrw of Lit    (* insert word into vector *)
type I_R32sXsXd =  | [< ISA "AVX"                   >]  Vpinsrd of Lit    (* insert long into vector *)
  (* I_R32sXsXd *) | [< ISA "AVX";  AKA "vcvtsi2sd" >]  Vcvtsi2sdl        (* convert long to float64 *)
  (* I_R32sXsXd *) | [< ISA "AVX";  AKA "vcvtsi2ss" >]  Vcvtsi2ssl        (* convert long to float32 *)
type I_R64sXsXd =  | [< ISA "AVX"                   >]  Vpinsrq of Lit    (* insert quad into vector *)
  (* I_R64sXsXd *) | [< ISA "AVX";  AKA "vcvtsi2sd" >]  Vcvtsi2sdq        (* convert quad to float64 *)
  (* I_R64sXsXd *) | [< ISA "AVX";  AKA "vcvtsi2ss" >]  Vcvtsi2ssq        (* convert quad to float32 *)

(*****************************************************************************************************************************************************************************)

type I_R8sXds =   | [< ISA "SSE4_1"                  >]  Pinsrb of Lit    (* insert byte into vector *)
type I_R16sXds =  |                                      Pinsrw of Lit    (* insert word into vector *)
type I_R32sXds =  | [<                AKA "cvtsi2sd" >]  Cvtsi2sdl        (* convert long to float64 *)
  (* I_R32sXds *) | [<                AKA "cvtsi2ss" >]  Cvtsi2ssl        (* convert long to float32 *)
  (* I_R32sXds *) | [< ISA "SSE4_1"                  >]  Pinsrd of Lit    (* insert long into vector *)
type I_R64sXds =  | [<                AKA "cvtsi2sd" >]  Cvtsi2sdq        (* convert quad to float64 *)
  (* I_R64sXds *) | [<                AKA "cvtsi2ss" >]  Cvtsi2ssq        (* convert quad to float32 *)
  (* I_R64sXds *) | [< ISA "SSE4_1"                  >]  Pinsrq of Lit    (* insert quad into vector *)

(*****************************************************************************************************************************************************************************)

type I_XsM8d =    | [< ISA "SSE4_1" >]  Pextrb of Lit        (* extract byte *)
  (* I_XsM8d *)   | [< ISA "AVX"    >]  Vpextrb of Lit       (* extract byte *)
type I_XsM16d =   |                     Pextrw of Lit        (* extract word *)
  (* I_XsM16d *)  | [< ISA "AVX"    >]  Vpextrw of Lit       (* extract word *)
type I_XsM32d =   |                     Extractps of Lit     (* extract packed float32 *)
  (* I_XsM32d *)  | [< ISA "AVX"    >]  Vextractps of Lit    (* extract packed float32 *)
  (* I_XsM32d *)  |                     Movd                 (* store long *)
  (* I_XsM32d *)  | [< ISA "AVX"    >]  Vmovd                (* store long *)
  (* I_XsM32d *)  | [< ISA "SSE4a"  >]  Movntss              (* store scalar float32 w/NT-hint *)
  (* I_XsM32d *)  |                     Movss                (* store scalar float32 *)
  (* I_XsM32d *)  | [< ISA "AVX"    >]  Vmovss               (* store scalar float32 *)
  (* I_XsM32d *)  | [< ISA "SSE4_1" >]  Pextrd of Lit        (* extract long *)
  (* I_XsM32d *)  | [< ISA "AVX"    >]  Vpextrd of Lit       (* extract long *)
type I_XsM64d =   |                     Movhpd               (* store data *)
  (* I_XsM64d *)  |                     Movhps               (* store data *)
  (* I_XsM64d *)  |                     Movlpd               (* store data *)
  (* I_XsM64d *)  |                     Movlps               (* store data *)
  (* I_XsM64d *)  |                     Movq                 (* store quad *)
  (* I_XsM64d *)  |                     Movsd                (* store scalar float64 *)
  (* I_XsM64d *)  | [< ISA "AVX"    >]  Vmovhpd              (* store data *)
  (* I_XsM64d *)  | [< ISA "AVX"    >]  Vmovhps              (* store data *)
  (* I_XsM64d *)  | [< ISA "AVX"    >]  Vmovlpd              (* store data *)
  (* I_XsM64d *)  | [< ISA "AVX"    >]  Vmovlps              (* store data *)
  (* I_XsM64d *)  | [< ISA "SSE4a"  >]  Movntsd              (* store scalar float64 w/NT-hint *)
  (* I_XsM64d *)  | [< ISA "AVX"    >]  Vmovq                (* store quad *)
  (* I_XsM64d *)  | [< ISA "AVX"    >]  Vmovsd               (* store scalar float64 *)
  (* I_XsM64d *)  | [< ISA "SSE4_1" >]  Pextrq of Lit        (* extract quad *)
  (* I_XsM64d *)  | [< ISA "AVX"    >]  Vpextrq of Lit       (* extract quad *)
type I_XsM128d =  |                     Movapd               (* aligned store packed float64s *)
  (* I_XsM128d *) |                     Movaps               (* aligned store packed float32s *)
  (* I_XsM128d *) |                     Movdqa               (* store data *)
  (* I_XsM128d *) |                     Movdqu               (* store data *)
  (* I_XsM128d *) |                     Movntdq              (* store double-quad w/NT-hint *)
  (* I_XsM128d *) |                     Movntdqa             (* store double-quad w/NTA-hint *)
  (* I_XsM128d *) |                     Movntpd              (* store packed float64s w/NT-hint *)
  (* I_XsM128d *) |                     Movntps              (* store packed float32s w/NT-hint *)
  (* I_XsM128d *) |                     Movupd               (* unaligned store packed float64s *)
  (* I_XsM128d *) |                     Movups               (* unaligned store packed float32s *)
  (* I_XsM128d *) | [< ISA "AVX"    >]  Vmovapd              (* aligned store packed float64s *)
  (* I_XsM128d *) | [< ISA "AVX"    >]  Vmovaps              (* aligned store packed float32s *)
  (* I_XsM128d *) | [< ISA "AVX"    >]  Vmovdqa              (* store data *)
  (* I_XsM128d *) | [< ISA "AVX"    >]  Vmovdqu              (* store data *)
  (* I_XsM128d *) | [< ISA "AVX"    >]  Vmovntdq             (* store double-quad w/NT-hint *)
  (* I_XsM128d *) | [< ISA "AVX"    >]  Vmovntdqa            (* store double-quad w/NTA-hint *)
  (* I_XsM128d *) | [< ISA "AVX"    >]  Vmovntpd             (* store packed float64s w/NT-hint *)
  (* I_XsM128d *) | [< ISA "AVX"    >]  Vmovntps             (* store packed float32s w/NT-hint *)
  (* I_XsM128d *) | [< ISA "AVX"    >]  Vmovupd              (* unaligned store packed float64s *)
  (* I_XsM128d *) | [< ISA "AVX"    >]  Vmovups              (* unaligned store packed float32s *)

(*****************************************************************************************************************************************************************************)

type I_R32sXd =  |                  Movd     (* move long to vector register *)
  (* I_R32sXd *) | [< ISA "AVX" >]  Vmovd    (* move long to vector register *)
type I_R64sXd =  |                  Movd     (* move long to vector register *)
  (* I_R64sXd *) |                  Movq     (* move quad to vector register *)
  (* I_R64sXd *) | [< ISA "AVX" >]  Vmovd    (* move long to vector register *)
  (* I_R64sXd *) | [< ISA "AVX" >]  Vmovq    (* move quad to vector register *)

(*****************************************************************************************************************************************************************************)
      
type I_XsXd =
  (* I_XsXd *) |                      Cvtdq2pd                   (* convert packed longs to packed float64s *)
  (* I_XsXd *) |                      Cvtdq2ps                   (* convert packed longs to packed float32s *)
  (* I_XsXd *) |                      Cvtpd2dq                   (* convert packed float64s to packed longs *)
  (* I_XsXd *) |                      Cvtpd2ps                   (* convert packed float64s to float32s *)
  (* I_XsXd *) |                      Cvtps2dq                   (* convert packed float32s to packed longs *)
  (* I_XsXd *) |                      Cvtps2pd                   (* convert packed float32s to double-precision *)
  (* I_XsXd *) |                      Cvttpd2dq                  (* convert w/truncation packed float64s to packed longs *)
  (* I_XsXd *) |                      Cvttps2dq                  (* convert w/truncation packed float32s to packed longs *)
  (* I_XsXd *) |                      Movapd                     (* copy packed float64s *)
  (* I_XsXd *) |                      Movaps                     (* copy packed float32s *)
  (* I_XsXd *) |                      Movdqa                     (* load data *)
  (* I_XsXd *) |                      Movdqu                     (* load data *)
  (* I_XsXd *) |                      Movq                       (* load data *)
  (* I_XsXd *) |                      Movupd                     (* load data *)
  (* I_XsXd *) |                      Movups                     (* load data *)
  (* I_XsXd *) |                      Pshufd of Lit              (* shuffle packed longs *)
  (* I_XsXd *) |                      Pshufhw of Lit             (* shuffle high packed words *)
  (* I_XsXd *) |                      Pshuflw of Lit             (* shuffle low packed words *)
  (* I_XsXd *) |                      Rcpps                      (* compute reciprocal of packed float32 *)
  (* I_XsXd *) |                      Rsqrtps                    (* compute reciprocal square roots of packed float32s *)
  (* I_XsXd *) |                      Sqrtpd                     (* compute square roots of packed float64s *)
  (* I_XsXd *) |                      Sqrtps                     (* compute square roots of packed float32s *)
  (* I_XsXd *) | [< ISA "AES"     >]  Aesimc                     (* perform the AES InvMixColumn transformation *)
  (* I_XsXd *) | [< ISA "AES"     >]  Aeskeygenassist of Lit     (* AES round key generation assist *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vcvtdq2pd                  (* convert packed longs to packed float64s *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vcvtdq2ps                  (* convert packed longs to packed float32s *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vcvtpd2dq                  (* convert packed float64s to packed longs *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vcvtpd2ps                  (* convert packed float64s to float32s *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vcvtps2dq                  (* convert packed float32s to packed longs *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vcvtps2pd                  (* convert packed float32s to double-precision *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vcvttpd2dq                 (* convert w/truncation packed float64s to packed longs *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vcvttps2dq                 (* convert w/truncation packed float32s to packed longs *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vmovapd                    (* aligned store packed float64s *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vmovaps                    (* aligned store packed float32s *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vmovddup                   (* copy float64 and duplicate *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vmovdqa                    (* load data *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vmovdqu                    (* load data *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vmovq                      (* load data *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vmovshdup                  (* load data *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vmovsldup                  (* load data *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vmovupd                    (* load data *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vmovups                    (* load data *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpabsb                     (* absolute value of packed bytes *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpabsd                     (* absolute value of packed longs *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpabsw                     (* absolute value of packed words *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpermilpd of Lit           (* permute float64s *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpermilps of Lit           (* permute float32s *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vphminposuw                (* packed horizontal word minimum *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovsxbd                  (* sign-extend packed bytes to packed longs *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovsxbq                  (* sign-extend packed bytes to packed quads *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovsxbw                  (* sign-extend packed bytes to packed words *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovsxdq                  (* sign-extend packed longs to packed quads *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovsxwd                  (* sign-extend packed words to packed longs *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovsxwq                  (* sign-extend packed words to packed quads *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovzxbd                  (* zero-extend packed bytes to packed longs *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovzxbq                  (* zero-extend packed bytes to packed quads *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovzxbw                  (* zero-extend packed bytes to packed words *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovzxdq                  (* zero-extend packed longs to packed quads *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovzxwd                  (* zero-extend packed words to packed longs *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpmovzxwq                  (* zero-extend packed words to packed quads *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpshufd of Lit             (* shuffle packed longs *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpshufhw of Lit            (* shuffle high packed words *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpshuflw of Lit            (* shuffle low packed words *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpslld of Lit              (* shift packed data left logical by <N> bits *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpslldq of Lit             (* shift packed data left logical by <N> bytes *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpsllq of Lit              (* shift packed data left logical by <N> bits *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpsllw of Lit              (* shift packed data left logical by <N> bits *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpsrad of Lit              (* shift packed data right arithmetic by <N> bits *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpsraw of Lit              (* shift packed data right arithmetic by <N> bits *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpsrld of Lit              (* shift packed data right logical by <N> bits *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpsrldq of Lit             (* shift packed data right logical by <N> bytes *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpsrlq of Lit              (* shift packed data right logical by <N> bits *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vpsrlw of Lit              (* shift packed data right logical by <N> bits *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vrcpps                     (* compute reciprocal of packed float32 *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vroundpd of Lit            (* round packed float64s *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vroundps of Lit            (* round packed float32s *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vrsqrtps                   (* compute reciprocal square roots of packed float32s *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vsqrtpd                    (* compute square roots of packed float64s *)
  (* I_XsXd *) | [< ISA "AVX"     >]  Vsqrtps                    (* compute square roots of packed float32s *)
  (* I_XsXd *) | [< ISA "AVX,AES" >]  Vaesimc                    (* perform the AES InvMixColumn transformation *)
  (* I_XsXd *) | [< ISA "AVX,AES" >]  Vaeskeygenassist of Lit    (* AES round key generation assist *)
  (* I_XsXd *) | [< ISA "SSE3"    >]  Movddup                    (* copy float64 and duplicate *)
  (* I_XsXd *) | [< ISA "SSE3"    >]  Movshdup                   (* load data *)
  (* I_XsXd *) | [< ISA "SSE3"    >]  Movsldup                   (* load data *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Phminposuw                 (* packed horizontal word minimum *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovsxbd                   (* sign-extend packed bytes to packed longs *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovsxbq                   (* sign-extend packed bytes to packed quads *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovsxbw                   (* sign-extend packed bytes to packed words *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovsxdq                   (* sign-extend packed longs to packed quads *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovsxwd                   (* sign-extend packed words to packed longs *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovsxwq                   (* sign-extend packed words to packed quads *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovzxbd                   (* zero-extend packed bytes to packed longs *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovzxbq                   (* zero-extend packed bytes to packed quads *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovzxbw                   (* zero-extend packed bytes to packed words *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovzxdq                   (* zero-extend packed longs to packed quads *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovzxwd                   (* zero-extend packed words to packed longs *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Pmovzxwq                   (* zero-extend packed words to packed quads *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Roundpd of Lit             (* round packed float64s *)
  (* I_XsXd *) | [< ISA "SSE4_1"  >]  Roundps of Lit             (* round packed float32s *)
  (* I_XsXd *) | [< ISA "SSSE3"   >]  Pabsb                      (* absolute value of packed bytes *)
  (* I_XsXd *) | [< ISA "SSSE3"   >]  Pabsd                      (* absolute value of packed longs *)
  (* I_XsXd *) | [< ISA "SSSE3"   >]  Pabsw                      (* absolute value of packed words *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vfrczpd                    (* extract fraction packed float64 *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vfrczps                    (* extract fraction packed float32 *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vfrczsd                    (* extract fraction scalar float64 *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vfrczss                    (* extract fraction scalar float32 *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphaddbd                   (* packed horizontal add signed byte to signed long *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphaddbq                   (* packed horizontal add signed byte to signed quad *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphaddbw                   (* packed horizontal add signed byte to signed word *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphadddq                   (* packed horizontal add signed long to signed quad *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphaddubd                  (* packed horizontal add unsigned byte to long *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphaddubq                  (* packed horizontal add unsigned byte to quad *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphaddubw                  (* packed horizontal add unsigned byte to word *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphaddudq                  (* packed horizontal add unsigned long to quad *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphadduwd                  (* packed horizontal add unsigned word to long *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphadduwq                  (* packed horizontal add unsigned word to quad *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphaddwd                   (* packed horizontal add signed word to signed long *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphaddwq                   (* packed horizontal add signed word to signed quad *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphsubbw                   (* packed horizontal subtract signed byte to signed word *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphsubdq                   (* packed horizontal subtract signed byte to signed word *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vphsubwd                   (* packed horizontal subtract signed word to signed long *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vprotb of Lit              (* packed rotate bytes *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vprotd of Lit              (* packed rotate longs *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vprotq of Lit              (* packed rotate quads *)
  (* I_XsXd *) | [< ISA "XOP"     >]  Vprotw of Lit              (* packed rotate words *)
 
(*****************************************************************************************************************************************************************************)

type I_M16sXd =   | [< ISA "AVX"     >]  Vpmovsxbq                  (* sign-extend packed bytes to packed quads *)
  (* I_M16sXd *)  | [< ISA "AVX"     >]  Vpmovzxbq                  (* zero-extend packed bytes to packed quads *)
  (* I_M16sXd *)  | [< ISA "SSE4_1"  >]  Pmovsxbq                   (* sign-extend packed bytes to packed quads *)
  (* I_M16sXd *)  | [< ISA "SSE4_1"  >]  Pmovzxbq                   (* zero-extend packed bytes to packed quads *)
type I_M32sXd =   |                      Movd                       (* load long *)
  (* I_M32sXd *)  |                      Movss                      (* load scalar float32 *)
  (* I_M32sXd *)  | [< ISA "AVX"     >]  Vbroadcastss               (* load float32 and broadcast *)
  (* I_M32sXd *)  | [< ISA "AVX"     >]  Vmovd                      (* load long *)
  (* I_M32sXd *)  | [< ISA "AVX"     >]  Vmovss                     (* load scalar float32 *)
  (* I_M32sXd *)  | [< ISA "AVX"     >]  Vpmovsxbd                  (* sign-extend packed bytes to packed longs *)
  (* I_M32sXd *)  | [< ISA "AVX"     >]  Vpmovsxwq                  (* sign-extend packed words to packed quads *)
  (* I_M32sXd *)  | [< ISA "AVX"     >]  Vpmovzxbd                  (* zero-extend packed bytes to packed longs *)
  (* I_M32sXd *)  | [< ISA "AVX"     >]  Vpmovzxwq                  (* zero-extend packed words to packed quads *)
  (* I_M32sXd *)  | [< ISA "SSE4_1"  >]  Pmovsxbd                   (* sign-extend packed bytes to packed longs *)
  (* I_M32sXd *)  | [< ISA "SSE4_1"  >]  Pmovsxwq                   (* sign-extend packed words to packed quads *)
  (* I_M32sXd *)  | [< ISA "SSE4_1"  >]  Pmovzxbd                   (* zero-extend packed bytes to packed longs *)
  (* I_M32sXd *)  | [< ISA "SSE4_1"  >]  Pmovzxwq                   (* zero-extend packed words to packed quads *)
type I_M64sXd =   |                      Movddup                    (* load float64 and duplicate *)
  (* I_M64sXd *)  |                      Movq                       (* load quad *)
  (* I_M64sXd *)  |                      Movsd                      (* load scalar float64 *)
  (* I_M64sXd *)  | [< ISA "AVX"     >]  Vmovddup                   (* load float64 and duplicate *)
  (* I_M64sXd *)  | [< ISA "AVX"     >]  Vmovq                      (* load quad *)
  (* I_M64sXd *)  | [< ISA "AVX"     >]  Vmovsd                     (* load scalar float64 *)
  (* I_M64sXd *)  | [< ISA "AVX"     >]  Vpmovsxbw                  (* sign-extend packed bytes to packed words *)
  (* I_M64sXd *)  | [< ISA "AVX"     >]  Vpmovsxdq                  (* sign-extend packed longs to packed quads *)
  (* I_M64sXd *)  | [< ISA "AVX"     >]  Vpmovsxwd                  (* sign-extend packed words to packed longs *)
  (* I_M64sXd *)  | [< ISA "AVX"     >]  Vpmovzxbw                  (* zero-extend packed bytes to packed words *)
  (* I_M64sXd *)  | [< ISA "AVX"     >]  Vpmovzxdq                  (* zero-extend packed longs to packed quads *)
  (* I_M64sXd *)  | [< ISA "AVX"     >]  Vpmovzxwd                  (* zero-extend packed words to packed longs *)
  (* I_M64sXd *)  | [< ISA "SSE4_1"  >]  Pmovsxbw                   (* sign-extend packed bytes to packed words *)
  (* I_M64sXd *)  | [< ISA "SSE4_1"  >]  Pmovsxdq                   (* sign-extend packed longs to packed quads *)
  (* I_M64sXd *)  | [< ISA "SSE4_1"  >]  Pmovsxwd                   (* sign-extend packed words to packed longs *)
  (* I_M64sXd *)  | [< ISA "SSE4_1"  >]  Pmovzxbw                   (* zero-extend packed bytes to packed words *)
  (* I_M64sXd *)  | [< ISA "SSE4_1"  >]  Pmovzxdq                   (* zero-extend packed longs to packed quads *)
  (* I_M64sXd *)  | [< ISA "SSE4_1"  >]  Pmovzxwd                   (* zero-extend packed words to packed longs *)
  (* I_M64sXd *)  |                      Cvtpi2pd                   (* convert packed longs to packed float64s *)
type I_M128sXd =  |                      Cvtdq2pd                   (* convert packed integers to packed float64s *)
  (* I_M128sXd *) |                      Cvtdq2ps                   (* convert packed integers to packed float32s *)
  (* I_M128sXd *) |                      Cvtpd2dq                   (* convert packed float64s to packed longs *)
  (* I_M128sXd *) |                      Cvtpd2ps                   (* convert packed float64s to packed float32s *)
  (* I_M128sXd *) |                      Cvtps2dq                   (* convert packed float32s to packed longs *)
  (* I_M128sXd *) |                      Cvtps2pd                   (* convert packed float32s to packed float64s *)
  (* I_M128sXd *) |                      Cvttpd2dq                  (* convert w/truncation packed float64s to packed longs *)
  (* I_M128sXd *) |                      Cvttps2dq                  (* convert w/truncation packed float32s to packed longs *)
  (* I_M128sXd *) |                      Movapd                     (* aligned load packed float64s *)
  (* I_M128sXd *) |                      Movaps                     (* aligned load packed float32s *)
  (* I_M128sXd *) |                      Movdqa                     (* aligned load packed double-quad data *)
  (* I_M128sXd *) |                      Movdqu                     (* unaligned load packed double-quad data *)
  (* I_M128sXd *) |                      Movupd                     (* unaligned load packed float64s *)
  (* I_M128sXd *) |                      Movups                     (* unaligned load packed float32s *)
  (* I_M128sXd *) |                      Pshufd  of Lit             (* shuffle packed longs *)
  (* I_M128sXd *) |                      Pshufhw of Lit             (* shuffle high packed words *)
  (* I_M128sXd *) |                      Pshuflw of Lit             (* shuffle low packed words *)
  (* I_M128sXd *) |                      Rcpps                      (* compute reciprocal of packed float32 *)
  (* I_M128sXd *) |                      Rsqrtps                    (* compute reciprocal square roots of packed float32s *)
  (* I_M128sXd *) |                      Sqrtpd                     (* compute square roots of packed float64s *)
  (* I_M128sXd *) |                      Sqrtps                     (* compute square roots of packed float32s *)
  (* I_M128sXd *) | [< ISA "AES"     >]  Aesimc                     (* perform the AES InvMixColumn transformation *)
  (* I_M128sXd *) | [< ISA "AES"     >]  Aeskeygenassist of Lit     (* AES round key generation assist *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vcvtdq2pd                  (* convert packed integers to packed float64s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vcvtdq2ps                  (* convert packed integers to packed float32s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vcvtpd2dq                  (* convert packed float64s to packed longs *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vcvtpd2ps                  (* convert packed float64s to packed float32s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vcvtps2dq                  (* convert packed float32s to packed longs *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vcvtps2pd                  (* convert packed float32s to packed float64s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vcvttpd2dq                 (* convert w/truncation packed float64s to packed longs *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vcvttps2dq                 (* convert w/truncation packed float32s to packed longs *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vlddqu                     (* unaligned load integer 128 bits *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vmovapd                    (* aligned load packed float64s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vmovaps                    (* aligned load packed float32s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vmovdqa                    (* aligned load packed double-quad data *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vmovdqu                    (* unaligned load packed double-quad data *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vmovupd                    (* unaligned load packed float64s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vmovups                    (* unaligned load packed float32s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vpabsb                     (* absolute value of packed bytes *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vpabsd                     (* absolute value of packed longs *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vpabsw                     (* absolute value of packed words *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vphminposuw                (* packed horizontal word minimum *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vpshufd of Lit             (* shuffle packed longs *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vpshufhw of Lit            (* shuffle high packed words *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vpshuflw of Lit            (* shuffle low packed words *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vrcpps                     (* compute reciprocal of packed float32 *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vroundpd of Lit            (* round packed float64s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vroundps of Lit            (* round packed float32s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vrsqrtps                   (* compute reciprocal square roots of packed float32s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vsqrtpd                    (* compute square roots of packed float64s *)
  (* I_M128sXd *) | [< ISA "AVX"     >]  Vsqrtps                    (* compute square roots of packed float32s *)
  (* I_M128sXd *) | [< ISA "AVX,AES" >]  Vaesimc                    (* perform the AES InvMixColumn transformation *)
  (* I_M128sXd *) | [< ISA "AVX,AES" >]  Vaeskeygenassist of Lit    (* AES round key generation assist *)
  (* I_M128sXd *) | [< ISA "SSE3"    >]  Lddqu                      (* unaligned load integer 128 bits *)
  (* I_M128sXd *) | [< ISA "SSE4_1"  >]  Phminposuw                 (* packed horizontal word minimum *)
  (* I_M128sXd *) | [< ISA "SSE4_1"  >]  Roundpd of Lit             (* round packed float64s *)
  (* I_M128sXd *) | [< ISA "SSE4_1"  >]  Roundps of Lit             (* round packed float32s *)
  (* I_M128sXd *) | [< ISA "SSSE3"   >]  Pabsb                      (* absolute value of packed bytes *)
  (* I_M128sXd *) | [< ISA "SSSE3"   >]  Pabsd                      (* absolute value of packed longs *)
  (* I_M128sXd *) | [< ISA "SSSE3"   >]  Pabsw                      (* absolute value of packed words *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vfrczpd                    (* extract fraction packed float64 *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vfrczps                    (* extract fraction packed float32 *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vfrczsd                    (* extract fraction scalar float64 *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vfrczss                    (* extract fraction scalar float32 *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphaddbd                   (* packed horizontal add signed byte to signed long *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphaddbq                   (* packed horizontal add signed byte to signed quad *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphaddbw                   (* packed horizontal add signed byte to signed word *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphadddq                   (* packed horizontal add signed long to signed quad *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphaddubd                  (* packed horizontal add unsigned byte to long *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphaddubq                  (* packed horizontal add unsigned byte to quad *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphaddubw                  (* packed horizontal add unsigned byte to word *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphaddudq                  (* packed horizontal add unsigned long to quad *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphadduwd                  (* packed horizontal add unsigned word to long *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphadduwq                  (* packed horizontal add unsigned word to quad *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphaddwd                   (* packed horizontal add signed word to signed long *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphaddwq                   (* packed horizontal add signed word to signed quad *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphsubbw                   (* packed horizontal subtract signed byte to signed word *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphsubdq                   (* packed horizontal subtract signed byte to signed word *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vphsubwd                   (* packed horizontal subtract signed word to signed long *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vprotb of Lit              (* packed rotate bytes *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vprotd of Lit              (* packed rotate longs *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vprotq of Lit              (* packed rotate quads *)
  (* I_M128sXd *) | [< ISA "XOP"     >]  Vprotw of Lit              (* packed rotate words *)

(*****************************************************************************************************************************************************************************)

type I_M8sXds =    | [< ISA "SSE4_1"                               >]  Pinsrb of Lit              (* insert byte into vector *)
type I_M16sXds =   |                                                   Pinsrw of Lit              (* insert word into vector *)
type I_M32sXds =   | [<                 AKA "cvtsi2sd"             >]  Cvtsi2sdl                  (* convert scalar longs to float64 *)
  (* I_M32sXds *)  |                                                   Cvtss2sd                   (* convert scalar float32 to float64 *)
  (* I_M32sXds *)  | [<                 AKA "cvtsi2ss"             >]  Cvtsi2ssl                  (* convert longs to scalar float32 *)
  (* I_M32sXds *)  |                                                   Addss                      (* add scalar float32s *)
  (* I_M32sXds *)  |                                                   Cmpss of Lit               (* arithmetic comparison of scalar float32s *)
  (* I_M32sXds *)  |                                                   Divss                      (* divide scalar float32s *)
  (* I_M32sXds *)  | [< ISA "SSE4_1"                               >]  Insertps of Lit            (* insert packed float32 *)
  (* I_M32sXds *)  |                                                   Maxss                      (* maximum of scalar float32s *)
  (* I_M32sXds *)  |                                                   Minss                      (* minimum of scalar float32s *)
  (* I_M32sXds *)  |                                                   Mulss                      (* multiply scalar float32s *)
  (* I_M32sXds *)  | [< ISA "SSE4_1"                               >]  Roundss of Lit             (* round scalar float32 *)
  (* I_M32sXds *)  |                                                   Rsqrtss                    (* compute reciprocal of square root of scalar float32 *)
  (* I_M32sXds *)  |                                                   Sqrtss                     (* compute square root of scalar float32 *)
  (* I_M32sXds *)  |                                                   Subss                      (* subtract scalar float32s *)
  (* I_M32sXds *)  | [< ISA "SSE4_1"                               >]  Pinsrd of Lit              (* insert long into vector *)
type I_M64sXds =   |                                                   Cvtsd2ss                   (* convert float64 to float32 *)
  (* I_M64sXds *)  |                                                   Cvtsi2sdq                  (* convert quad to float64 *)
  (* I_M64sXds *)  |                                                   Cvtsi2ssq                  (* convert quad to scalar float32 *)
  (* I_M64sXds *)  |                                                   Addsd                      (* add scalar float64s *)
  (* I_M64sXds *)  |                                                   Cmpsd of Lit               (* arithmetic comparison of scalar float64s *)
  (* I_M64sXds *)  |                                                   Divsd                      (* divide scalar float64s *)
  (* I_M64sXds *)  |                                                   Maxsd                      (* maximum of scalar float64s *)
  (* I_M64sXds *)  |                                                   Minsd                      (* minimum of scalar float64s *)
  (* I_M64sXds *)  |                                                   Movhpd                     (* load high part of packed float64s *)
  (* I_M64sXds *)  |                                                   Movhps                     (* load high parts of packed float32s *)
  (* I_M64sXds *)  |                                                   Movlpd                     (* load low part of packed float64s *)
  (* I_M64sXds *)  |                                                   Movlps                     (* load low parts of packed float32s *)
  (* I_M64sXds *)  |                                                   Mulsd                      (* multiply scalar float64s *)
  (* I_M64sXds *)  | [< ISA "SSE4_1"                               >]  Roundsd of Lit             (* round scalar float64 *)
  (* I_M64sXds *)  |                                                   Sqrtsd                     (* compute square root of scalar float64 *)
  (* I_M64sXds *)  |                                                   Subsd                      (* subtract scalar float64s *)
  (* I_M64sXds *)  | [< ISA "SSE4_1"                               >]  Pinsrq of Lit              (* insert quad into vector *)
  (* I_M64sXds *)  |                                                   Cvtpi2ps                   (* convert packed longs to packed float32s *)
type I_M128sXds =  |                                                   Addpd                      (* add packed float64s *)
  (* I_M128sXds *) |                                                   Addps                      (* add packed float32s *)
  (* I_M128sXds *) | [< ISA "SSE3"                                 >]  Addsubpd                   (* mixed-add-subtract packed float64s *)
  (* I_M128sXds *) | [< ISA "SSE3"                                 >]  Addsubps                   (* mixed-add-subtract packed float32s *)
  (* I_M128sXds *) | [< ISA "AES"                                  >]  Aesdec                     (* perform one round of an AES decryption flow *)
  (* I_M128sXds *) | [< ISA "AES"                                  >]  Aesdeclast                 (* perform last round of an AES decryption flow *)
  (* I_M128sXds *) | [< ISA "AES"                                  >]  Aesenc                     (* perform one round of an AES encryption flow *)
  (* I_M128sXds *) | [< ISA "AES"                                  >]  Aesenclast                 (* perform last round of an AES encryption flow *)
  (* I_M128sXds *) |                                                   Andnpd                     (* logical AND-NOT packed float64s *)
  (* I_M128sXds *) |                                                   Andnps                     (* logical AND-NOT packed float32s *)
  (* I_M128sXds *) |                                                   Andpd                      (* logical AND packed float64s *)
  (* I_M128sXds *) |                                                   Andps                      (* logical AND packed float32s *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Blendpd of Lit             (* blend packed float64s *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Blendps of Lit             (* blend packed float32s *)
  (* I_M128sXds *) | [< ISA "SSE4_1";                    Xs "xmm0" >]  Blendvpd                   (* variable blend packed float64s *)
  (* I_M128sXds *) | [< ISA "SSE4_1";                    Xs "xmm0" >]  Blendvps                   (* variable blend packed float32s *)
  (* I_M128sXds *) |                                                   Cmppd of Lit               (* arithmetic comparison of packed float64s *)
  (* I_M128sXds *) |                                                   Cmpps of Lit               (* arithmetic comparison of packed float32s *)
  (* I_M128sXds *) |                                                   Divpd                      (* divide packed float64s *)
  (* I_M128sXds *) |                                                   Divps                      (* divide packed float32s *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Dppd of Lit                (* dot-product of packed float64s *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Dpps of Lit                (* dot-product of packed float32s *)
  (* I_M128sXds *) | [< ISA "SSE3"                                 >]  Haddpd                     (* horizontal add packed float64 *)
  (* I_M128sXds *) | [< ISA "SSE3"                                 >]  Haddps                     (* horizontal add packed float32 *)
  (* I_M128sXds *) | [< ISA "SSE3"                                 >]  Hsubpd                     (* horizontal subtract packed float64 *)
  (* I_M128sXds *) | [< ISA "SSE3"                                 >]  Hsubps                     (* horizontal subtract packed float32 *)
  (* I_M128sXds *) |                                                   Maxpd                      (* maximum of packed float64s *)
  (* I_M128sXds *) |                                                   Maxps                      (* maximum of packed float32s *)
  (* I_M128sXds *) |                                                   Minpd                      (* minimum of packed float64s *)
  (* I_M128sXds *) |                                                   Minps                      (* minimum of packed float32s *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Mpsadbw of Lit             (* compute multiple packed sums of absolute difference *)
  (* I_M128sXds *) |                                                   Mulpd                      (* multiply packed float64s *)
  (* I_M128sXds *) |                                                   Mulps                      (* multiply packed float32s *)
  (* I_M128sXds *) |                                                   Orpd                       (* logical inclusive-OR packed float64s *)
  (* I_M128sXds *) |                                                   Orps                       (* logical inclusive-OR packed float32s *)
  (* I_M128sXds *) |                                                   Packssdw                   (* convert packed longs to words w/signed saturation *)
  (* I_M128sXds *) |                                                   Packsswb                   (* convert packed words to bytes w/signed saturation *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Packusdw                   (* convert packed longs to words w/unsigned saturation *)
  (* I_M128sXds *) |                                                   Packuswb                   (* convert packed words to bytes w/unsigned saturation *)
  (* I_M128sXds *) |                                                   Paddb                      (* add packed bytes *)
  (* I_M128sXds *) |                                                   Paddd                      (* add packed longs *)
  (* I_M128sXds *) |                                                   Paddq                      (* add packed quads *)
  (* I_M128sXds *) |                                                   Paddsb                     (* add packed signed bytes (w/saturation) *)
  (* I_M128sXds *) |                                                   Paddsw                     (* add packed signed words (w/saturation) *)
  (* I_M128sXds *) |                                                   Paddusb                    (* add packed unsigned bytes (w/saturation) *)
  (* I_M128sXds *) |                                                   Paddusw                    (* add packed unsigned words (w/saturation) *)
  (* I_M128sXds *) |                                                   Paddw                      (* add packed words *)
  (* I_M128sXds *) |                                                   Pand                       (* logical AND packed integers *)
  (* I_M128sXds *) |                                                   Pandn                      (* logical AND-NOT packed integers *)
  (* I_M128sXds *) |                                                   Pavgb                      (* average packed bytes *)
  (* I_M128sXds *) |                                                   Pavgw                      (* average packed words *)
  (* I_M128sXds *) | [< ISA "SSE4_1";                    Xs "xmm0" >]  Pblendvb                   (* variable blend packed bytes *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pblendw of Lit             (* blend packed words *)
  (* I_M128sXds *) | [< ISA "CLMUL"                                >]  Pclmulqdq of Lit           (* carry-less multiply quads *)
  (* I_M128sXds *) |                                                   Pcmpeqb                    (* arithmetic compare-is-equal packed bytes *)
  (* I_M128sXds *) |                                                   Pcmpeqd                    (* arithmetic compare-is-equal packed longs *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pcmpeqq                    (* arithmetic compare-is-equal packed quads *)
  (* I_M128sXds *) |                                                   Pcmpeqw                    (* arithmetic compare-is-equal packed words *)
  (* I_M128sXds *) |                                                   Pcmpgtb                    (* arithmetic compare-is-greater-than packed bytes *)
  (* I_M128sXds *) |                                                   Pcmpgtd                    (* arithmetic compare-is-greater-than packed longs *)
  (* I_M128sXds *) | [< ISA "SSE4_2"                               >]  Pcmpgtq                    (* arithmetic compare-is-greater-than packed quads *)
  (* I_M128sXds *) |                                                   Pcmpgtw                    (* arithmetic compare-is-greater-than packed words *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Phaddd                     (* horizontal add packed longs *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Phaddsw                    (* horizontal add packed words (w/saturation) *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Phaddw                     (* horizontal add packed words *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Phsubd                     (* horizontal subtract packed longs *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Phsubsw                    (* horizontal subtract packed words (w/saturation) *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Phsubw                     (* horizontal subtract packed words *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Pmaddubsw                  (* multiply and add packed signed and unsigned bytes *)
  (* I_M128sXds *) |                                                   Pmaddwd                    (* multiply and add packed integers *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pmaxsb                     (* maximum of packed signed bytes *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pmaxsd                     (* maximum of packed signed longs *)
  (* I_M128sXds *) |                                                   Pmaxsw                     (* maximum of packed signed words *)
  (* I_M128sXds *) |                                                   Pmaxub                     (* maximum of packed unsigned bytes *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pmaxud                     (* maximum of packed unsigned longs *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pmaxuw                     (* maximum of packed unsigned words *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pminsb                     (* minimum of packed signed bytes *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pminsd                     (* minimum of packed signed longs *)
  (* I_M128sXds *) |                                                   Pminsw                     (* minimum of packed signed words *)
  (* I_M128sXds *) |                                                   Pminub                     (* minimum of packed unsigned bytes *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pminud                     (* minimum of packed unsigned longs *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pminuw                     (* minimum of packed unsigned words *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pmuldq                     (* multiply packed signed longs to quads *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Pmulhrsw                   (* packed multiply high w/round and scale *)
  (* I_M128sXds *) |                                                   Pmulhuw                    (* multiply packed unsigned words and store high result *)
  (* I_M128sXds *) |                                                   Pmulhw                     (* multiply packed signed words and store high result *)
  (* I_M128sXds *) | [< ISA "SSE4_1"                               >]  Pmulld                     (* multiply packed signed longs and store low result *)
  (* I_M128sXds *) |                                                   Pmullw                     (* multiply packed signed words and store low result *)
  (* I_M128sXds *) |                                                   Pmuludq                    (* multiply packed integers *)
  (* I_M128sXds *) |                                                   Por                        (* logical inclusive-OR packed integers *)
  (* I_M128sXds *) |                                                   Psadbw                     (* compute sum of absolute differences *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Pshufb                     (* packed shuffle bytes *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Psignb                     (* packed sign bytes *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Psignd                     (* packed sign longs *)
  (* I_M128sXds *) | [< ISA "SSSE3"                                >]  Psignw                     (* packed sign words *)
  (* I_M128sXds *) |                                                   Pslld                      (* shift packed data left logical by <N> bits *)
  (* I_M128sXds *) |                                                   Psllq                      (* shift packed data left logical by <N> bits *)
  (* I_M128sXds *) |                                                   Psllw                      (* shift packed data left logical by <N> bits *)
  (* I_M128sXds *) |                                                   Psrad                      (* shift packed data right arithmetic by <N> bits *)
  (* I_M128sXds *) |                                                   Psraw                      (* shift packed data right arithmetic by <N> bits *)
  (* I_M128sXds *) |                                                   Psrld                      (* shift packed data right logical by <N> bits *)
  (* I_M128sXds *) |                                                   Psrlq                      (* shift packed data right logical by <N> bits *)
  (* I_M128sXds *) |                                                   Psrlw                      (* shift packed data right logical by <N> bits *)
  (* I_M128sXds *) |                                                   Psubb                      (* subtract packed bytes *)
  (* I_M128sXds *) |                                                   Psubd                      (* subtract packed longs *)
  (* I_M128sXds *) |                                                   Psubq                      (* subtract packed quads *)
  (* I_M128sXds *) |                                                   Psubsb                     (* subtract packed signed bytes (w/saturation) *)
  (* I_M128sXds *) |                                                   Psubsw                     (* subtract packed signed words (w/saturation) *)
  (* I_M128sXds *) |                                                   Psubusb                    (* subtract packed unsigned bytes (w/saturation) *)
  (* I_M128sXds *) |                                                   Psubusw                    (* subtract packed unsigned words (w/saturation) *)
  (* I_M128sXds *) |                                                   Psubw                      (* subtract packed words *)
  (* I_M128sXds *) |                                                   Punpckhbw                  (* unpack high bytes to words *)
  (* I_M128sXds *) |                                                   Punpckhdq                  (* unpack high longs to quads *)
  (* I_M128sXds *) |                                                   Punpckhqdq                 (* unpack high quads to double-quads *)
  (* I_M128sXds *) |                                                   Punpckhwd                  (* unpack high words to longs *)
  (* I_M128sXds *) |                                                   Punpcklbw                  (* unpack low bytes to words *)
  (* I_M128sXds *) |                                                   Punpckldq                  (* unpack low longs to quads *)
  (* I_M128sXds *) |                                                   Punpcklqdq                 (* unpack low quads to double-quads *)
  (* I_M128sXds *) |                                                   Punpcklwd                  (* unpack low words to longs *)
  (* I_M128sXds *) |                                                   Pxor                       (* logical exclusive-OR packed integers *)
  (* I_M128sXds *) |                                                   Shufpd of Lit              (* shuffle packed float64s *)
  (* I_M128sXds *) |                                                   Shufps of Lit              (* shuffle packed float32s *)
  (* I_M128sXds *) |                                                   Subpd                      (* subtract packed float64s *)
  (* I_M128sXds *) |                                                   Subps                      (* subtract packed float32s *)
  (* I_M128sXds *) |                                                   Unpckhpd                   (* unpack and interleave high packed float64s *)
  (* I_M128sXds *) |                                                   Unpckhps                   (* unpack and interleave high packed float32s *)
  (* I_M128sXds *) |                                                   Unpcklpd                   (* unpack and interleave low packed float64s *)
  (* I_M128sXds *) |                                                   Unpcklps                   (* unpack and interleave low packed float32s *)
  (* I_M128sXds *) |                                                   Xorpd                      (* logical exclusive-OR packed float64s *)
  (* I_M128sXds *) |                                                   Xorps                      (* logical exclusive-OR packed float32s *)

(*****************************************************************************************************************************************************************************)

type I_XsXds =  |                                                  Addpd                      (* add packed float64s *)
  (* I_XsXds *) |                                                  Addps                      (* add packed float32s *)
  (* I_XsXds *) |                                                  Addsd                      (* add scalar float64s *)
  (* I_XsXds *) |                                                  Addss                      (* add scalar float32s *)
  (* I_XsXds *) |                                                  Andnpd                     (* logical AND-NOT packed float64s *)
  (* I_XsXds *) |                                                  Andnps                     (* logical AND-NOT packed float32s *)
  (* I_XsXds *) |                                                  Andpd                      (* logical AND packed float64s *)
  (* I_XsXds *) |                                                  Andps                      (* logical AND packed float32s *)
  (* I_XsXds *) |                                                  Cmppd of Lit               (* arithmetic comparison of packed float64s *)
  (* I_XsXds *) |                                                  Cmpps of Lit               (* arithmetic comparison of packed float32s *)
  (* I_XsXds *) |                                                  Cmpsd of Lit               (* arithmetic comparison of scalar float64s *)
  (* I_XsXds *) |                                                  Cmpss of Lit               (* arithmetic comparison of scalar float32s *)
  (* I_XsXds *) |                                                  Cvtsd2ss                   (* convert scalar float64 to float32 *)
  (* I_XsXds *) |                                                  Cvtss2sd                   (* convert scalar float32 to float64 *)
  (* I_XsXds *) |                                                  Divpd                      (* divide packed float64s *)
  (* I_XsXds *) |                                                  Divps                      (* divide packed float32s *)
  (* I_XsXds *) |                                                  Divsd                      (* divide scalar float64s *)
  (* I_XsXds *) |                                                  Divss                      (* divide scalar float32s *)
  (* I_XsXds *) |                                                  Maxpd                      (* maximum of packed float64s *)
  (* I_XsXds *) |                                                  Maxps                      (* maximum of packed float32s *)
  (* I_XsXds *) |                                                  Maxsd                      (* maximum of scalar float64s *)
  (* I_XsXds *) |                                                  Maxss                      (* maximum of scalar float32s *)
  (* I_XsXds *) |                                                  Minpd                      (* minimum of packed float64s *)
  (* I_XsXds *) |                                                  Minps                      (* minimum of packed float32s *)
  (* I_XsXds *) |                                                  Minsd                      (* minimum of scalar float64s *)
  (* I_XsXds *) |                                                  Minss                      (* minimum of scalar float32s *)
  (* I_XsXds *) |                                                  Movhlps                    (* move part of vector *)
  (* I_XsXds *) |                                                  Movlhps                    (* move part of vector *)
  (* I_XsXds *) |                                                  Movsd                      (* move scalar float64 *)
  (* I_XsXds *) |                                                  Movss                      (* move scalar float32 *)
  (* I_XsXds *) |                                                  Mulpd                      (* multiply packed float64s *)
  (* I_XsXds *) |                                                  Mulps                      (* multiply packed float32s *)
  (* I_XsXds *) |                                                  Mulsd                      (* multiply scalar float64s *)
  (* I_XsXds *) |                                                  Mulss                      (* multiply scalar float32s *)
  (* I_XsXds *) |                                                  Orpd                       (* logical inclusive-OR packed float64s *)
  (* I_XsXds *) |                                                  Orps                       (* logical inclusive-OR packed float32s *)
  (* I_XsXds *) |                                                  Packssdw                   (* convert packed longs to packed words w/signed saturation *)
  (* I_XsXds *) |                                                  Packsswb                   (* convert packed words to packed bytes w/signed saturation *)
  (* I_XsXds *) |                                                  Packuswb                   (* convert packed words to packed bytes w/unsigned saturation *)
  (* I_XsXds *) |                                                  Paddb                      (* add packed bytes *)
  (* I_XsXds *) |                                                  Paddd                      (* add packed longs *)
  (* I_XsXds *) |                                                  Paddq                      (* add packed quads *)
  (* I_XsXds *) |                                                  Paddsb                     (* add packed signed bytes (w/saturation) *)
  (* I_XsXds *) |                                                  Paddsw                     (* add packed signed words (w/saturation) *)
  (* I_XsXds *) |                                                  Paddusb                    (* add packed unsigned bytes (w/saturation) *)
  (* I_XsXds *) |                                                  Paddusw                    (* add packed unsigned words (w/saturation) *)
  (* I_XsXds *) |                                                  Paddw                      (* add packed words *)
  (* I_XsXds *) |                                                  Pand                       (* logical AND packed integers *)
  (* I_XsXds *) |                                                  Pandn                      (* logical AND-NOT packed integers *)
  (* I_XsXds *) |                                                  Pavgb                      (* average packed bytes *)
  (* I_XsXds *) |                                                  Pavgw                      (* average packed words *)
  (* I_XsXds *) |                                                  Pcmpeqb                    (* arithmetic compare-is-equal bytes *)
  (* I_XsXds *) |                                                  Pcmpeqd                    (* arithmetic compare-is-equal longs *)
  (* I_XsXds *) |                                                  Pcmpeqw                    (* arithmetic compare-is-equal words *)
  (* I_XsXds *) |                                                  Pcmpgtb                    (* arithmetic compare-is-greater-than packed bytes *)
  (* I_XsXds *) |                                                  Pcmpgtd                    (* arithmetic compare-is-greater-than packed longs *)
  (* I_XsXds *) |                                                  Pcmpgtw                    (* arithmetic compare-is-greater-than packed words *)
  (* I_XsXds *) |                                                  Pmaddwd                    (* multiply and add packed integers    *)
  (* I_XsXds *) |                                                  Pmaxsw                     (* maximum of packed signed words *)
  (* I_XsXds *) |                                                  Pmaxub                     (* maximum of packed unsigned bytes *)
  (* I_XsXds *) |                                                  Pminsw                     (* minimum of packed signed words *)
  (* I_XsXds *) |                                                  Pminub                     (* minimum of packed unsigned bytes *)
  (* I_XsXds *) |                                                  Pmulhuw                    (* multiply packed unsigned words and store high result *)
  (* I_XsXds *) |                                                  Pmulhw                     (* multiply packed signed words and store high result *)
  (* I_XsXds *) |                                                  Pmullw                     (* multiply packed signed words and store low result *)
  (* I_XsXds *) |                                                  Pmuludq                    (* multiply packed integers *)
  (* I_XsXds *) |                                                  Por                        (* logical inclusive-OR packed integers *)
  (* I_XsXds *) |                                                  Psadbw                     (* compute sum of absolute differences *)
  (* I_XsXds *) |                                                  Pslld                      (* shift packed data left logical by <N> bits *)
  (* I_XsXds *) |                                                  Psllq                      (* shift packed data left logical by <N> bits *)
  (* I_XsXds *) |                                                  Psllw                      (* shift packed data left logical by <N> bits *)
  (* I_XsXds *) |                                                  Psrad                      (* shift packed data right arithmetic by <N> bits *)
  (* I_XsXds *) |                                                  Psraw                      (* shift packed data right arithmetic by <N> bits *)
  (* I_XsXds *) |                                                  Psrld                      (* shift packed data right logical by <N> bits *)
  (* I_XsXds *) |                                                  Psrlq                      (* shift packed data right logical by <N> bits *)
  (* I_XsXds *) |                                                  Psrlw                      (* shift packed data right logical by <N> bits *)
  (* I_XsXds *) |                                                  Psubb                      (* subtract packed bytes *)
  (* I_XsXds *) |                                                  Psubd                      (* subtract packed longs *)
  (* I_XsXds *) |                                                  Psubq                      (* subtract packed quads *)
  (* I_XsXds *) |                                                  Psubsb                     (* subtract packed signed bytes (w/saturation) *)
  (* I_XsXds *) |                                                  Psubsw                     (* subtract packed signed words (w/saturation) *)
  (* I_XsXds *) |                                                  Psubusb                    (* subtract packed unsigned bytes (w/saturation) *)
  (* I_XsXds *) |                                                  Psubusw                    (* subtract packed unsigned words (w/saturation) *)
  (* I_XsXds *) |                                                  Psubw                      (* subtract packed words *)
  (* I_XsXds *) |                                                  Punpckhbw                  (* unpack high bytes to words *)
  (* I_XsXds *) |                                                  Punpckhdq                  (* unpack high longs to quads *)
  (* I_XsXds *) |                                                  Punpckhqdq                 (* unpack high quads to double-quads *)
  (* I_XsXds *) |                                                  Punpckhwd                  (* unpack high words to longs *)
  (* I_XsXds *) |                                                  Punpcklbw                  (* unpack low bytes to words *)
  (* I_XsXds *) |                                                  Punpckldq                  (* unpack low longs to quads *)
  (* I_XsXds *) |                                                  Punpcklqdq                 (* unpack low quads to double-quads *)
  (* I_XsXds *) |                                                  Punpcklwd                  (* unpack low words to longs *)
  (* I_XsXds *) |                                                  Pxor                       (* logical exclusive-OR packed integers *)
  (* I_XsXds *) |                                                  Rcpss                      (* compute reciprocal of scalar float32 *)
  (* I_XsXds *) |                                                  Rsqrtss                    (* compute reciprocal of square root of scalar float32 *)
  (* I_XsXds *) |                                                  Shufpd of Lit              (* shuffle packed float64s *)
  (* I_XsXds *) |                                                  Shufps of Lit              (* shuffle packed float32s *)
  (* I_XsXds *) |                                                  Sqrtsd                     (* compute square root of scalar float64 *)
  (* I_XsXds *) |                                                  Sqrtss                     (* compute square root of scalar float32 *)
  (* I_XsXds *) |                                                  Subpd                      (* subtract packed float64s *)
  (* I_XsXds *) |                                                  Subps                      (* subtract packed float32s *)
  (* I_XsXds *) |                                                  Subsd                      (* subtract scalar float64s *)
  (* I_XsXds *) |                                                  Subss                      (* subtract scalar float32s *)
  (* I_XsXds *) |                                                  Unpckhpd                   (* unpack and interleave high packed float64s *)
  (* I_XsXds *) |                                                  Unpckhps                   (* unpack and interleave high packed float32s *)
  (* I_XsXds *) |                                                  Unpcklpd                   (* unpack and interleave low packed float64s *)
  (* I_XsXds *) |                                                  Unpcklps                   (* unpack and interleave low packed float32s *)
  (* I_XsXds *) |                                                  Xorpd                      (* logical exclusive-OR packed float64s *)
  (* I_XsXds *) |                                                  Xorps                      (* logical exclusive-OR packed float32s *)
  (* I_XsXds *) | [< ISA "AES"                                 >]  Aesdec                     (* perform one round of an AES decryption flow *)
  (* I_XsXds *) | [< ISA "AES"                                 >]  Aesdeclast                 (* perform last round of an AES decryption flow *)
  (* I_XsXds *) | [< ISA "AES"                                 >]  Aesenc                     (* perform one round of an AES encryption flow *)
  (* I_XsXds *) | [< ISA "AES"                                 >]  Aesenclast                 (* perform last round of an AES encryption flow *)
  (* I_XsXds *) | [< ISA "CLMUL"                               >]  Pclmulqdq of Lit           (* carry-less multiply quads *)
  (* I_XsXds *) | [< ISA "SSE3"                                >]  Addsubpd                   (* packed float64 add/subtract *)
  (* I_XsXds *) | [< ISA "SSE3"                                >]  Addsubps                   (* packed float32 add/subtract *)
  (* I_XsXds *) | [< ISA "SSE3"                                >]  Haddpd                     (* horizontal add packed float64s *)
  (* I_XsXds *) | [< ISA "SSE3"                                >]  Haddps                     (* horizontal add packed float32s *)
  (* I_XsXds *) | [< ISA "SSE3"                                >]  Hsubpd                     (* horizontal subtract packed float64s *)
  (* I_XsXds *) | [< ISA "SSE3"                                >]  Hsubps                     (* horizontal subtract packed float32s *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Blendpd of Lit             (* blend packed float64s *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Blendps of Lit             (* blend packed float32s *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Dppd of Lit                (* dot-product of packed float64s *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Dpps of Lit                (* dot-product of packed float32s *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Insertps of Lit            (* insert packed float32 *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Mpsadbw of Lit             (* compute multiple packed sums of absolute difference *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Packusdw                   (* convert packed longs to packed words w/unsigned saturation *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pblendw of Lit             (* blend packed words *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pcmpeqq                    (* arithmetic compare-is-equal quads *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pmaxsb                     (* maximum of packed signed bytes *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pmaxsd                     (* maximum of packed signed longs *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pmaxud                     (* maximum of packed unsigned longs *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pmaxuw                     (* maximum of packed unsigned words *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pminsb                     (* minimum of packed signed bytes *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pminsd                     (* minimum of packed signed longs *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pminud                     (* minimum of packed unsigned longs *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pminuw                     (* minimum of packed unsigned words *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pmuldq                     (* multiply packed signed longs to quads *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Pmulld                     (* multiply packed signed longs and store low result *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Roundsd of Lit             (* round scalar float64 *)
  (* I_XsXds *) | [< ISA "SSE4_1"                              >]  Roundss of Lit             (* round scalar float32 *)
  (* I_XsXds *) | [< ISA "SSE4_1";                   Xs "xmm0" >]  Blendvpd                   (* variable blend packed float64s *)
  (* I_XsXds *) | [< ISA "SSE4_1";                   Xs "xmm0" >]  Blendvps                   (* variable blend packed float32s *)
  (* I_XsXds *) | [< ISA "SSE4_1";                   Xs "xmm0" >]  Pblendvb                   (* variable blend packed bytes *)
  (* I_XsXds *) | [< ISA "SSE4_2"                              >]  Pcmpgtq                    (* arithmetic compare-is-greater-than packed quads *)
  (* I_XsXds *) | [< ISA "SSE4a"                               >]  Extrq                      (* extract field from register *)
  (* I_XsXds *) | [< ISA "SSE4a"                               >]  Insertq                    (* insert field *)
  (* I_XsXds *) | [< ISA "SSE4a";    AKA "insertq"             >]  Insertq_ of Lit * Lit      (* insert field *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Palignr of Lit             (* packed align right *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Phaddd                     (* horizontal add packed longs *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Phaddsw                    (* horizontal add packed words (w/saturation) *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Phaddw                     (* horizontal add packed words *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Phsubd                     (* horizontal subtract packed longs *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Phsubsw                    (* horizontal subtract packed words (w/saturation) *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Phsubw                     (* horizontal subtract packed words *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Pmaddubsw                  (* multiply and add packed signed and unsigned bytes *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Pmulhrsw                   (* packed multiply high w/round and scale *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Pshufb                     (* packed shuffle bytes *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Psignb                     (* packed sign bytes *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Psignd                     (* packed sign longs *)
  (* I_XsXds *) | [< ISA "SSSE3"                               >]  Psignw                     (* packed sign words *)

(*****************************************************************************************************************************************************************************)

type I_M32sXs =   | [<                                                      Fd "Z,P,C,O,A,S" >]  Comiss       (* ordered compare scalar float32s and set RFLAGS *)
  (* I_M32sXs *)  | [<                                                      Fd "Z,P,C,O,A,S" >]  Ucomiss      (* unordered compare scalar float32s and set RFLAGS *)
  (* I_M32sXs *)  | [< ISA "AVX";                                           Fd "Z,P,C,O,A,S" >]  Vcomiss      (* ordered compare scalar float32s and set RFLAGS *)
  (* I_M32sXs *)  | [< ISA "AVX";                                           Fd "Z,P,C,O,A,S" >]  Vucomiss     (* unordered compare scalar float32s and set RFLAGS *)
type I_M64sXs =   | [<                                                      Fd "Z,P,C,O,A,S" >]  Comisd       (* ordered compare scalar float64s and set RFLAGS *)
  (* I_M64sXs *)  | [<                                                      Fd "Z,P,C,O,A,S" >]  Ucomisd      (* unordered compare scalar float64s and set RFLAGS *)
  (* I_M64sXs *)  | [< ISA "AVX";                                           Fd "Z,P,C,O,A,S" >]  Vcomisd      (* ordered compare scalar float64s and set RFLAGS *)
  (* I_M64sXs *)  | [< ISA "AVX";                                           Fd "Z,P,C,O,A,S" >]  Vucomisd     (* unordered compare scalar float64s and set RFLAGS *)
type I_M128sXs =  | [< ISA "SSE4_2";  Rs "rax,rdx";  Rd "rcx";              Fd "C,Z,S,O,A,P" >]  Pcmpestri    (* packed compare explicit length strings, return index *)
  (* I_M128sXs *) | [< ISA "SSE4_2";  Rs "rax,rdx";  Rd "rcx";  Xd "xmm0";  Fd "C,Z,S,O,A,P" >]  Pcmpestrm    (* packed compare explicit length strings, return mask *)
  (* I_M128sXs *) | [< ISA "SSE4_2";                 Rd "rcx";              Fd "C,Z,S,O,A,P" >]  Pcmpistri    (* packed compare implicit length strings, return index *)
  (* I_M128sXs *) | [< ISA "SSE4_2";                 Rd "rcx";  Xd "xmm0";  Fd "C,Z,S,O,A,P" >]  Pcmpistrm    (* packed compare implicit length strings, return mask *)
  (* I_M128sXs *) | [< ISA "AVX";     Rs "rax,rdx";  Rd "rcx";              Fd "C,Z,S,O,A,P" >]  Vpcmpestri   (* packed compare explicit length strings, return index *)
  (* I_M128sXs *) | [< ISA "AVX";     Rs "rax,rdx";  Rd "rcx";  Xd "xmm0";  Fd "C,Z,S,O,A,P" >]  Vpcmpestrm   (* packed compare explicit length strings, return mask *)
  (* I_M128sXs *) | [< ISA "AVX";                    Rd "rcx";              Fd "C,Z,S,O,A,P" >]  Vpcmpistri   (* packed compare implicit length strings, return index *)
  (* I_M128sXs *) | [< ISA "AVX";                    Rd "rcx";  Xd "xmm0";  Fd "C,Z,S,O,A,P" >]  Vpcmpistrm   (* packed compare implicit length strings, return mask *)
  (* I_M128sXs *) | [< ISA "SSE4_1";                                        Fd "O,A,P,S,Z,C" >]  Ptest        (* logical compare *)
  (* I_M128sXs *) | [< ISA "AVX";                                           Fd "O,A,P,S,Z,C" >]  Vptest       (* packed bit test *)
  (* I_M128sXs *) | [< ISA "AVX";                                           Fd "O,A,P,S,Z,C" >]  Vtestpd      (* packed float64 bit test *)
  (* I_M128sXs *) | [< ISA "AVX";                                           Fd "O,A,P,S,Z,C" >]  Vtestps      (* packed float32 bit test *)

(*****************************************************************************************************************************************************************************)

type I_XsXs =  | [<                                                      Fd "Z,P,C,O,A,S" >]  Comisd         (* ordered compare scalar float64s and set RFLAGS *)
  (* I_XsXs *) | [<                                                      Fd "Z,P,C,O,A,S" >]  Comiss         (* ordered compare scalar float32s and set RFLAGS *)
  (* I_XsXs *) | [<                                                      Fd "Z,P,C,O,A,S" >]  Ucomisd        (* unordered compare scalar float64s and set RFLAGS *)
  (* I_XsXs *) | [<                                                      Fd "Z,P,C,O,A,S" >]  Ucomiss        (* unordered compare scalar float32s and set RFLAGS *)
  (* I_XsXs *) | [<                Rs "rdi"                                               >]  Maskmovdqu     (* store selected bytes of double quad w/NT-hint *)
  (* I_XsXs *) | [< ISA "AVX";                                           Fd "O,A,P,S,Z,C" >]  Vptest         (* packed bit test *)
  (* I_XsXs *) | [< ISA "AVX";                                           Fd "O,A,P,S,Z,C" >]  Vtestpd        (* packed float64 bit test *)
  (* I_XsXs *) | [< ISA "AVX";                                           Fd "O,A,P,S,Z,C" >]  Vtestps        (* packed float32 bit test *)
  (* I_XsXs *) | [< ISA "AVX";                                           Fd "Z,P,C,O,A,S" >]  Vcomisd        (* ordered compare scalar float64s and set RFLAGS *)
  (* I_XsXs *) | [< ISA "AVX";                                           Fd "Z,P,C,O,A,S" >]  Vcomiss        (* ordered compare scalar float32s and set RFLAGS *)
  (* I_XsXs *) | [< ISA "AVX";                                           Fd "Z,P,C,O,A,S" >]  Vucomisd       (* unordered compare scalar float64s and set RFLAGS *)
  (* I_XsXs *) | [< ISA "AVX";                                           Fd "Z,P,C,O,A,S" >]  Vucomiss       (* unordered compare scalar float32s and set RFLAGS *)
  (* I_XsXs *) | [< ISA "AVX";                    Rd "rcx";              Fd "C,Z,S,O,A,P" >]  Vpcmpistri     (* packed compare implicit length strings, return index *)
  (* I_XsXs *) | [< ISA "AVX";                    Rd "rcx";  Xd "xmm0";  Fd "C,Z,S,O,A,P" >]  Vpcmpistrm     (* packed compare implicit length strings, return mask *)
  (* I_XsXs *) | [< ISA "AVX";     Rs "rax,rdx";  Rd "rcx";              Fd "C,Z,S,O,A,P" >]  Vpcmpestri     (* packed compare explicit length strings, return index *)
  (* I_XsXs *) | [< ISA "AVX";     Rs "rax,rdx";  Rd "rcx";  Xd "xmm0";  Fd "C,Z,S,O,A,P" >]  Vpcmpestrm     (* packed compare explicit length strings, return mask *)
  (* I_XsXs *) | [< ISA "AVX";     Rs "rdi"                                               >]  Vmaskmovdqu    (* store selected bytes of double quad w/NT-hint *)
  (* I_XsXs *) | [< ISA "SSE4_1";                                        Fd "O,A,P,S,Z,C" >]  Ptest          (* logical compare *)
  (* I_XsXs *) | [< ISA "SSE4_2";                 Rd "rcx";              Fd "C,Z,S,O,A,P" >]  Pcmpistri      (* packed compare implicit length strings, return index *)
  (* I_XsXs *) | [< ISA "SSE4_2";                 Rd "rcx";  Xd "xmm0";  Fd "C,Z,S,O,A,P" >]  Pcmpistrm      (* packed compare implicit length strings, return mask *)
  (* I_XsXs *) | [< ISA "SSE4_2";  Rs "rax,rdx";  Rd "rcx";              Fd "C,Z,S,O,A,P" >]  Pcmpestri      (* packed compare explicit length strings, return index *)
  (* I_XsXs *) | [< ISA "SSE4_2";  Rs "rax,rdx";  Rd "rcx";  Xd "xmm0";  Fd "C,Z,S,O,A,P" >]  Pcmpestrm      (* packed compare explicit length strings, return mask *)

(*****************************************************************************************************************************************************************************)

type I_XsXsXd =  | [< ISA "AVX"       >]  Vaddpd               (* add packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vaddps               (* add packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vaddsd               (* add scalar float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vaddss               (* add scalar float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vaddsubpd            (* mixed-add-subtract packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vaddsubps            (* mixed-add-subtract packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vandnpd              (* bitwise logical AND-NOT of packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vandnps              (* bitwise logical AND-NOT of packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vandpd               (* bitwise logical AND of packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vandps               (* bitwise logical AND of packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vblendpd of Lit      (* blend packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vblendps of Lit      (* blend packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vcmppd of Lit        (* compare packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vcmpps of Lit        (* compare packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vcmpsd of Lit        (* compare scalar float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vcmpss of Lit        (* compare scalar float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vcvtsd2ss            (* convert scalar float64 to float32 *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vcvtss2sd            (* convert scalar float32 to float64 *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vdivpd               (* divide packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vdivps               (* divide packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vdivsd               (* divide scalar float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vdivss               (* divide scalar float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vdppd of Lit         (* dot-product of packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vdpps of Lit         (* dot-product of packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vhaddpd              (* horizontal add packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vhaddps              (* horizontal add packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vhsubpd              (* horizontal sub packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vhsubps              (* horizontal sub packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vinsertps of Lit     (* insert scalar float32 *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmaxpd               (* maximum of packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmaxps               (* maximum of packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmaxsd               (* maximum of scalar float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmaxss               (* maximum of scalar float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vminpd               (* minimum of packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vminps               (* minimum of packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vminsd               (* minimum of scalar float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vminss               (* minimum of scalar float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmovhlps             (* move part of vector *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmovlhps             (* move part of vector *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmovsd               (* move scalar float64 *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmovss               (* move scalar float32 *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmpsadbw of Lit      (* compute multiple packed sums of abs. diff. *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmulpd               (* multiply packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmulps               (* multiply packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmulsd               (* multiply scalar float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vmulss               (* multiply scalar float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vorpd                (* bitwise logical OR of packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vorps                (* bitwise logical OR of packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpackssdw            (* convert packed longs to packed words w/signed saturation *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpacksswb            (* convert packed words to packed bytes w/signed saturation *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpackusdw            (* convert packed longs to packed words w/unsigned saturation *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpackuswb            (* convert packed words to packed bytes w/unsigned saturation *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpaddb               (* add packed bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpaddd               (* add packed longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpaddq               (* add packed quads *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpaddsb              (* add packed signed bytes (w/saturation) *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpaddsw              (* add packed signed words (w/saturation) *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpaddusb             (* add packed unsigned bytes (w/saturation) *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpaddusw             (* add packed unsigned words (w/saturation) *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpaddw               (* add packed words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpalignr of Lit      (* packed align right *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpand                (* logical AND packed integers *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpandn               (* logical AND-NOT packed integers *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpavgb               (* average packed bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpavgw               (* average packed words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpcmpeqb             (* arithmetic compare-is-equal packed bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpcmpeqd             (* arithmetic compare-is-equal packed longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpcmpeqq             (* arithmetic compare-is-equal packed quads *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpcmpeqw             (* arithmetic compare-is-equal packed words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpcmpgtb             (* arithmetic compare-is-greater-than packed bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpcmpgtd             (* arithmetic compare-is-greater-than packed longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpcmpgtq             (* arithmetic compare-is-greater-than packed quads *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpcmpgtw             (* arithmetic compare-is-greater-than packed words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpermilpd            (* permute float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpermilps            (* permute float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vphaddd              (* horizontal add packed longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vphaddsw             (* horizontal add packed words (w/saturation) *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vphaddw              (* horizontal add packed words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vphsubd              (* horizontal subtract packed longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vphsubsw             (* horizontal subtract packed words (w/saturation) *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vphsubw              (* horizontal subtract packed words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmaddubsw           (* multiply and add packed signed and unsigned bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmaddwd             (* multiply and add packed integers *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmaxsb              (* maximum of packed signed bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmaxsd              (* maximum of packed signed longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmaxsw              (* maximum of packed signed words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmaxub              (* maximum of packed unsigned bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmaxud              (* maximum of packed unsigned longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmaxuw              (* maximum of packed unsigned words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpminsb              (* minimum of packed signed bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpminsd              (* minimum of packed signed longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpminsw              (* minimum of packed signed words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpminub              (* minimum of packed unsigned bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpminud              (* minimum of packed unsigned longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpminuw              (* minimum of packed unsigned words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmuldq              (* multiply packed signed longs to quads *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmulhrsw            (* packed multiply high w/round and scale *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmulhuw             (* multiply packed unsigned words and store high result *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmulhw              (* multiply packed signed words and store high result *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmulld              (* multiply packed signed longs and store low result *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmullw              (* multiply packed signed words and store low result *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpmuludq             (* multiply packed integers *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpor                 (* logical inclusive-OR packed integers *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsadbw              (* compute sum of absolute differences *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpshufb              (* packed shuffle bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsignb              (* packed sign bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsignd              (* packed sign longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsignw              (* packed sign words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpslld               (* shift packed data left logical by <N> bits *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsllq               (* shift packed data left logical by <N> bits *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsllw               (* shift packed data left logical by <N> bits *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsrad               (* shift packed data right arithmetic by <N> bits *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsraw               (* shift packed data right arithmetic by <N> bits *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsrld               (* shift packed data right logical by <N> bits *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsrlq               (* shift packed data right logical by <N> bits *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsrlw               (* shift packed data right logical by <N> bits *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsubb               (* subtract packed bytes *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsubd               (* subtract packed longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsubq               (* subtract packed quads *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsubsb              (* subtract packed signed bytes (w/saturation) *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsubsw              (* subtract packed signed words (w/saturation) *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsubusb             (* subtract packed unsigned bytes (w/saturation) *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsubusw             (* subtract packed unsigned words (w/saturation) *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpsubw               (* subtract packed words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpunpckhbw           (* unpack high bytes to words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpunpckhdq           (* unpack high longs to quads *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpunpckhqdq          (* unpack high quads to double-quads *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpunpckhwd           (* unpack high words to longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpunpcklbw           (* unpack low bytes to words *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpunpckldq           (* unpack low longs to quads *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpunpcklqdq          (* unpack low quads to double-quads *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpunpcklwd           (* unpack low words to longs *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vpxor                (* logical exclusive-OR packed integers *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vrcpss               (* reciprocate of scalar float32 *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vroundsd of Lit      (* round scalar float64 *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vroundss of Lit      (* round scalar float32 *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vrsqrtss             (* reciprocate sqrt of scalar float32 *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vshufpd of Lit       (* shuffle packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vshufps of Lit       (* shuffle packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vsqrtsd              (* compute square root of scalar float64 *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vsqrtss              (* compute square root of scalar float32 *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vsubpd               (* subtract packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vsubps               (* subtract packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vsubsd               (* subtract scalar float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vsubss               (* subtract scalar float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vunpckhpd            (* unpack and interleave high packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vunpckhps            (* unpack and interleave high packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vunpcklpd            (* unpack and interleave low packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vunpcklps            (* unpack and interleave low packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vxorpd               (* logical exclusive-OR packed float64s *)
  (* I_XsXsXd *) | [< ISA "AVX"       >]  Vxorps               (* logical exclusive-OR packed float32s *)
  (* I_XsXsXd *) | [< ISA "AVX,AES"   >]  Vaesdec              (* perform one round of an AES decryption flow *)
  (* I_XsXsXd *) | [< ISA "AVX,AES"   >]  Vaesdeclast          (* perform last round of an AES decryption flow *)
  (* I_XsXsXd *) | [< ISA "AVX,AES"   >]  Vaesenc              (* perform one round of an AES encryption flow *)
  (* I_XsXsXd *) | [< ISA "AVX,AES"   >]  Vaesenclast          (* perform last round of an AES encryption flow *)
  (* I_XsXsXd *) | [< ISA "AVX,CLMUL" >]  Vpclmulqdq of Lit    (* carry-less multiply quads *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpcomb of Lit        (* compare vector signed bytes *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpcomd of Lit        (* compare vector signed longs *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpcomq of Lit        (* compare vector signed quads *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpcomub of Lit       (* compare vector unsigned bytes *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpcomud of Lit       (* compare vector unsigned longs *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpcomuq of Lit       (* compare vector unsigned quads *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpcomuw of Lit       (* compare vector unsigned words *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpcomw of Lit        (* compare vector signed words *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vprotb               (* packed rotate bytes *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vprotd               (* packed rotate longs *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vprotq               (* packed rotate quads *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vprotw               (* packed rotate words *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpshab               (* packed shift arithmetic bytes *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpshad               (* packed shift arithmetic longs *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpshaq               (* packed shift arithmetic quads *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpshaw               (* packed shift arithmetic words *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpshlb               (* packed shift logical bytes *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpshld               (* packed shift logical longs *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpshlq               (* packed shift logical quads *)
  (* I_XsXsXd *) | [< ISA "XOP"       >]  Vpshlw               (* packed shift logical words *)

(*****************************************************************************************************************************************************************************)

type I_XsXsM128ds =  | [< ISA "AVX" >]  Vmaskmovpd    (* conditional store packed float64s using mask *) 
  (* I_XsXsM128ds *) | [< ISA "AVX" >]  Vmaskmovps    (* conditional store packed float32s using mask *) 

(*****************************************************************************************************************************************************************************)

type I_M8sXsXd =    | [< ISA "AVX"                                  >]  Vpinsrb of Lit       (* insert byte into vector *)
type I_M16sXsXd =   | [< ISA "AVX"                                  >]  Vpinsrw of Lit       (* insert word into vector *)
type I_M32sXsXd =   | [< ISA "AVX"                                  >]  Vaddss               (* add scalar float32s *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vcmpss of Lit        (* compare scalar float32s *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vdivss               (* divide scalar float32s *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vmaxss               (* maximum of scalar float32s *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vminss               (* minimum of scalar float32s *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vmulss               (* multiply scalar float32s *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vrcpss               (* compute reciprocal of scalar float32 *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vroundss of Lit      (* round scalar float32 *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vrsqrtss             (* compute reciprocal of square root of scalar float32 *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vsqrtss              (* compute square root of scalar float32 *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vsubss               (* subtract scalar float32s *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vinsertps of Lit     (* insert scalar float32 *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vpinsrd of Lit       (* insert long into vector *)
  (* I_M32sXsXd *)  | [< ISA "AVX";      AKA "vcvtsi2sd"            >]  Vcvtsi2sdl           (* convert scalar longs to float64 *)
  (* I_M32sXsXd *)  | [< ISA "AVX"                                  >]  Vcvtss2sd            (* convert scalar float32 to float64 *)
  (* I_M32sXsXd *)  | [< ISA "AVX";      AKA "vcvtsi2ss"            >]  Vcvtsi2ssl           (* convert longs to scalar float32 *)
type I_M64sXsXd =   | [< ISA "AVX"                                  >]  Vaddsd               (* add scalar float64s *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vcmpsd of Lit        (* compare scalar float64s *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vdivsd               (* divide scalar float64s *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vmaxsd               (* maximum of scalar float64s *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vminsd               (* minimum of scalar float64s *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vmulsd               (* multiply scalar float64s *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vmovlps              (* load low parts of packed float32s *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vmovlpd              (* load low parts of packed float64s *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vmovhps              (* load high parts of packed float32s *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vmovhpd              (* load high parts of packed float64s *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vroundsd of Lit      (* round scalar float64 *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vsqrtsd              (* compute square root of scalar float64 *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vsubsd               (* subtract scalar float64s *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vpinsrq of Lit       (* insert quad into vector *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vcvtsd2ss            (* convert float64 to float32 *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vcvtsi2sdq           (* convert quad to float64 *)
  (* I_M64sXsXd *)  | [< ISA "AVX"                                  >]  Vcvtsi2ssq           (* convert quad to scalar float32 *)
type I_M128sXsXd =  | [< ISA "AVX"                                  >]  Vaddpd               (* add packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vaddps               (* add packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vaddsubpd            (* mixed-add-subtract packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vaddsubps            (* mixed-add-subtract packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX,AES"                              >]  Vaesdec              (* perform one round of an AES decryption flow *)
  (* I_M128sXsXd *) | [< ISA "AVX,AES"                              >]  Vaesdeclast          (* perform last round of an AES decryption flow *)
  (* I_M128sXsXd *) | [< ISA "AVX,AES"                              >]  Vaesenc              (* perform one round of an AES encryption flow *)
  (* I_M128sXsXd *) | [< ISA "AVX,AES"                              >]  Vaesenclast          (* perform last round of an AES encryption flow *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vandnpd              (* bitwise logical AND-NOT of packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vandnps              (* bitwise logical AND-NOT of packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vandpd               (* bitwise logical AND of packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vandps               (* bitwise logical AND of packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vblendpd of Lit      (* blend packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vblendps of Lit      (* blend packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vcmppd of Lit        (* compare packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vcmpps of Lit        (* compare packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vdivpd               (* divide packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vdivps               (* divide packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vdppd of Lit         (* dot-product of packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vdpps of Lit         (* dot-product of packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vhaddpd              (* horizontal add packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vhaddps              (* horizontal add packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vhsubpd              (* horizontal sub packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vhsubps              (* horizontal sub packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vmaskmovpd           (* conditional load packed float64s using mask *) 
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vmaskmovps           (* conditional load packed float32s using mask *) 
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vmaxpd               (* maximum of packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vmaxps               (* maximum of packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vminpd               (* minimum of packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vminps               (* minimum of packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vmpsadbw of Lit      (* compute multiple packed sums of absolute difference *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vmulpd               (* multiply packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vmulps               (* multiply packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vorpd                (* bitwise logical OR of packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vorps                (* bitwise logical OR of packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpackssdw            (* convert packed longs to packed words w/signed saturation *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpacksswb            (* convert packed words to packed bytes w/signed saturation *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpackusdw            (* convert packed longs to packed words w/unsigned saturation *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpackuswb            (* convert packed words to packed bytes w/unsigned saturation *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpaddb               (* add packed bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpaddd               (* add packed longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpaddq               (* add packed quads *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpaddsb              (* add packed signed bytes (w/saturation) *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpaddsw              (* add packed signed words (w/saturation) *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpaddusb             (* add packed unsigned bytes (w/saturation) *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpaddusw             (* add packed unsigned words (w/saturation) *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpaddw               (* add packed words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpalignr of Lit      (* packed align right *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpand                (* logical AND packed integers *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpandn               (* logical AND-NOT packed integers *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpavgb               (* average packed bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpavgw               (* average packed words *)
  (* I_M128sXsXd *) | [< ISA "AVX,CLMUL"                            >]  Vpclmulqdq of Lit    (* carry-less multiply quads *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpcmpeqb             (* arithmetic compare-is-equal packed bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpcmpeqd             (* arithmetic compare-is-equal packed longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpcmpeqq             (* arithmetic compare-is-equal packed quads *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpcmpeqw             (* arithmetic compare-is-equal packed words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpcmpgtb             (* arithmetic compare-is-greater-than packed bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpcmpgtd             (* arithmetic compare-is-greater-than packed longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpcmpgtq             (* arithmetic compare-is-greater-than packed quads *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpcmpgtw             (* arithmetic compare-is-greater-than packed words *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpcomb of Lit        (* compare vector signed bytes *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpcomd of Lit        (* compare vector signed longs *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpcomq of Lit        (* compare vector signed quads *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpcomub of Lit       (* compare vector unsigned bytes *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpcomud of Lit       (* compare vector unsigned longs *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpcomuq of Lit       (* compare vector unsigned quads *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpcomuw of Lit       (* compare vector unsigned words *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpcomw of Lit        (* compare vector signed words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpermilpd            (* permute float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpermilps            (* permute float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vphaddd              (* horizontal add packed longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vphaddsw             (* horizontal add packed words (w/saturation) *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vphaddw              (* horizontal add packed words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vphsubd              (* horizontal subtract packed longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vphsubsw             (* horizontal subtract packed words (w/saturation) *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vphsubw              (* horizontal subtract packed words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmaddubsw           (* multiply and add packed signed and unsigned bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmaddwd             (* multiply and add packed integers *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmaxsb              (* maximum of packed signed bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmaxsd              (* maximum of packed signed longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmaxsw              (* maximum of packed signed words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmaxub              (* maximum of packed unsigned bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmaxud              (* maximum of packed unsigned longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmaxuw              (* maximum of packed unsigned words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpminsb              (* minimum of packed signed bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpminsd              (* minimum of packed signed longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpminsw              (* minimum of packed signed words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpminub              (* minimum of packed unsigned bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpminud              (* minimum of packed unsigned longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpminuw              (* minimum of packed unsigned words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmuldq              (* multiply packed signed longs to quads *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmulhrsw            (* packed multiply high w/round and scale *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmulhuw             (* multiply packed unsigned words and store high result *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmulhw              (* multiply packed signed words and store high result *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmulld              (* multiply packed signed longs and store low result *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmullw              (* multiply packed signed words and store low result *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpmuludq             (* multiply packed integers *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpor                 (* logical inclusive-OR packed integers *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vprotb               (* packed rotate bytes *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vprotd               (* packed rotate longs *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vprotq               (* packed rotate quads *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vprotw               (* packed rotate words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsadbw              (* compute sum of absolute differences *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpshab               (* packed shift arithmetic bytes *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpshad               (* packed shift arithmetic longs *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpshaq               (* packed shift arithmetic quads *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpshaw               (* packed shift arithmetic words *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpshlb               (* packed shift logical bytes *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpshld               (* packed shift logical longs *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpshlq               (* packed shift logical quads *)
  (* I_M128sXsXd *) | [< ISA "XOP"                                  >]  Vpshlw               (* packed shift logical words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpshufb              (* packed shuffle bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsignb              (* packed sign bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsignd              (* packed sign longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsignw              (* packed sign words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsubb               (* subtract packed bytes *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsubd               (* subtract packed longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsubq               (* subtract packed quads *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsubsb              (* subtract packed signed bytes (w/saturation) *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsubsw              (* subtract packed signed words (w/saturation) *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsubusb             (* subtract packed unsigned bytes (w/saturation) *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsubusw             (* subtract packed unsigned words (w/saturation) *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsubw               (* subtract packed words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpunpckhbw           (* unpack high bytes to words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpunpckhdq           (* unpack high longs to quads *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpunpckhqdq          (* unpack high quads to double-quads *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpunpckhwd           (* unpack high words to longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpunpcklbw           (* unpack low bytes to words *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpunpckldq           (* unpack low longs to quads *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpunpcklqdq          (* unpack low quads to double-quads *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpunpcklwd           (* unpack low words to longs *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpxor                (* logical exclusive-OR packed integers *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vshufpd of Lit       (* shuffle packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vshufps of Lit       (* shuffle packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vsubpd               (* subtract packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vsubps               (* subtract packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vunpckhpd            (* unpack and interleave high packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vunpckhps            (* unpack and interleave high packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vunpcklpd            (* unpack and interleave low packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vunpcklps            (* unpack and interleave low packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vxorpd               (* logical exclusive-OR packed float64s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vxorps               (* logical exclusive-OR packed float32s *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpslld               (* shift packed data left logical by <N> bits *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsllq               (* shift packed data left logical by <N> bits *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsllw               (* shift packed data left logical by <N> bits *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsrad               (* shift packed data right arithmetic by <N> bits *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsraw               (* shift packed data right arithmetic by <N> bits *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsrld               (* shift packed data right logical by <N> bits *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsrlq               (* shift packed data right logical by <N> bits *)
  (* I_M128sXsXd *) | [< ISA "AVX"                                  >]  Vpsrlw               (* shift packed data right logical by <N> bits *)

(*****************************************************************************************************************************************************************************)

type I_XsM128sXd =  | [< ISA "XOP" >]  Vprotb    (* packed rotate bytes *)
  (* I_XsM128sXd *) | [< ISA "XOP" >]  Vprotd    (* packed rotate longs *)
  (* I_XsM128sXd *) | [< ISA "XOP" >]  Vprotq    (* packed rotate quads *)
  (* I_XsM128sXd *) | [< ISA "XOP" >]  Vprotw    (* packed rotate words *)
  (* I_XsM128sXd *) | [< ISA "XOP" >]  Vpshab    (* packed shift arithmetic bytes *)
  (* I_XsM128sXd *) | [< ISA "XOP" >]  Vpshad    (* packed shift arithmetic longs *)
  (* I_XsM128sXd *) | [< ISA "XOP" >]  Vpshaq    (* packed shift arithmetic quads *)
  (* I_XsM128sXd *) | [< ISA "XOP" >]  Vpshaw    (* packed shift arithmetic words *)
  (* I_XsM128sXd *) | [< ISA "XOP" >]  Vpshlb    (* packed shift logical bytes *)
  (* I_XsM128sXd *) | [< ISA "XOP" >]  Vpshld    (* packed shift logical longs *)
  (* I_XsM128sXd *) | [< ISA "XOP" >]  Vpshlq    (* packed shift logical quads *)
  (* I_XsM128sXd *) | [< ISA "XOP" >]  Vpshlw    (* packed shift logical words *)

(*****************************************************************************************************************************************************************************)

type I_XsXsXsXd =  | [< ISA "AVX"  >]  Vblendvpd            (* blend packed float64s *)
  (* I_XsXsXsXd *) | [< ISA "AVX"  >]  Vblendvps            (* blend packed float32s *)
  (* I_XsXsXsXd *) | [< ISA "AVX"  >]  Vpblendvb            (* variable blend packed bytes *)
  (* I_XsXsXsXd *) | [< ISA "AVX"  >]  Vpblendw of Lit      (* blend packed words *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmaddpd             (* multiply and add packed float64 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmaddps             (* multiply and add packed float32 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmaddsd             (* multiply and add scalar float64 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmaddss             (* multiply and add scalar float32 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmaddsubpd          (* multiply w/alternating add/subtract of packed float64 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmaddsubps          (* multiply w/alternating add/subtract of packed float32 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmsubaddpd          (* multiply w/alternating subtract/add of packed float64 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmsubaddps          (* multiply w/alternating subtract/add of packed float32 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmsubpd             (* multiply and subtract packed float64 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmsubps             (* multiply and subtract packed float32 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmsubsd             (* multiply and subtract scalar float64 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfmsubss             (* multiply and subtract scalar float32 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfnmaddpd            (* negative multiply and add packed float64 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfnmaddps            (* negative multiply and add packed float32 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfnmaddsd            (* negative multiply and add scalar float64 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfnmaddss            (* negative multiply and add scalar float32 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfnmsubpd            (* negative multiply and subtract packed float64 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfnmsubps            (* negative multiply and subtract packed float32 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfnmsubsd            (* negative multiply and subtract scalar float64 *)
  (* I_XsXsXsXd *) | [< ISA "FMA4" >]  Vfnmsubss            (* negative multiply and subtract scalar float32 *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpcmov               (* vector conditional moves *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpermil2pd of Lit    (* permute two-source float64s *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpermil2ps of Lit    (* permute two-source float32s *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmacsdd             (* packed multiply accumulate signed long to signed long *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmacsdqh            (* packed multiply accumulate signed high long to signed quad *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmacsdql            (* packed multiply accumulate signed low long to signed quad *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmacssdd            (* packed multiply accumulate signed long to signed long w/saturation *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmacssdqh           (* packed multiply accumulate signed high long to signed quad w/saturation *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmacssdql           (* packed multiply accumulate signed low long to signed quad w/saturation *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmacsswd            (* packed multiply accumulate signed word to signed long w/saturation *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmacssww            (* packed multiply accumulate signed word to signed word w/saturation *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmacswd             (* packed multiply accumulate signed word to signed long *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmacsww             (* packed multiply accumulate signed word to signed word *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmadcsswd           (* packed multiply, add & accumulate signed word to signed long w/saturation *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpmadcswd            (* packed multiply add & accumulate signed word to signed long *)
  (* I_XsXsXsXd *) | [< ISA "XOP"  >]  Vpperm               (* packed permute bytes *)

(*****************************************************************************************************************************************************************************)

type I_XsM32sXsXd =   | [< ISA "FMA4" >]  Vfmaddss             (* multiply and add scalar float32 *)
  (* I_XsM32sXsXd *)  | [< ISA "FMA4" >]  Vfmsubss             (* multiply and subtract scalar float32 *)
  (* I_XsM32sXsXd *)  | [< ISA "FMA4" >]  Vfnmaddss            (* negative multiply and add scalar float32 *)
  (* I_XsM32sXsXd *)  | [< ISA "FMA4" >]  Vfnmsubss            (* negative multiply and subtract scalar float32 *)
type I_XsM64sXsXd =   | [< ISA "FMA4" >]  Vfmaddsd             (* multiply and add scalar float64 *)
  (* I_XsM64sXsXd *)  | [< ISA "FMA4" >]  Vfmsubsd             (* multiply and subtract scalar float64 *)
  (* I_XsM64sXsXd *)  | [< ISA "FMA4" >]  Vfnmaddsd            (* negative multiply and add scalar float64 *)
  (* I_XsM64sXsXd *)  | [< ISA "FMA4" >]  Vfnmsubsd            (* negative multiply and subtract scalar float64 *)
type I_XsM128sXsXd =  | [< ISA "AVX"  >]  Vblendvpd            (* blend packed float64s *)
  (* I_XsM128sXsXd *) | [< ISA "AVX"  >]  Vblendvps            (* blend packed float32s *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfmaddpd             (* multiply and add packed float64 *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfmaddps             (* multiply and add packed float32 *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfmaddsubpd          (* multiply w/alternating add/subtract of packed float64 *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfmaddsubps          (* multiply w/alternating add/subtract of packed float32 *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfmsubaddpd          (* multiply w/alternating subtract/add of packed float64 *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfmsubaddps          (* multiply w/alternating subtract/add of packed float32 *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfmsubpd             (* multiply and subtract packed float64 *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfmsubps             (* multiply and subtract packed float32 *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfnmaddpd            (* negative multiply and add packed float64 *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfnmaddps            (* negative multiply and add packed float32 *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfnmsubpd            (* negative multiply and subtract packed float64 *)
  (* I_XsM128sXsXd *) | [< ISA "FMA4" >]  Vfnmsubps            (* negative multiply and subtract packed float32 *)
  (* I_XsM128sXsXd *) | [< ISA "AVX"  >]  Vpblendvb            (* variable blend packed bytes *)
  (* I_XsM128sXsXd *) | [< ISA "AVX"  >]  Vpblendw of Lit      (* blend packed words *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpcmov               (* vector conditional moves *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpermil2pd of Lit    (* permute two-source float64s *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpermil2ps of Lit    (* permute two-source float32s *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmacsdd             (* packed multiply accumulate signed long to signed long *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmacsdqh            (* packed multiply accumulate signed high long to signed quad *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmacsdql            (* packed multiply accumulate signed low long to signed quad *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmacssdd            (* packed multiply accumulate signed long to signed long w/saturation *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmacssdqh           (* packed multiply accumulate signed high long to signed quad w/saturation *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmacssdql           (* packed multiply accumulate signed low long to signed quad w/saturation *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmacsswd            (* packed multiply accumulate signed word to signed long w/saturation *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmacssww            (* packed multiply accumulate signed word to signed word w/saturation *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmacswd             (* packed multiply accumulate signed word to signed long *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmacsww             (* packed multiply accumulate signed word to signed word *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmadcsswd           (* packed multiply, add & accumulate signed word to signed long w/saturation *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpmadcswd            (* packed multiply add & accumulate signed word to signed long *)
  (* I_XsM128sXsXd *) | [< ISA "XOP"  >]  Vpperm               (* packed permute bytes *)

(*****************************************************************************************************************************************************************************)

type I_M32sXsXsXd =   | [< ISA "FMA4" >]  Vfmaddss             (* multiply and add scalar float32 *)
  (* I_M32sXsXsXd *)  | [< ISA "FMA4" >]  Vfmsubss             (* multiply and subtract scalar float32 *)
  (* I_M32sXsXsXd *)  | [< ISA "FMA4" >]  Vfnmaddss            (* negative multiply and add scalar float32 *)
  (* I_M32sXsXsXd *)  | [< ISA "FMA4" >]  Vfnmsubss            (* negative multiply and subtract scalar float32 *)
type I_M64sXsXsXd =   | [< ISA "FMA4" >]  Vfmaddsd             (* multiply and add scalar float64 *)
  (* I_M64sXsXsXd *)  | [< ISA "FMA4" >]  Vfmsubsd             (* multiply and subtract scalar float64 *)
  (* I_M64sXsXsXd *)  | [< ISA "FMA4" >]  Vfnmaddsd            (* negative multiply and add scalar float64 *)
  (* I_M64sXsXsXd *)  | [< ISA "FMA4" >]  Vfnmsubsd            (* negative multiply and subtract scalar float64 *)
type I_M128sXsXsXd =  | [< ISA "FMA4" >]  Vfmaddpd             (* multiply and add packed float64 *)
  (* I_M128sXsXsXd *) | [< ISA "FMA4" >]  Vfmaddps             (* multiply and add packed float32 *)
  (* I_M128sXsXsXd *) | [< ISA "FMA4" >]  Vfmaddsubpd          (* multiply w/alternating add/subtract of packed float64 *)
  (* I_M128sXsXsXd *) | [< ISA "FMA4" >]  Vfmaddsubps          (* multiply w/alternating add/subtract of packed float32 *)
  (* I_M128sXsXsXd *) | [< ISA "FMA4" >]  Vfmsubaddpd          (* multiply w/alternating subtract/add of packed float64 *)
  (* I_M128sXsXsXd *) | [< ISA "FMA4" >]  Vfmsubaddps          (* multiply w/alternating subtract/add of packed float32 *)
  (* I_M128sXsXsXd *) | [< ISA "FMA4" >]  Vfmsubpd             (* multiply and subtract packed float64 *)
  (* I_M128sXsXsXd *) | [< ISA "FMA4" >]  Vfmsubps             (* multiply and subtract packed float32 *)
  (* I_M128sXsXsXd *) | [< ISA "FMA4" >]  Vfnmaddpd            (* negative multiply and add packed float64 *)
  (* I_M128sXsXsXd *) | [< ISA "FMA4" >]  Vfnmaddps            (* negative multiply and add packed float32 *)
  (* I_M128sXsXsXd *) | [< ISA "FMA4" >]  Vfnmsubpd            (* negative multiply and subtract packed float64 *)
  (* I_M128sXsXsXd *) | [< ISA "FMA4" >]  Vfnmsubps            (* negative multiply and subtract packed float32 *)
  (* I_M128sXsXsXd *) | [< ISA "XOP"  >]  Vpermil2pd of Lit    (* permute two-source float64s *)
  (* I_M128sXsXsXd *) | [< ISA "XOP"  >]  Vpermil2ps of Lit    (* permute two-source float32s *)
  (* I_M128sXsXsXd *) | [< ISA "XOP"  >]  Vpperm               (* packed permute bytes *)

(*****************************************************************************************************************************************************************************)

(* 256-bit AVX vector instructions *)

type I_XsYsYd = | [< ISA "AVX" >]  Vinsertf128 of Lit    (* insert packed FP values *)

(*****************************************************************************************************************************************************************************)

type I_YsXd = | [< ISA "AVX" >]  Vextractf128 of Lit    (* extract packed FP values *)

(*****************************************************************************************************************************************************************************)

type I_M32sYd =   | [< ISA "AVX" >]  Vbroadcastss       (* load float32 and broadcast *)
type I_M64sYd =   | [< ISA "AVX" >]  Vbroadcastsd       (* load float64 and broadcast *)
type I_M128sYd =  | [< ISA "AVX" >]  Vbroadcastf128     (* load 128-bit and broadcast *)
type I_M256sYd =  | [< ISA "AVX" >]  Vlddqu             (* unaligned load integer *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vmovapd            (* aligned load packed float64s *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vmovaps            (* aligned load packed float32s *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vmovdqa            (* aligned load packed integers *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vmovdqu            (* unaligned load packed integers *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vmovupd            (* unaligned load packed float64s *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vmovups            (* unaligned load packed float32s *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vmovddup           (* copy float64s and duplicate *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vrcpps             (* reciprocate packed float32s *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vroundpd of Lit    (* round packed float64s *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vroundps of Lit    (* round packed float32s *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vrsqrtps           (* reciprocate sqrts of packed float32s *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vpermilpd of Lit   (* permute float64s *)
  (* I_M256sYd *) | [< ISA "AVX" >]  Vpermilps of Lit   (* permute float32s *)

(*****************************************************************************************************************************************************************************)

type I_YsYd =  | [< ISA "AVX" >]  Vmovddup           (* copy float64s and duplicate *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vmovapd            (* aligned load packed float64s *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vmovaps            (* aligned load packed float32s *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vmovdqa            (* aligned load packed integers *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vmovdqu            (* unaligned load packed integers *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vmovupd            (* unaligned load packed float64s *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vmovups            (* unaligned load packed float32s *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vrcpps             (* reciprocate packed float32s *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vroundpd of Lit    (* round packed float64s *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vroundps of Lit    (* round packed float32s *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vrsqrtps           (* reciprocate sqrts of packed float32s *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vpermilpd of Lit   (* permute float64s *)
  (* I_YsYd *) | [< ISA "AVX" >]  Vpermilps of Lit   (* permute float32s *)

(*****************************************************************************************************************************************************************************)

type I_YsM128d =  | [< ISA "AVX" >]  Vextractf128 of Lit    (* extract packed FP values *)
type I_YsM256d =  | [< ISA "AVX" >]  Vmovapd                (* aligned store packed float64s *)
  (* I_XsM256d *) | [< ISA "AVX" >]  Vmovaps                (* aligned store packed float32s *)
  (* I_XsM256d *) | [< ISA "AVX" >]  Vmovdqa                (* load data *)
  (* I_XsM256d *) | [< ISA "AVX" >]  Vmovdqu                (* load data *)
  (* I_XsM256d *) | [< ISA "AVX" >]  Vmovntdq               (* store double-quad w/NT-hint *)
  (* I_XsM256d *) | [< ISA "AVX" >]  Vmovntpd               (* store packed float64s w/NT-hint *)
  (* I_XsM256d *) | [< ISA "AVX" >]  Vmovntps               (* store packed float32s w/NT-hint *)
  (* I_XsM256d *) | [< ISA "AVX" >]  Vmovupd                (* unaligned store packed float64s *)
  (* I_XsM256d *) | [< ISA "AVX" >]  Vmovups                (* unaligned store packed float32s *)
 
(*****************************************************************************************************************************************************************************)

type I_YsR32d =  | [< ISA "AVX";  AKA "Vmovmskpd" >]  Vmovmskpsl    (* extract float64s sign masks *)
  (* I_YsR32d *) | [< ISA "AVX";  AKA "Vmovmskps" >]  Vmovmskpdl    (* extract float32s sign masks *)
type I_YsR64d =  | [< ISA "AVX";  AKA "Vmovmskpd" >]  Vmovmskpdq    (* extract float64s sign masks *)
  (* I_YsR64d *) | [< ISA "AVX";  AKA "Vmovmskps" >]  Vmovmskpsq    (* extract float32s sign masks *)

(*****************************************************************************************************************************************************************************)

type I_YsYsYd =  | [< ISA "AVX" >]  Vperm2f128 of Lit    (* permute floating-point values *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vpermilpd            (* permute float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vpermilps            (* permute float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vaddpd               (* add packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vaddps               (* add packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vaddsubpd            (* mixed-add-subtract packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vaddsubps            (* mixed-add-subtract packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vandnpd              (* bitwise logical AND-NOT of packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vandnps              (* bitwise logical AND-NOT of packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vandpd               (* bitwise logical AND of packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vandps               (* bitwise logical AND of packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vblendpd of Lit      (* blend packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vblendps of Lit      (* blend packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vcmppd of Lit        (* compare packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vcmpps of Lit        (* compare packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vdivpd               (* divide packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vdivps               (* divide packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vdppd of Lit         (* dot-product of packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vdpps of Lit         (* dot-product of packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vhaddpd              (* horizontal add packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vhaddps              (* horizontal add packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vhsubpd              (* horizontal sub packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vhsubps              (* horizontal sub packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vmaxpd               (* maximum of packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vmaxps               (* maximum of packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vminpd               (* minimum of packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vminps               (* minimum of packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vmulpd               (* multiply packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vmulps               (* multiply packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vorpd                (* bitwise logical OR of packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vorps                (* bitwise logical OR of packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vshufpd of Lit       (* shuffle packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vshufps of Lit       (* shuffle packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vsubpd               (* subtract packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vsubps               (* subtract packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vunpckhpd            (* unpack and interleave high packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vunpckhps            (* unpack and interleave high packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vunpcklpd            (* unpack and interleave low packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vunpcklps            (* unpack and interleave low packed float32s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vxorpd               (* logical exclusive-OR packed float64s *)
  (* I_YsYsYd *) | [< ISA "AVX" >]  Vxorps               (* logical exclusive-OR packed float32s *)
  
(*****************************************************************************************************************************************************************************)

type I_M128sYsYd = | [< ISA "AVX" >]  Vinsertf128 of Lit    (* insert packed FP values *)

(*****************************************************************************************************************************************************************************)

type I_M256sYsYd =  | [< ISA "AVX" >]  Vperm2f128 of Lit    (* permute floating-point values *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vpermilpd            (* permute float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vpermilps            (* permute float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vaddpd               (* add packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vaddps               (* add packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vaddsubpd            (* mixed-add-subtract packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vaddsubps            (* mixed-add-subtract packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vandnpd              (* bitwise logical AND-NOT of packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vandnps              (* bitwise logical AND-NOT of packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vandpd               (* bitwise logical AND of packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vandps               (* bitwise logical AND of packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vblendpd of Lit      (* blend packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vblendps of Lit      (* blend packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vcmppd of Lit        (* compare packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vcmpps of Lit        (* compare packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vdivpd               (* divide packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vdivps               (* divide packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vdppd of Lit         (* dot-product of packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vdpps of Lit         (* dot-product of packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vhaddpd              (* horizontal add packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vhaddps              (* horizontal add packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vhsubpd              (* horizontal sub packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vhsubps              (* horizontal sub packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vmaxpd               (* maximum of packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vmaxps               (* maximum of packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vminpd               (* minimum of packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vminps               (* minimum of packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vmulpd               (* multiply packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vmulps               (* multiply packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vorpd                (* bitwise logical OR of packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vorps                (* bitwise logical OR of packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vshufpd of Lit       (* shuffle packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vshufps of Lit       (* shuffle packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vsubpd               (* subtract packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vsubps               (* subtract packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vunpckhpd            (* unpack and interleave high packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vunpckhps            (* unpack and interleave high packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vunpcklpd            (* unpack and interleave low packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vunpcklps            (* unpack and interleave low packed float32s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vxorpd               (* logical exclusive-OR packed float64s *)
  (* I_M256sYsYd *) | [< ISA "AVX" >]  Vxorps               (* logical exclusive-OR packed float32s *)

(*****************************************************************************************************************************************************************************)

type Instr =  |            I0           of I_I0                                          (*"  pushfq                                       "*)
  (* Instr *) | [< Vec >]  V0           of I_V0                                          (*"  vzeroupper                                   "*)
  (* Instr *) |            C            of I_C            * Iargs_S<Iexpr>               (*"  jng             .L1                          "*)
  (* Instr *) |            M8d          of I_M8d          * Iargs_D<M8>                  (*"  movb            $1, (%rax)                   "*)
  (* Instr *) |            M16d         of I_M16d         * Iargs_D<M16>                 (*"  movw            $1, (%rax)                   "*)
  (* Instr *) |            M32d         of I_M32d         * Iargs_D<M32>                 (*"  movl            $1, (%rax)                   "*)
  (* Instr *) |            M64d         of I_M64d         * Iargs_D<M64>                 (*"  movq            $1, (%rax)                   "*)
  (* Instr *) |            M8ds         of I_M8ds         * Iargs_DS<M8>                 (*"  addb            $2, (%rax)                   "*)
  (* Instr *) |            M16ds        of I_M16ds        * Iargs_DS<M16>                (*"  addw            $2, (%rax)                   "*)
  (* Instr *) |            M32ds        of I_M32ds        * Iargs_DS<M32>                (*"  addl            $2, (%rax)                   "*)
  (* Instr *) |            M64ds        of I_M64ds        * Iargs_DS<M64>                (*"  addq            $2, (%rax)                   "*)
  (* Instr *) |            M128ds       of I_M128ds       * Iargs_DS<M128>               (*"  cmpxchg16       (%rdi)                       "*)
  (* Instr *) |            R8d          of I_R8d          * Iargs_D<R8>                  (*"  movb            $1, %al                      "*)
  (* Instr *) |            R16d         of I_R16d         * Iargs_D<R16>                 (*"  movw            $1, %dx                      "*)
  (* Instr *) |            R32d         of I_R32d         * Iargs_D<R32>                 (*"  movl            $1, %esi                     "*)
  (* Instr *) |            R64d         of I_R64d         * Iargs_D<R64>                 (*"  movq            $1, %r8                      "*)
  (* Instr *) |            R8ds         of I_R8ds         * Iargs_DS<R8>                 (*"  negb            %dh                          "*)
  (* Instr *) |            R16ds        of I_R16ds        * Iargs_DS<R16>                (*"  notw            %si                          "*)
  (* Instr *) |            R32ds        of I_R32ds        * Iargs_DS<R32>                (*"  bswapl          %eax                         "*)
  (* Instr *) |            R64ds        of I_R64ds        * Iargs_DS<R64>                (*"  bswapq          %rdx                         "*)
  (* Instr *) |            NR64s        of I_NR64s        * Iargs_S<NR64>                (*"  jmpq            *%rdx                        "*)
  (* Instr *) |            NM64s        of I_NM64s        * Iargs_S<NM64>                (*"  call            *64(%rbx,%rax)               "*)
  (* Instr *) |            R8s          of I_R8s          * Iargs_S<R8>                  (*"  cmpb            $42, %cl                     "*)
  (* Instr *) |            R16s         of I_R16s         * Iargs_S<R16>                 (*"  testw           $42, %cx                     "*)
  (* Instr *) |            R32s         of I_R32s         * Iargs_S<R32>                 (*"  testl           $11, %edx                    "*)
  (* Instr *) |            R64s         of I_R64s         * Iargs_S<R64>                 (*"  testq           $12, %rcx                    "*)
  (* Instr *) |            R8sR8s       of I_R8sR8s       * Iargs_S_S<R8,R8>             (*"  cmpb            %al, %bl                     "*)
  (* Instr *) |            R16sR16s     of I_R16sR16s     * Iargs_S_S<R16,R16>           (*"  cmpw            %ax, %bx                     "*)
  (* Instr *) |            R32sR32s     of I_R32sR32s     * Iargs_S_S<R32,R32>           (*"  cmpl            %eax, %ebx                   "*)
  (* Instr *) |            R64sR64s     of I_R64sR64s     * Iargs_S_S<R64,R64>           (*"  cmpq            %rax, %rbx                   "*)
  (* Instr *) |            M8s          of I_M8s          * Iargs_S<M8>                  (*"  cmpb            $1, (%rdx)                   "*)
  (* Instr *) |            M16s         of I_M16s         * Iargs_S<M16>                 (*"  cmpw            $1, (%rdx)                   "*)
  (* Instr *) |            M32s         of I_M32s         * Iargs_S<M32>                 (*"  cmpl            $1, (%rdx)                   "*)
  (* Instr *) |            M64s         of I_M64s         * Iargs_S<M64>                 (*"  cmpq            $1, (%rdx)                   "*)
  (* Instr *) |            R8sR8d       of I_R8sR8d       * Iargs_S_D<R8,R8>             (*"  movb            %al, %ah                     "*)
  (* Instr *) |            R8sR16d      of I_R8sR16d      * Iargs_S_D<R8,R16>            (*"  movsbw          %al, %bx                     "*)
  (* Instr *) |            R8sR32d      of I_R8sR32d      * Iargs_S_D<R8,R32>            (*"  movsbl          %cl, %eax                    "*)
  (* Instr *) |            R8sR64d      of I_R8sR64d      * Iargs_S_D<R8,R64>            (*"  movsbq          %dl, %rax                    "*)
  (* Instr *) |            R16sR16d     of I_R16sR16d     * Iargs_S_D<R16,R16>           (*"  popcntw         %ax, %bx                     "*)
  (* Instr *) |            R16sR32d     of I_R16sR32d     * Iargs_S_D<R16,R32>           (*"  movswl          %bx, %eax                    "*)
  (* Instr *) |            R16sR64d     of I_R16sR64d     * Iargs_S_D<R16,R64>           (*"  movswq          %bx, %rax                    "*)
  (* Instr *) |            R32sR32d     of I_R32sR32d     * Iargs_S_D<R32,R32>           (*"  bsrl            %eax, %edx                   "*)
  (* Instr *) |            R32sR64d     of I_R32sR64d     * Iargs_S_D<R32,R64>           (*"  movslq          %eax, %rbx                   "*)
  (* Instr *) |            R64sR64d     of I_R64sR64d     * Iargs_S_D<R64,R64>           (*"  popcntq         %eax, %esi                   "*)
  (* Instr *) |            R8sM8d       of I_R8sM8d       * Iargs_S_D<R8,M8>             (*"  movb            %al, (%rbx)                  "*)
  (* Instr *) |            R16sM16d     of I_R16sM16d     * Iargs_S_D<R16,M16>           (*"  movw            %ax, (%rbx)                  "*)
  (* Instr *) |            R32sM32d     of I_R32sM32d     * Iargs_S_D<R32,M32>           (*"  movl            %eax, (%rbx)                 "*)
  (* Instr *) |            R64sM64d     of I_R64sM64d     * Iargs_S_D<R64,M64>           (*"  movq            %rax, (%rbx)                 "*)
  (* Instr *) |            R8dsR8ds     of I_R8dsR8ds     * Iargs_DS_DS<R8,R8>           (*"  xaddb           %al, %bl                     "*)
  (* Instr *) |            R16dsR16ds   of I_R16dsR16ds   * Iargs_DS_DS<R16,R16>         (*"  xaddw           %ax, %bx                     "*)
  (* Instr *) |            R32dsR32ds   of I_R32dsR32ds   * Iargs_DS_DS<R32,R32>         (*"  xaddl           %eax, %ebx                   "*)
  (* Instr *) |            R64dsR64ds   of I_R64dsR64ds   * Iargs_DS_DS<R64,R64>         (*"  xaddq           %rax, %rbx                   "*)
  (* Instr *) |            M8sR8ds      of I_M8sR8ds      * Iargs_S_DS<M8,R8>            (*"  addb            (%rbx), %al                  "*)
  (* Instr *) |            M16sR16ds    of I_M16sR16ds    * Iargs_S_DS<M16,R16>          (*"  addw            (%rbx), %ax                  "*)
  (* Instr *) |            M32sR32ds    of I_M32sR32ds    * Iargs_S_DS<M32,R32>          (*"  addl            (%rbx), %eax                 "*)
  (* Instr *) |            M64sR64ds    of I_M64sR64ds    * Iargs_S_DS<M64,R64>          (*"  addq            (%rbx), %rax                 "*)
  (* Instr *) |            R8sM8ds      of I_R8sM8ds      * Iargs_S_DS<R8,M8>            (*"  addb            %al, (%rbx)                  "*)
  (* Instr *) |            R16sM16ds    of I_R16sM16ds    * Iargs_S_DS<R16,M16>          (*"  addw            %ax, (%rbx)                  "*)
  (* Instr *) |            R32sM32ds    of I_R32sM32ds    * Iargs_S_DS<R32,M32>          (*"  addl            %eax, (%rbx)                 "*)
  (* Instr *) |            R64sM64ds    of I_R64sM64ds    * Iargs_S_DS<R64,M64>          (*"  addq            %rax, (%rbx)                 "*)
  (* Instr *) |            R8sR16sM16ds of I_R8sR16sM16ds * Iargs_S_S_DS<R8,R16,M16>     (*"  shrdw           %cl, %bx, (%rdx)             "*)
  (* Instr *) |            R8sR32sM32ds of I_R8sR32sM32ds * Iargs_S_S_DS<R8,R32,M32>     (*"  shrdl           %cl, %ebx, (%rdx)            "*)
  (* Instr *) |            R8sR64sM64ds of I_R8sR64sM64ds * Iargs_S_S_DS<R8,R64,M64>     (*"  shrdq           %cl, %rbx, (%rdx)            "*)
  (* Instr *) |            M8dsR8ds     of I_M8dsR8ds     * Iargs_DS_DS<M8,R8>           (*"  xaddb           (%rbx), %al                  "*)
  (* Instr *) |            M16dsR16ds   of I_M16dsR16ds   * Iargs_DS_DS<M16,R16>         (*"  xaddw           (%rbx), %ax                  "*)
  (* Instr *) |            M32dsR32ds   of I_M32dsR32ds   * Iargs_DS_DS<M32,R32>         (*"  xaddl           (%rbx), %eax                 "*)
  (* Instr *) |            M64dsR64ds   of I_M64dsR64ds   * Iargs_DS_DS<M64,R64>         (*"  xaddq           (%rbx), %rax                 "*)
  (* Instr *) |            R8sR16sR16ds of I_R8sR16sR16ds * Iargs_S_S_DS<R8,R16,R16>     (*"  shldw           %cl, %ax, %bx                "*)
  (* Instr *) |            R8sR32sR32ds of I_R8sR32sR32ds * Iargs_S_S_DS<R8,R32,R32>     (*"  shldl           %cl, %eax, %ebx              "*)
  (* Instr *) |            R8sR64sR64ds of I_R8sR64sR64ds * Iargs_S_S_DS<R8,R64,R64>     (*"  shldq           %cl, %rax, %rbx              "*)
  (* Instr *) |            R8dsM8ds     of I_R8dsM8ds     * Iargs_DS_DS<R8,M8>           (*"  xaddb           %al, (%rbx)                  "*)
  (* Instr *) |            R16dsM16ds   of I_R16dsM16ds   * Iargs_DS_DS<R16,M16>         (*"  xaddw           %ax, (%rbx)                  "*)
  (* Instr *) |            R32dsM32ds   of I_R32dsM32ds   * Iargs_DS_DS<R32,M32>         (*"  xaddl           %eax, (%rbx)                 "*)
  (* Instr *) |            R64dsM64ds   of I_R64dsM64ds   * Iargs_DS_DS<R64,M64>         (*"  xaddq           %rax, (%rbx)                 "*)
  (* Instr *) |            R8sR8ds      of I_R8sR8ds      * Iargs_S_DS<R8,R8>            (*"  addb            %al, %cl                     "*)
  (* Instr *) |            R8sR16ds     of I_R8sR16ds     * Iargs_S_DS<R8,R16>           (*"  rolw            %cl, %ax                     "*)
  (* Instr *) |            R8sR32ds     of I_R8sR32ds     * Iargs_S_DS<R8,R32>           (*"  shll            %cl, %eax                    "*)
  (* Instr *) |            R8sR64ds     of I_R8sR64ds     * Iargs_S_DS<R8,R64>           (*"  shrq            %cl, %rax                    "*)
  (* Instr *) |            R16sR16ds    of I_R16sR16ds    * Iargs_S_DS<R16,R16>          (*"  addw            %ax, %bx                     "*)
  (* Instr *) |            R32sR32ds    of I_R32sR32ds    * Iargs_S_DS<R32,R32>          (*"  addl            %eax, %ebx                   "*)
  (* Instr *) |            R64sR64ds    of I_R64sR64ds    * Iargs_S_DS<R64,R64>          (*"  addq            %rax, %rbx                   "*)
  (* Instr *) |            M8sR8d       of I_M8sR8d       * Iargs_S_D<M8,R8>             (*"  movb            (%rbx), %al                  "*)
  (* Instr *) |            M8sR16d      of I_M8sR16d      * Iargs_S_D<M8,R16>            (*"  movsbw          (%rbx), %ax                  "*)
  (* Instr *) |            M8sR32d      of I_M8sR32d      * Iargs_S_D<M8,R32>            (*"  movsbl          (%rbx), %eax                 "*)
  (* Instr *) |            M8sR64d      of I_M8sR64d      * Iargs_S_D<M8,R64>            (*"  movsbq          (%rbx), %rax                 "*)
  (* Instr *) |            M16sR16d     of I_M16sR16d     * Iargs_S_D<M16,R16>           (*"  movw            (%rbx), %ax                  "*)
  (* Instr *) |            M16sR32d     of I_M16sR32d     * Iargs_S_D<M16,R32>           (*"  movswl          (%rbx), %eax                 "*)
  (* Instr *) |            M16sR64d     of I_M16sR64d     * Iargs_S_D<M16,R64>           (*"  movswq          (%rbx), %rsi                 "*)
  (* Instr *) |            M32sR32d     of I_M32sR32d     * Iargs_S_D<M32,R32>           (*"  movl            (%rbx), %eax                 "*)
  (* Instr *) |            M32sR64d     of I_M32sR64d     * Iargs_S_D<M32,R64>           (*"  movslq          (%rbx), %rax                 "*)
  (* Instr *) |            M64sR64d     of I_M64sR64d     * Iargs_S_D<M64,R64>           (*"  movq            (%rbx), %rsi                 "*)
  (* Instr *) |            R8sM8s       of I_R8sM8s       * Iargs_S_S<R8,M8>             (*"  cmpb            %al, (%rbx)                  "*)
  (* Instr *) |            R16sM16s     of I_R16sM16s     * Iargs_S_S<R16,M16>           (*"  cmpw            %ax, (%rbx)                  "*)
  (* Instr *) |            R32sM32s     of I_R32sM32s     * Iargs_S_S<R32,M32>           (*"  cmpl            %eax, (%rbx)                 "*)
  (* Instr *) |            R64sM64s     of I_R64sM64s     * Iargs_S_S<R64,M64>           (*"  cmpq            %rax, (%rbx)                 "*)
  (* Instr *) |            M8sR8s       of I_M8sR8s       * Iargs_S_S<M8,R8>             (*"  cmpb            (%rbx), %al                  "*)
  (* Instr *) |            M16sR16s     of I_M16sR16s     * Iargs_S_S<M16,R16>           (*"  cmpw            (%rbx), %ax                  "*)
  (* Instr *) |            M32sR32s     of I_M32sR32s     * Iargs_S_S<M32,R32>           (*"  cmpl            (%rbx), %eax                 "*)
  (* Instr *) |            M64sR64s     of I_M64sR64s     * Iargs_S_S<M64,R64>           (*"  cmpq            (%rbx), %rdx                 "*)
  (* Instr *) | [< Vec >]  R8sXsXd      of I_R8sXsXd      * Iargs_S_S_D<R8,X,X>          (*"  vpinsrb         $1, %al, %xmm1, %xmm2        "*)
  (* Instr *) | [< Vec >]  R16sXsXd     of I_R16sXsXd     * Iargs_S_S_D<R16,X,X>         (*"  vpinsrw         $0, %bx, %xmm2, %xmm3        "*)
  (* Instr *) | [< Vec >]  R32sXsXd     of I_R32sXsXd     * Iargs_S_S_D<R32,X,X>         (*"  vpinsrd         $1, %eax, %xmm0, %xmm1       "*)
  (* Instr *) | [< Vec >]  R64sXsXd     of I_R64sXsXd     * Iargs_S_S_D<R64,X,X>         (*"  vpinsrq         $1, %rax, %xmm0, %xmm3       "*)
  (* Instr *) | [< Vec >]  R8sXds       of I_R8sXds       * Iargs_S_DS<R8,X>             (*"  pinsrb          $1, %al, %xmm1               "*)
  (* Instr *) | [< Vec >]  R16sXds      of I_R16sXds      * Iargs_S_DS<R16,X>            (*"  pinsrw          $0, %ax, %xmm2               "*)
  (* Instr *) | [< Vec >]  R32sXds      of I_R32sXds      * Iargs_S_DS<R32,X>            (*"  pinsrd          $1, %eax, %xmm3              "*)
  (* Instr *) | [< Vec >]  R64sXds      of I_R64sXds      * Iargs_S_DS<R64,X>            (*"  pinsrq          $0, %rax, %xmm2              "*)
  (* Instr *) | [< Vec >]  XsR32d       of I_XsR32d       * Iargs_S_D<X,R32>             (*"  pextrb          $1, %xmm1, %eax              "*)
  (* Instr *) | [< Vec >]  XsR64d       of I_XsR64d       * Iargs_S_D<X,R64>             (*"  pextrq          $0, %xmm0, %rax              "*)
  (* Instr *) | [< Vec >]  R32sXd       of I_R32sXd       * Iargs_S_D<R32,X>             (*"  movd            %eax, %xmm0                  "*)
  (* Instr *) | [< Vec >]  R64sXd       of I_R64sXd       * Iargs_S_D<R64,X>             (*"  movq            %rax, %xmm1                  "*)
  (* Instr *) | [< Vec >]  M16sXd       of I_M16sXd       * Iargs_S_D<M16,X>             (*"  pmovsxbq        (%rbx), %xmm0                "*)
  (* Instr *) | [< Vec >]  M32sXd       of I_M32sXd       * Iargs_S_D<M32,X>             (*"  movd            (%rbx), %xmm0                "*)
  (* Instr *) | [< Vec >]  M64sXd       of I_M64sXd       * Iargs_S_D<M64,X>             (*"  movq            (%rbx), %xmm1                "*)
  (* Instr *) | [< Vec >]  M128sXd      of I_M128sXd      * Iargs_S_D<M128,X>            (*"  movapd          (%rbx), %xmm8                "*)
  (* Instr *) | [< Vec >]  XsM8d        of I_XsM8d        * Iargs_S_D<X,M8>              (*"  pextrb          $0, %xmm0, (%rbx)            "*)
  (* Instr *) | [< Vec >]  XsM16d       of I_XsM16d       * Iargs_S_D<X,M16>             (*"  pextrw          $0, %xmm0, (%rbx)            "*)
  (* Instr *) | [< Vec >]  XsM32d       of I_XsM32d       * Iargs_S_D<X,M32>             (*"  pextrd          $0, %xmm0, (%rbx)            "*)
  (* Instr *) | [< Vec >]  XsM64d       of I_XsM64d       * Iargs_S_D<X,M64>             (*"  pextrq          $0, %xmm0, (%rbx)            "*)
  (* Instr *) | [< Vec >]  XsM128d      of I_XsM128d      * Iargs_S_D<X,M128>            (*"  movaps          %xmm0, (%rbx)                "*)
  (* Instr *) | [< Vec >]  M32sXs       of I_M32sXs       * Iargs_S_S<M32,X>             (*"  comiss          (%rax), %xmm1                "*)
  (* Instr *) | [< Vec >]  M64sXs       of I_M64sXs       * Iargs_S_S<M64,X>             (*"  comisd          (%rbx), %xmm2                "*)
  (* Instr *) | [< Vec >]  M128sXs      of I_M128sXs      * Iargs_S_S<M128,X>            (*"  ptest           (%rbx), %xmm1                "*)
  (* Instr *) | [< Vec >]  M8sXds       of I_M8sXds       * Iargs_S_DS<M8,X>             (*"  pinsrb          $1, (%rbx), %xmm1            "*)
  (* Instr *) | [< Vec >]  M16sXds      of I_M16sXds      * Iargs_S_DS<M16,X>            (*"  pinsrw          $0, (%rbx), %xmm1            "*)
  (* Instr *) | [< Vec >]  M32sXds      of I_M32sXds      * Iargs_S_DS<M32,X>            (*"  subss           (%rbx), %xmm1                "*)
  (* Instr *) | [< Vec >]  M64sXds      of I_M64sXds      * Iargs_S_DS<M64,X>            (*"  subsd           (%rbx), %xmm1                "*)
  (* Instr *) | [< Vec >]  M128sXds     of I_M128sXds     * Iargs_S_DS<M128,X>           (*"  addps           (%rbx), %xmm3                "*)
  (* Instr *) | [< Vec >]  Xds          of I_Xds          * Iargs_DS<X>                  (*"  psllw           $1, %xmm1                    "*)
  (* Instr *) | [< Vec >]  XsXd         of I_XsXd         * Iargs_S_D<X,X>               (*"  movaps          %xmm2, %xmm3                 "*)
  (* Instr *) | [< Vec >]  XsXds        of I_XsXds        * Iargs_S_DS<X,X>              (*"  addpd           %xmm10, %xmm9                "*)
  (* Instr *) | [< Vec >]  XsXs         of I_XsXs         * Iargs_S_S<X,X>               (*"  comisd          %xmm0, %xmm1                 "*)
  (* Instr *) | [< Vec >]  XsXsM128ds   of I_XsXsM128ds   * Iargs_S_S_DS<X,X,M128>       (*"  vmaskmovps      %xmm1, %xmm2, (%rbx)         "*)
  (* Instr *) | [< Vec >]  XsXsXd       of I_XsXsXd       * Iargs_S_S_D<X,X,X>           (*"  vaddps          %xmm0, %xmm1, %xmm2          "*)
  (* Instr *) | [< Vec >]  M8sXsXd      of I_M8sXsXd      * Iargs_S_S_D<M8,X,X>          (*"  vpinsrb         $2, (%rbx), %xmm1, %xmm2     "*)
  (* Instr *) | [< Vec >]  M16sXsXd     of I_M16sXsXd     * Iargs_S_S_D<M16,X,X>         (*"  vpinsrw         $4, (%rbx), %xmm1, %xmm2     "*)
  (* Instr *) | [< Vec >]  M32sXsXd     of I_M32sXsXd     * Iargs_S_S_D<M32,X,X>         (*"  vaddss          (%rbx), %xmm1, %xmm2         "*)
  (* Instr *) | [< Vec >]  M64sXsXd     of I_M64sXsXd     * Iargs_S_S_D<M64,X,X>         (*"  vaddsd          (%rbx), %xmm1, %xmm2         "*)
  (* Instr *) | [< Vec >]  M128sXsXd    of I_M128sXsXd    * Iargs_S_S_D<M128,X,X>        (*"  vaddps          (%rbx), %xmm1, %xmm2         "*)
  (* Instr *) | [< Vec >]  XsM128sXd    of I_XsM128sXd    * Iargs_S_S_D<X,M128,X>        (*"  vprotw          %xmm0, (%rbx), %xmm1         "*)
  (* Instr *) | [< Vec >]  XsXsXsXd     of I_XsXsXsXd     * Iargs_S_S_S_D<X,X,X,X>       (*"  vfmaddpd        %xmm1, %xmm2, %xmm3, %xmm4   "*)
  (* Instr *) | [< Vec >]  M32sXsXsXd   of I_M32sXsXsXd   * Iargs_S_S_S_D<M32,X,X,X>     (*"  vfmaddss        (%rbx), %xmm1, %xmm5, %xmm6  "*)
  (* Instr *) | [< Vec >]  M64sXsXsXd   of I_M64sXsXsXd   * Iargs_S_S_S_D<M64,X,X,X>     (*"  vfmaddsd        (%rbx), %xmm1, %xmm5, %xmm6  "*)
  (* Instr *) | [< Vec >]  M128sXsXsXd  of I_M128sXsXsXd  * Iargs_S_S_S_D<M128,X,X,X>    (*"  vfmaddpd        (%rbx), %xmm1, %xmm5, %xmm6  "*)
  (* Instr *) | [< Vec >]  XsM32sXsXd   of I_XsM32sXsXd   * Iargs_S_S_S_D<X,M32,X,X>     (*"  vfmaddss        %xmm1, (%rbx), %xmm3, %xmm4  "*)
  (* Instr *) | [< Vec >]  XsM64sXsXd   of I_XsM64sXsXd   * Iargs_S_S_S_D<X,M64,X,X>     (*"  vfmaddsd        %xmm1, (%rbx), %xmm3, %xmm4  "*)
  (* Instr *) | [< Vec >]  XsM128sXsXd  of I_XsM128sXsXd  * Iargs_S_S_S_D<X,M128,X,X>    (*"  vfmaddpd        %xmm1, (%rbx), %xmm3, %xmm4  "*)
  (* Instr *) | [< Vec >]  YsYd         of I_YsYd         * Iargs_S_D<Y,Y>               (*"  vmovapd         %ymm1, %ymm2                 "*)
  (* Instr *) | [< Vec >]  M32sYd       of I_M32sYd       * Iargs_S_D<M32,Y>             (*"  vbroadcastss    (%rbx), %ymm1                "*)
  (* Instr *) | [< Vec >]  M64sYd       of I_M64sYd       * Iargs_S_D<M64,Y>             (*"  vbroadcastsd    (%rbx), %ymm2                "*)
  (* Instr *) | [< Vec >]  M128sYd      of I_M128sYd      * Iargs_S_D<M128,Y>            (*"  vbroadcastf128  (%rbx), %ymm3                "*)
  (* Instr *) | [< Vec >]  M256sYd      of I_M256sYd      * Iargs_S_D<M256,Y>            (*"  vmovapd         (%rbx), %ymm1                "*)
  (* Instr *) | [< Vec >]  YsM128d      of I_YsM128d      * Iargs_S_D<Y,M128>            (*"  vextractf128    $0, %ymm1, (%rbx)            "*)
  (* Instr *) | [< Vec >]  YsM256d      of I_YsM256d      * Iargs_S_D<Y,M256>            (*"  vmovaps         %ymm9, (%rdx)                "*)
  (* Instr *) | [< Vec >]  YsYsYd       of I_YsYsYd       * Iargs_S_S_D<Y,Y,Y>           (*"  vminps          %ymm1, %ymm3, %ymm5          "*)
  (* Instr *) | [< Vec >]  M256sYsYd    of I_M256sYsYd    * Iargs_S_S_D<M256,Y,Y>        (*"  vaddpd          (%rbx), %ymm1, %ymm3         "*)
  (* Instr *) | [< Vec >]  XsYsYd       of I_XsYsYd       * Iargs_S_S_D<X,X,Y>           (*"  vinsertf128     $0, %xmm1, %ymm5             "*)
  (* Instr *) | [< Vec >]  YsXd         of I_YsXd         * Iargs_S_D<Y,X>               (*"  vextractf128    $1, %ymm1, %xmm3             "*)
  (* Instr *) | [< Vec >]  YsR32d       of I_YsR32d       * Iargs_S_D<Y,R32>             (*"  vmovmskpsl      %ymm1, (%rbx)                "*)
  (* Instr *) | [< Vec >]  YsR64d       of I_YsR64d       * Iargs_S_D<Y,R64>             (*"  vmovmskpdq      %ymm3, (%rcx)                "*)
with static member Nop = I0(I_I0.Nop)

(***)

let GetImpliedLiteral = let lut = new Dictionary<string * int, string * int>()
                        for pfx  (* arity 1   *) in ["sa";"sh";"ro";"rc"] do for lr in ["l";"r"] do for sfx in ["";"b";"w";"l";"q"] do let s = pfx+lr+sfx in lut.Add((s,1),(s,1))
                        let pfx  (* arity 2   *) = "sh" in for lr in ["l";"r"] do for sfx in ["";"l";"q"] do let s = pfx+lr+"d"+sfx in lut.Add((s,2),(s,1))
                        for (pfx,(* arity 2|3 *) n) in [("pcmul",2);("vpcmul",3)] do for (sfx,l) in [("lqlqdq",0); ("hqlqdq",1); ("lqhdq",16); ("hqhdq",17)] do lut.Add((pfx+sfx,n),(pfx+"qdq",l))
                        for (pfx,(* arity 2|3 *) n) in [("cmp",2);("vcmp",3)] do ["eq";"lt";"le";"unord";"neq";"nlt";"nle";"ord"] |> List.iteri (fun l mid -> for sfx in ["ss";"sd";"ps";"pd"] do lut.Add((pfx+mid+sfx,n),(pfx+sfx,l)))
                        let pfx  (* arity 3   *) = "vpcom" in ["lt";"le";"gt";"ge";"eq";"neq";"false";"true"] |> List.iteri (fun l mid -> for sfx in ["b";"ub";"d";"ud";"q";"uq";"w";"uw"] do lut.Add((pfx+mid+sfx,3),(pfx+sfx,l)))

                        fun name arity -> let (found,value) = lut.TryGetValue((name,arity)) in if found then Some(value) else None

(***)

(* test if an instruction has no effect at all (ie. if it qualifies as a nop) *)
(* [remark] many ALU instructions alter the flags as a side-effect; these don't count as nops even if they don't do anything else *)
let Instr_has_no_effect = function R8dsR8ds(I_R8dsR8ds.Xchgb,     {Iargs_DS_DS.DS1=s1; Iargs_DS_DS.DS2=s2}) -> s1 = s2
                                 | R16dsR16ds(I_R16dsR16ds.Xchgw, {Iargs_DS_DS.DS1=s1; Iargs_DS_DS.DS2=s2}) -> s1 = s2
                                 | R64dsR64ds(I_R64dsR64ds.Xchgq, {Iargs_DS_DS.DS1=s1; Iargs_DS_DS.DS2=s2}) -> s1 = s2
                                 | R8sR8d(I_R8sR8d.Movb,     {Iargs_S_D.S=s; Iargs_S_D.D=d}) -> s = d
                                 | R16sR16d(I_R16sR16d.Movw, {Iargs_S_D.S=s; Iargs_S_D.D=d}) -> s = d
                                 | R64sR64d(I_R64sR64d.Movq, {Iargs_S_D.S=s; Iargs_S_D.D=d}) -> s = d
                                 | R64d(I_R64d.Leaq(EA64(Zero,Some({Base=s}),None)), {Iargs_D.D=d}) -> s = d
                                 | XsXd((I_XsXd.Movapd | I_XsXd.Movaps | I_XsXd.Vmovapd | I_XsXd.Vmovaps), {Iargs_S_D.S=s; Iargs_S_D.D=d}) -> s = d
                                 | XsXd((I_XsXd.Movdqa | I_XsXd.Movdqu | I_XsXd.Vmovdqa | I_XsXd.Vmovdqu), {Iargs_S_D.S=s; Iargs_S_D.D=d}) -> s = d
                                 | XsXd((I_XsXd.Movupd | I_XsXd.Movups | I_XsXd.Vmovupd | I_XsXd.Vmovups), {Iargs_S_D.S=s; Iargs_S_D.D=d}) -> s = d
                                 | _ -> false

(***)

let stackOfsOption (m:M) : option<Int32> =
                   match m
                    with | EA64(Iexpr.Int(o), Some{Base=r}, None)                   when r = R64.rsp            -> (* check that it is negative *) Some(int(o))
                         | EA64(_,            _,            Some{Index=i;Scale=s})  when i = R64.rsp && s <> 1L -> failwith "stackLocOption: EA64 with rsp as index reg"
                         | EA64(Iexpr.Int(o), None,         Some{Index=i;Scale=1L}) when i = R64.rsp            -> (* check that it is negative *) Some(int(o))
                         | EA32(Iexpr.Int(o), Some{Base=r}, None)                   when r = R32.esp            -> (* check that it is negative *) Some(int(o))
                         | EA32(_,            _,            Some{Index=i;Scale=s})  when i = R32.esp && s <> 1L -> failwith "stackLocOption: EA32 with esp as index reg"
                         | EA32(Iexpr.Int(o), None,         Some{Index=i;Scale=1L}) when i = R32.esp            -> (* check that it is negative *) Some(int(o))
                         | _ -> None

let stackLocOption (size,m) =
                   match stackOfsOption(m)
                    with | Some(ofs) -> Some(size,ofs)
                         | None -> None

let (|StackLoc|_|) x = stackLocOption(x)

